:- use_module(library(fastrw)).

go :-
    prolog_flag(argv,Args),
    ( open_streams(Args,In,Out) ->
        call_cleanup(inspect(In,Out),(flush_output(Out),flush_output(user_error)))
    ; otherwise ->
        write(user_error,'Usage: inspect <file_name>\n'),
        fail).

open_streams([],user_input,user_output).
open_streams([InputFilename],S,user_output) :-
    open(InputFilename,read,S,[type(binary)]).
open_streams([InputFilename,OutputFilename],In,Out) :-
    open_streams([InputFilename],In,_),
    open(OutputFilename,write,Out).

inspect(In,Out) :-
    catch( ( fast_read(In,Term), writeq(Out,Term), write(Out,'.\n') ),
           error(E,_),
           ( E=syntax_error(_) -> 
               format(user_error,'Consistency error when reading from stream: ~w~n',[E]),
               fail
           ; E=permission_error(_,_,_) ->
               End=true % end of file
           ; E=consistency_error(_,_,_) ->
               format(user_error,'Consistency error when reading from stream: ~w~n',[E]),
               fail
           ; otherwise ->
               format(user_error,'Unknown error when reading from stream: ~w~n',[E]),
               fail
           )
         ),
    ( End == true -> true
    ; otherwise ->
        inspect(In,Out)).

runtime_entry(start) :- go.

save :- save_program(inspect).
