#include <stdio.h>
#include <stdlib.h>

#include "ltlarray.h"

static const unsigned int max = 100000;
static long convert(unsigned int x);
static void check_array(ltlarray * array);
static void initf(void *);
static const long testvalue = 34754;

int main(int argc, char *argv[]) {
  unsigned int i;
  long *elem;
  ltlarray array;

  init_array(&array, sizeof(long), initf, "array1");
  for (i=0; i<max; i++) {
    assure_arraysize(&array, i+1);
    elem = (long *) get_array_element(&array, i);
    if (*elem != testvalue) {
      printf("expected initial value %li, but was %li for element %u\n", testvalue, *elem, i);
      exit(1);
    }
    *elem = convert(i);
  }
  check_array(&array);
  delete_array(&array);

  init_array(&array, sizeof(long), 0, "array2");
  for (i=max; i>0; i--) {
    assure_arraysize(&array, i);
    elem = (long *) get_array_element(&array, i-1);
    *elem = convert(i-1);
  }
  check_array(&array);
  delete_array(&array);

  printf("ltlarray: test passed.\n");
  return 0;
}

long convert(unsigned int x) {
  return ((long) x) * 41L;
}

void initf(void *mem) {
  long *p = (long*) mem;
  *p = testvalue;
}

void check_array(ltlarray *array) {
  unsigned int i;
  long *elem,test;
  for (i=0; i<max; i++) {
    test = convert(i);
    elem = (long *) get_array_element(array, i);
    if (!elem) {
      printf("error: get_array_element() returned NULL for element %u.\n", i);
      exit(1);
    }
    if (*elem != test) {
      printf("error: expected %li, was %li for element %u.\n", test, *elem, i);
      exit(1);
    }
  }
}
