#include <sys/stat.h>
#include <stdlib.h>

//#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <string.h>

#include "mapper.h"
#include "queue.h"
#include "hash.h"

#include <dirent.h>

#include "../lib/include/zmq.h"
#include "../lib/include/czmq.h"

//static void work_file(mmap_env *env, size_t pos, const size_t flen, wQueue *q, size_t bufsize);
static void work_file(mmap_env *env, const size_t flen, wQueue *q, size_t bufsize);
static void process_file(char *filename, wQueue *q);





#include <stdio.h>


static zmsg_t *msg = NULL;

wQueue *recover_queues(char *dirname, zctx_t *ctx, char *masteraddr, int port_start) {
    // TODO: known, unchecked hashes have to be propagated to the master
    msg = zmsg_new();
    wQueue *q = init_queue(0);

    DIR *dir = opendir(dirname);
    assert(dir);
    struct dirent *ent;

    while ((ent = readdir(dir)) != NULL) {
        char *fname = ent->d_name;
        if (strncmp(fname, "zmq-", 4) == 0 || strncmp(fname, "zmqw-", 5) == 0) {
            int len = strlen(fname) + strlen(dirname) + 2;
            char *buf = malloc(len);
            memset(buf, 0, len);

            strcat(buf, dirname);
            if (dirname[strlen(dirname) -1] != '/') {
                strcat(buf, "/");
            }
            strcat(buf, fname);

            if (0 != strcmp(buf, q->env->mmap_env->filename)) {
                process_file(buf, q);
            }
            free(buf);
        }
    }
    closedir(dir);
    void *zocket = zsocket_new(ctx, ZMQ_REQ);
    zsocket_connect(zocket, "tcp://%s:%d", masteraddr, port_start + 8); // TODO: what happens if socket is not bound yet?
    if (zmsg_size(msg) == 0) {
        zmsg_addstr(msg, "empty");
    }
    zmsg_send(&msg, zocket);
    zmsg_destroy(&msg);
    char *ok = zstr_recv(zocket);
    assert(0 == strcmp(ok, "ok"));
    free(ok);
    zsocket_destroy(ctx, zocket);
    printf("queuelength: %d\n", q->c);

    return q;
}


wQueue *recover_file(char *filename) {
    wQueue *q = init_queue(0);

    process_file(filename, q);

    return q;
}


static void process_file(char *filename, wQueue *q) {
    struct stat st;
    stat(filename, &st);
    size_t len = st.st_size;

    struct flock lock;
    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;

    int fd = open(filename, O_RDWR);
    assert(fd >= 0);


    int rc = fcntl(fd, F_SETLK, &lock);
    if (rc == -1) {
        // most likely somebody else locked the file
        return;
    }

    size_t initial_bufsize = 1048576;

    mmap_env *env = mapper_init(fd, filename);
    work_file(env, len, q, initial_bufsize);

    ftruncate(fd, 0);
    lock.l_type = F_UNLCK;
    rc = fcntl(fd, F_SETLK, &lock);
    assert(rc != -1);
}


static void work_file(mmap_env *env, const size_t flen, wQueue *q, size_t bufsize) {
    int counta = 0;
    int corrupted = 0;
    size_t pos = 0;
    while (pos < flen) {
        size_t end;
        if (pos + bufsize <= flen) {
            end = pos + bufsize;
        } else {
            end = flen;
        }
        char *content = get_in_mmap(env, pos, bufsize, 0);

        size_t offset = 0;
        // skip many zeros
        int count = 0;
        while (pos + offset + sizeof(char*) < end) {
            if (((char**) content + count) != NULL) {
                break;
            }
            count++;
            offset += sizeof(char*);
        }
        
        // skip 8 zeros at max
        while (pos + offset < end && content[offset] == 0)  {
            offset++;
        }

        if (pos + offset >= end || content[offset] != 'X') {
            pos += offset;
            continue;
        }

        // we found a state
        content = get_in_mmap(env, pos+offset, 1 + sizeof(size_t), 0);
        size_t len = *((size_t*) (content + 1));
        offset += 1 + sizeof(size_t);

        if (TRIE_HASHLEN + len > bufsize) {
            bufsize = TRIE_HASHLEN + len;
        }

        content = get_in_mmap(env, pos+offset, TRIE_HASHLEN + len, 0);
        char *digest = (char*) malloc(TRIE_HASHLEN);
        char *term = (char*) malloc(len);

        memcpy(digest, content, TRIE_HASHLEN);
        memcpy(term, content + TRIE_HASHLEN, len);
        char buf[TRIE_HASHLEN];
        hash(term, len, buf);

        if (0 == memcmp(buf, digest, TRIE_HASHLEN)) {
            zmsg_addmem(msg, digest, TRIE_HASHLEN);
            enqueue(q, term, len, digest);
            counta++;
        } else {
            // e.g. if a hdd is full and not the whole state is written
            assert(corrupted == 0);
            puts("encountered a term that does not match its SHA-1 digest");
            corrupted++;
        }
        offset += TRIE_HASHLEN + len;
        pos += offset;
    }

    printf("recovered %d wps from file %s\n", counta, env->filename);
}

//int main(int argc, char *argv[]) {
//    assert(argc == 2);
//    wQueue *q = recover_queues(argv[1]);
//    printf("%d\n", q_size(q));
//    teardown_queue(&q);
//    return 0;
//}
