#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <semaphore.h>
#include <fcntl.h>
#include "shm_common.h"
#include "shm_srv.h"

static char *semap_names[32];
static char *semap_multiwrite_names[32];
static int names_initialised = 0;

static void init_semaphore_names(void) {
    if (names_initialised) {
        return;
    }
    int i;
    for (i = 0; i < 32; i++) {
        int len = strlen("/hashes-semap_nodes") + 3; // "31" + zero termination
        semap_names[i] = (char*) malloc(len);
        sprintf(semap_names[i], "/hashes-semap_nodes%d", i);
        len = strlen("/hashes-semap_meta") + 3;
        semap_multiwrite_names[i] = (char*) malloc(len);
        sprintf(semap_multiwrite_names[i], "/hashes-semap_meta%d", i);
    }
    names_initialised = 1;
}

static void rmsem(void) {
    if (!names_initialised) {
        init_semaphore_names();
    }
    sem_unlink("/hashes-semap_infos");
    sem_unlink("/hashes-semap_digests");
    sem_unlink("/hashes-semap_digestswrite");
    int i;
    for (i = 0; i < 32; i++) {
        sem_unlink(semap_names[i]);
        sem_unlink(semap_multiwrite_names[i]);
    }
}

struct semap_env *create_semaphores(void) {
    init_semaphore_names();

    rmsem();

    struct semap_env *r = (struct semap_env *) malloc(sizeof(struct semap_env));
    r->semap_meta   = sem_open("/hashes-semap_infos", O_CREAT | O_EXCL, 0666, 1);
    r->semap_hashes = sem_open("/hashes-semap_digests", O_CREAT | O_EXCL, 0666, SEMAP_DIGESTS_VAL);
    r->semap_hashwrite = sem_open("/hashes-semap_digestswrite", O_CREAT | O_EXCL, 0666, 1);
    assert(r->semap_meta   != SEM_FAILED);
    assert(r->semap_hashes != SEM_FAILED);
    assert(r->semap_hashwrite != SEM_FAILED);
    int i;
    for (i = 0; i < 32; i++) {
        r->semap_nodes[i] = sem_open(semap_names[i], O_CREAT | O_EXCL, 0666, SEMAP_NODES_VAL);
        assert(r->semap_nodes[i] != SEM_FAILED);
        r->semap_multiwrite[i] = sem_open(semap_multiwrite_names[i], O_CREAT | O_EXCL, 0666, 1);
        assert(r->semap_multiwrite[i] != SEM_FAILED);
    }

    return r;
}

void init_trie(void) {
    sem_wait(semaphores->semap_meta);
    memset(shm->shm_meta_infos->ptr, 0, sizeof(struct hash_meta_infos));

//  this pointer should be initialized already
//  tNode *shm_nodes_ptr = (tNode*) mmap(NULL, 33 * sizeof(tNode), PROT_WRITE, MAP_SHARED, fd_nodes, 0);

    int i;
    tNode *root = (tNode*) (shm->shm_nodes[0]->ptr);
    for (i = 0; i < 32; i++) {
        // init node
        tNode *node = root + i + 1;
        memset(node->indices, 0xFF, 32 * sizeof(index_t));
        ((struct hash_meta_infos*) (shm->shm_meta_infos->ptr))->pos_nodes++;

        index_t index = id2index(i+1);
        root->indices[i] = index; // set successors in root node
    }
    ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_nodes++;
    sem_post(semaphores->semap_meta);
}

void shm_srv_cleanup(struct semap_env *semaphores) {
    if (!names_initialised) {
        init_semaphore_names();
    }
    int rc = sem_close(semaphores->semap_meta);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashes);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashwrite);
    assert(rc == 0);
    int i;
    for (i = 0; i < 32; i++) {
        rc = sem_close(semaphores->semap_nodes[i]);
        assert(rc == 0);
        rc = sem_close(semaphores->semap_multiwrite[i]);
        assert(rc == 0);
    }
    rc = sem_unlink("/hashes-semap_infos");
    assert(rc == 0);
    rc = sem_unlink("/hashes-semap_digests");
    assert(rc == 0);
    rc = sem_unlink("/hashes-semap_digestswrite");
    assert(rc == 0);
    for (i = 0; i < 32; i++) {
        rc = sem_unlink(semap_names[i]);
        rc = sem_unlink(semap_multiwrite_names[i]);
        assert(rc == 0);
        free(semap_names[i]);
        free(semap_multiwrite_names[i]);
    }
    names_initialised = 0;
}
