% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_records,[is_a_record_wf/3, not_is_a_record_wf/3,
    construct_record/2,
    access_record/3,
    normalise_record_type/2, record_has_multiple_field_names/2,
    check_field_name_compatibility/2
    ]).


:- use_module(self_check).
:- use_module(debug).
:- use_module(tools).
:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module contains predicates that operate on the record type.').

:- assert_must_succeed(( kernel_records:normalise_record_type(
 record([field(name,global('ID')),field(balance,integer)]),X),
 X= record([field(balance,integer),field(name,global('ID'))])  )).
:- assert_must_succeed(( kernel_records:normalise_record_type(
  record([field(balance,integer),field(name,global('ID'))]),X),
 X= record([field(balance,integer),field(name,global('ID'))]) )).
:- assert_must_succeed(( kernel_records:normalise_record_type(
  record([field(b,d),field(a,c)]),X),
 X= record([field(a,c),field(b,d)]) )).

normalise_record_type(record(F),record(SF)) :-
   sort_fields(F,SF).

:- use_module(kernel_waitflags).
:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
       kernel_records:is_a_record_wf(rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
              rec([field(balance,global_set('NAT')),field(name,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
       kernel_records:is_a_record_wf(rec([field(a,int(0)),field(b,fd(2,'Name'))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(0))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).


is_a_record_wf(rec(Fields),rec(FieldSets),WF) :-  %print_message(check_is_a_record(Record,FieldSets)),
   check_field_values(Fields,FieldSets,WF).


:-use_module(kernel_objects,[check_element_of_wf/3, membership_test_wf/4]).

:- block check_field_values(-,-,?).
check_field_values([],[],_WF).
check_field_values([field(Name,V)|ValueRest],[field(Name,VT)|TypeRest],WF) :-
  kernel_objects:check_element_of_wf(V,VT,WF),
  check_field_values(ValueRest,TypeRest,WF).



 :- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(balance,int(0)),field(name,fd(2,'Name'))]),
              rec([field(balance,global_set('NAT1')),field(name,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
 :- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(0))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
 :- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(1))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).

not_is_a_record_wf(rec(Fields),rec(FieldSets),WF) :-
   not_check_field_values(Fields,FieldSets,WF).
   %print_message(not_is_a_record(Record,FieldSets,WF)).


:- block not_check_field_values(-,-,?).
not_check_field_values([field(Name,V)|ValueRest],[field(Name,VT)|TypeRest],WF) :-
   membership_test_wf(VT,V,MemRes,WF),
   not_check_field_values2(MemRes,ValueRest,TypeRest,WF).

:- block not_check_field_values2(-,?,?,?).
not_check_field_values2(pred_false,_,_,_).
not_check_field_values2(pred_true,ValueRest,TypeRest,WF) :-
      not_check_field_values(ValueRest,TypeRest,WF).





:- assert_must_fail(( X = [field(b,3),field(a,2)], kernel_records:construct_record(X,X) )).
:- assert_must_succeed(( kernel_records:construct_record(X,rec([field(a,2),field(b,3)])),
    X = [field(b,3),field(a,2)] )).
:- assert_must_succeed(( kernel_records:construct_record(X,rec([field(a,2),field(b,3)])),
    X = [field(a,2),field(b,3)] )).
:- assert_must_succeed(( kernel_records:construct_record(X,Y),
    X = [field(b,4),field(a,2)], Y = rec([field(a,2),field(b,4)]) )).

:- block construct_record(-,?).
construct_record(Fields,rec(SortedFields)) :- % print_bt_message(sort_fields(Fields,SortedFields)),trace,
  sort_fields(Fields,SortedFields).
   % , print(construct_record(Fields,rec(SortedFields))),nl

sort_fields(Fields,SortedFields) :-
  known_field_names(Fields,Known), %print(sort(Known,Fields)),nl,
  block_sort(Known,Fields,SortedFields).

 :- block block_sort(-,?,?).
block_sort(K,F,SF) :- (K=sorted -> F=SF ; sort(F,SF)).

% instantiate output argument when all field names known + indicate whether sorted or not
:- block known_field_names(-,?).
known_field_names([],sorted).
known_field_names([field(Name,_)|T],Known) :- known_field_names4(T,Name,sorted,Known).

:- block known_field_names4(-,?,?,?),known_field_names4(?,-,?,?).
known_field_names4([],_,K,K).
known_field_names4([field(Name,_)|T],PrevName,SortedSoFar,Known) :-
    (PrevName @< Name -> S2=SortedSoFar ; S2=not_sorted),
    known_field_names4(T,Name,S2,Known).



% check if a sorted records has multiple field names
record_has_multiple_field_names([field(N,_)|T],Res) :- multiple_fields_aux(T,N,Res).
multiple_fields_aux([field(N,_)|T],N1,Res) :-
  (N1=N -> Res=N ; multiple_fields_aux(T,N,Res)).

:- use_module(library(lists)).

%%:- assert_must_fail(( kernel_records:access_record(rec([]),a,_) )). %% actually: must generate error; not just fail
:- assert_must_fail(( kernel_records:access_record(rec([field(a,int(2))]),a,int(3)) )).
:- assert_must_succeed(( kernel_records:access_record(rec(X),a,int(2)), X=[field(b,_),field(a,int(2))|_] )).
:- assert_must_succeed(( kernel_records:access_record(rec(X),b,int(2)), X=[field(b,_),field(a,int(3))|_] )).

%access_record(X,Y,Z) :- print_message(access_record(X,Y,Z)),fail.
:- block access_record(?,-,?). % when we add access_record(-,?,?) we do not instantiate to rec(_), but PROB-356 runs slower.
access_record(rec(Fields),FieldName,Value) :-
%   access_record2(Fields,FieldName,Value).
%:- block access_record2(?,-,?).
%access_record2(Fields,FieldName,Value) :-
   get_field(Fields,FieldName,Value,Fields).

:- block get_field(-,?,?,?).
get_field([],FieldName,_Value,OrigFields) :-
   add_internal_error('Could not get field: ', FieldName:OrigFields),fail.
get_field([field(SFieldName,SValue)|T], FieldName,Value,OrigFields) :-
   get_field2(SFieldName,SValue,T, FieldName,Value,OrigFields).

:- block get_field2(-,?,?, ?,?,?).
get_field2(SFieldName,SValue,T, FieldName,Value,OrigFields) :-
     (SFieldName=FieldName
       -> kernel_objects:equal_object(Value,SValue,get_field2)
       ;  get_field(T,FieldName,Value,OrigFields)
     ).



:- assert_must_succeed(( kernel_records:check_field_name_compatibility(a,a))).
check_field_name_compatibility(Name1,Name2) :-
   nonvar(Name1), nonvar(Name2), Name1 \= Name2, !,
   add_internal_error('Incompatible fields: ',check_field_name_compatibility(Name1,Name2)),
   fail.
check_field_name_compatibility(Name,Name).
