:- module(module_information,[module_info/2,module_info/3,get_module_info/3,number_of_groups/1,number_of_modules/1,number_of_modules/2]).

% a simple mechanism to provide annotations for various modules
% examples:
% :- module_info(description,'....').
% :- module_info(group,kernel).
:- meta_predicate module_info(0,*).
:- dynamic get_module_info/3.
module_info(Module:Attribute,Value) :-
   !,%print(describe_module(Module,Attribute,Value)),nl,
   assert(get_module_info(Module,Attribute,Value)).
module_info(Attribute,Value) :- nl,print('### Attribute has no module prefix: '), Attribute,nl,
   print('### Value: '), print(Value),nl.
% called for non-module prolog files
module_info(Module,Attribute,Value) :-
   assert(get_module_info(Module,Attribute,Value)).

:- module_info(group,infrastructure).
:- module_info(description,'This module contains functionality to store basic documentation and information about all modules in ProB.').

:- use_module(library(lists)).

% module_information:get_maximal_module_info(revision,X).
%  module_information:get_maximal_module_info(lastchanged,X).

number_of_groups(X) :-
    findall(Group,get_module_info(_Module,group,Group),List),
    remove_dups(List,NList),
    length(NList,X).

number_of_modules(X) :-
    findall(Module,get_module_info(Module,_A,_B),List),
    remove_dups(List,NList),
    length(NList,X).

number_of_modules(Group,X) :-
    findall(Module,get_module_info(Module,group,Group),List),
    remove_dups(List,NList),
    length(NList,X).


