:- module(synthesis_util,[get_distinguishing_state_or_trans/8,
                          get_valid_transitions_for_operation/6,
                          get_all_location_vars_for_operators/3,
                          get_involved_variables/4,
                          get_guard_list_from_operation/2,
                          get_solution_or_none/2,
                          get_single_constant_or_set_operator_id/3,

                          add_distinguishing_states_from_cache/6,
                          expand_transition_cache_if_solution/5,
                          expand_state_cache_if_guard_and_solution/6,
                          exclude_transition_constraint/4,
                          conjunct_previous_invariant_or_guard_if_necessary/4,
                          check_transitions_against_machine_operations/4,
                          check_transition_against_machine_operations/4,
                          pre_select_transitions/3,
                          delete_numbers_from_atom/2,
                          find_typed_identifier_uses_with_info/2,
                          get_binding_for_name_or_value/3,
                          sequence_well_definedness_constraint/2,
                          replace_preds_by_subst_if_neccessary/3,
                          parallel_to_conjunction/2,
                          adjust_distinguishing_state_binding/3,
                          contains_equivalent_state/3,

                          pretty_example_to_ast_nodes/3,
                          example_bindings_to_ast_nodes/2,
                          get_current_vars_from_example/2,
                          get_distinct_states/5,
                          mark_examples_as_initial/4,
                          get_valid_states/2,
                          reduce_input_to_violating_vars/3,
                          set_type_if_necessary/2,
                          get_varname_from_node_info/2,
                          get_varname_from_id_node/2,
                          untyped/1,
                          value_to_ast/2,

                          check_if_any_action_fits_example/4,
                          check_if_guard_fits_example/3,

                          create_equality_nodes_from_examples/5,
                          create_equality_nodes_from_examples/3,

                          get_violated_and_valid_machine_invariants/4,
                          get_invariant_violating_transition/6,
                          get_invariant_violating_example/4,

                          get_input_output_nodes_from_bindings/3,
                          get_input_nodes_from_bindings/2,
                          get_output_nodes_from_bindings/2,

                          get_varname_from_binding/2,
                          remove_pre_condition_from_body/2,
                          prime_or_unprime_variables_in_ast/3,
                          remove_prime_from_node_info_var/2,
                          remove_prime/2,
                          is_input_location_var/1,
                          is_seq/1,
                          is_ast/1,
                          zip/3]).

:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(b_interpreter),[b_test_boolean_expression_wf/3]).
:- use_module(probsrc(state_space)).
:- use_module(probsrc(solver_interface),[solve_predicate/3]).
:- use_module(probsrc(bmachine_eventb),[prime_variable/2,is_primed_id/1]).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(before_after_predicates),[before_after_predicate_with_equalities/3]).

:- use_module(extension('plspec/plspec/plspec')).
:- use_module(extension('plspec/plspec/plspec_checker')).

:- use_module(library(lists)).
:- use_module(library(avl),[avl_to_list/2]).
:- use_module(library(sets),[subtract/3]).

:- use_module(synthesis(logging)).

get_machinevars_by_name(VarNames,CurrentVars) :-
    b_get_machine_variables(MachineVars) ,
    get_machinevars_by_name_aux(VarNames,MachineVars,[],CurrentVars).
get_machinevars_by_name_aux([],_,Acc,Acc).
get_machinevars_by_name_aux([VarName|T],MachineVars,Acc,CurrentVars) :-
    member(b(identifier(VarName),Type,Info),MachineVars) ,
    get_machinevars_by_name_aux(T,MachineVars,[b(identifier(VarName),Type,Info)|Acc],CurrentVars).

% True if the given set of states contains an equivalent state.
contains_equivalent_state(_,[],_) :- fail.
contains_equivalent_state(State,[ExampleState|_],EquivalentState) :-
  create_equality_nodes_from_example(State,StateEqualityNodes) ,
  create_equality_nodes_from_example(ExampleState,ExampleStateEqualityNodes) ,
  append(StateEqualityNodes,ExampleStateEqualityNodes,EqualityNodes) ,
  conjunct_predicates(EqualityNodes,Equality) ,
  b_test_boolean_expression_wf(Equality,[],_) , ! ,
  EquivalentState = ExampleState.
contains_equivalent_state(State,[_|T],EquivalentState) :-
  contains_equivalent_state(State,T,EquivalentState).

% Get the internal names like add1 for the used components in a synthesized solution.
%get_used_operators_from_solution(b(Node,_,Info),Acc,UsedOperators) :-
    % if is output
%    member(synthesis(ComponentName,output),Info) , ! ,
%    accumulate_used_operator(ComponentName,Acc,NewAcc) ,
%    get_used_operators_from_solution_aux(Node,NewAcc,UsedOperators).
%get_used_operators_from_solution(b(Node,_,_),Acc,UsedOperators) :-
%    get_used_operators_from_solution_aux(Node,Acc,UsedOperators).

%get_used_operators_from_solution_aux(Node,Acc,UsedOperators) :-
%    Node =.. [_,LeftAst,RightAST] ,
%    get_used_operators_from_solution(LeftAst,Acc,Acc1) ,
%    get_used_operators_from_solution(RightAST,Acc1,UsedOperators).
%get_used_operators_from_solution_aux(Node,Acc,UsedOperators) :-
%    Node =.. [_,AST] , is_ast(AST) ,
%    get_used_operators_from_solution(AST,Acc,UsedOperators).
%get_used_operators_from_solution_aux(_,Acc,Acc).

%accumulate_used_operator(ComponentName,Acc,[(RawName,ComponentName)|Acc]) :-
%    delete_numbers_from_atom(ComponentName,RawName) ,
%    \+member((RawName,ComponentName),Acc).
%accumulate_used_operator(_,Acc,Acc).

is_input_location_var(b(identifier(_),_,Info)) :-
    member(synthesis(_,left_input),Info) ; member(synthesis(_,right_input),Info).

% Get all the location vars from a list of operators. The result is a map with entries like
% Operator:ListOfLocationVars. Just one iteration of the list of location variables for all operators.
get_all_location_vars_for_operators(Operators,LocationVars,OperatorLVarsMap) :-
    findall(Op:[],member(Op,Operators),EmptyAcc) ,
    get_all_location_vars_for_operators_aux(LocationVars,EmptyAcc,OperatorLVarsMap).
get_all_location_vars_for_operators_aux([],Acc,Acc).
get_all_location_vars_for_operators_aux([LocationVar|T],Acc,OperatorLVarsMap) :-
    LocationVar = b(_,_,Info) ,
    member(synthesis(ComponentName,LocationType),Info) ,
    member(LocationType,[left_input,right_input,output]) ,
    delete_numbers_from_atom(ComponentName,Operator) ,
    accumulate_locationvar_for_operator(LocationVar,Operator,Acc,NewAcc) ,
    get_all_location_vars_for_operators_aux(T,NewAcc,OperatorLVarsMap).
get_all_location_vars_for_operators_aux([_|T],Acc,OperatorLVarsMap) :-
    get_all_location_vars_for_operators_aux(T,Acc,OperatorLVarsMap).

% If the operator the location variable refers to is one that we search for, we accumulate the variable respectively.
accumulate_locationvar_for_operator(LocationVar,Operator,Acc,NewAcc) :-
    member(Operator:OperatorLocationVars,Acc) ,
    delete(Acc,Operator:_,TempAcc) ,
    NewAcc = [Operator:[LocationVar|OperatorLocationVars]|TempAcc].
% Otherwise skip the location variable.
accumulate_locationvar_for_operator(_,Operator,Acc,Acc) :-
    \+member(Operator:_,Acc).

% seq(_), string_set, integer_set, bool_set
get_single_constant_or_set_operator_id(seq-Type,Out,b(seq(Set),set(seq(NodeType)),[])) :-
    get_single_constant_or_set_operator_id(Type,_,Set) ,
    Set = b(_,TInnerType,_) ,
    TInnerType =.. [_|InnerType] ,
    NodeType =.. InnerType ,
    unique_typed_id("o",set(seq(NodeType)),TempOut) ,
    % we add the internal type of infinite set constants here
    % so we are able to call this predicate when converting location variables to ASTs later on
    add_texpr_infos(TempOut,[synthesis(type,seq-Type)],Out).
get_single_constant_or_set_operator_id(ConstantSet,Out,Constant) :-
    get_constant_node_or_id(ConstantSet,Constant) ,
    Constant = b(_,Type,_) ,
    unique_typed_id("o",Type,TempOut) ,
    add_texpr_infos(TempOut,[synthesis(type,ConstantSet)],Out).

get_constant_node_or_id(string_set,b(string_set,set(string),[])).
get_constant_node_or_id(bool_set,b(bool_set,set(boolean),[])).
get_constant_node_or_id(integer_set,b(integer_set('INTEGER'),set(integer),[])).
get_constant_node_or_id(integer_set_natural1,b(integer_set('NATURAL1'),set(integer),[])).
get_constant_node_or_id(integer_set_natural,b(integer_set('NATURAL'),set(integer),[])).
get_constant_node_or_id(constant-Type,Constant) :-
    unique_typed_id("constant",Type,Constant).

% Remove the prime from the identifier name in the term synthesis/2 in the node's info.
remove_prime_from_node_info_var(b(Node,Type,Info),b(Node,Type,[synthesis(machinevar,UnprimedVarName)|TInfo])) :-
    member(synthesis(machinevar,VarName),Info) ,
    remove_prime(VarName,UnprimedVarName) ,
    delete(Info,synthesis(machinevar,VarName),TInfo).

add_distinguishing_states_from_cache(action,_,PosInput,NegInput,PosInput,NegInput) :- !.
add_distinguishing_states_from_cache(_,(DistIn,DistOut),PosInput,NegInput,NewPosInput,NewNegInput) :-
    append(DistIn,PosInput,TempPosInput) ,
    append(DistOut,NegInput,TempNegInput) ,
    remove_dups(TempPosInput,NewPosInput) ,
    remove_dups(TempNegInput,NewNegInput).

% If the domain is enumerated from 1 to n we suppose set(couple(integer,_)) is a sequence.
is_seq(b(Node,Type,Info)) :-
    (Type = set(couple(integer,_))  ; Type = seq(_)) ,
    b_interpreter:b_test_boolean_expression_wf(b(equal(b(domain(b(Node,Type,Info)),set(integer),[]),
        b(interval(b(integer(1),integer,[]),
            b(size(b(Node,Type,Info)),integer,[])),set(integer),[])),pred,[]),[],[]).

%get_value_from_binding(bind(_,Value),Value).
%get_value_from_binding(binding(_,Value,_),Value).

get_distinguishing_state_or_trans(DistTransitionCache,CurrentVars,Type,ValidInvariantsOrGuards,AST1,
                                  AST2,NewDistTransitionCache,Distinguishing) :-
    b_get_invariant_from_machine(Invariant) ,
    % add WD-constraints for sequences
    find_typed_identifier_uses(AST1,AST1Nodes) , find_typed_identifier_uses(AST2,AST2Nodes) ,
    findall(Seq,(member(Seq,AST1Nodes) , is_seq(Seq)),Seq1) ,
    findall(Seq,(member(Seq,AST2Nodes) , is_seq(Seq)),Seq2) ,
    append(Seq1,Seq2,TSequences) , remove_dups(TSequences,Sequences) ,
    maplist(sequence_well_definedness_constraint,Sequences,SeqConstraintList) ,
    conjunct_predicates(SeqConstraintList,SeqConstraints) ,
    get_distinguishing_state_or_trans_aux(DistTransitionCache,Type,ValidInvariantsOrGuards,
                b(conjunct(SeqConstraints,b(conjunct(b(negation(
                    b(equivalence(AST1,AST2),pred,[])),pred,[]),Invariant),pred,[])),pred,[]),Distinguishing) ,
    expand_transition_cache_if_solution(Type,DistTransitionCache,CurrentVars,Distinguishing,NewDistTransitionCache) ,
    synthesis_log(distinguishing_transition_cache_changed(DistTransitionCache,NewDistTransitionCache)).
% When synthesizing an action only search for a distinguishing transition that doesn't violate the
% so far synthesized guard (in case there is one).
% Furthermore, we exclude already visited distinguishing transitions using the cache in order to
% prevent cycles when a distinguishing transition's output is changed by the user
% (otherwise we would find the same transition over and over again because the one with an adapted
% output is added to the examples).
get_distinguishing_state_or_trans_aux(DistTransitionCache,action,ValidInvariantsOrGuards,Constraint,Distinguishing) :-
    ValidInvariantsOrGuards \= any , ValidInvariantsOrGuards \= none , ! ,
    DistConstraint = b(conjunct(b(conjunct(Constraint,ValidInvariantsOrGuards),pred,[]),
                            DistTransitionCache),pred,[]) ,
    synthesis_log(distinguishing_constraint(action,DistConstraint)) ,
    solve_predicate(DistConstraint,_,TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).
get_distinguishing_state_or_trans_aux(DistTransitionCache,action,_,Constraint,Distinguishing) :- ! ,
    DistConstraint = b(conjunct(Constraint,DistTransitionCache),pred,[]) ,
    synthesis_log(distinguishing_constraint(action,Constraint)) ,
    solve_predicate(DistConstraint,_,TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).
% DistTransitionCache is empty when synthesizing guard/invariant
get_distinguishing_state_or_trans_aux(_,Type,_,Constraint,Distinguishing) :-
    synthesis_log(distinguishing_constraint(Type,Constraint)) ,
    solve_predicate(Constraint,_,TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).

get_solution_or_none(solution(Solution),solution(Solution)).
get_solution_or_none(Solution,none) :-
    Solution \= solution(_).

:- spec_pre(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),var]).
:- spec_invariant(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),ast]).
:- spec_post(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),var],
                                                   [atom,ast,list(ast),one_of([atom,solution_compound]),ast]).
expand_transition_cache_if_solution(action,DistTransitionCache,CurrentVars,solution(Solution),NewDistTransitionCache) :- ! ,
    get_input_output_nodes_from_bindings(Solution,Input,Output) ,
    exclude_transition_constraint(Input,Output,CurrentVars,Exclusion) ,
    NewDistTransitionCache = b(conjunct(DistTransitionCache,Exclusion),pred,[]).
expand_transition_cache_if_solution(action,DistTransitionCache,_,_,DistTransitionCache).
expand_transition_cache_if_solution(Type,DistTransitionCache,_,_,DistTransitionCache) :-
    Type \= action.

:- spec_pre(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                        ast_examples,ast_examples,var]).
:- spec_invariant(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                              ast_examples,ast_examples,pair(ast_examples,ast_examples)]).
:- spec_post(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                         ast_examples,ast_examples,var],
                                                        [atom,solution_compound,pair(ast_examples,ast_examples),
                                                         ast_examples,ast_examples,pair(ast_examples,ast_examples)]).
expand_state_cache_if_guard_and_solution(guard,solution(_),(UsedPositives,UsedNegatives),PosInput,
                                         NegInput,(PositiveStates,NegativeStates)) :-
    append(UsedPositives,PosInput,TPositiveStates) ,
    append(UsedNegatives,NegInput,TNegativeStates) ,
    remove_dups(TPositiveStates,PositiveStates) ,
    remove_dups(TNegativeStates,NegativeStates).
expand_state_cache_if_guard_and_solution(_,_,DistStateCache,_,_,DistStateCache).

:- spec_pre(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),var]).
:- spec_invariant(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),ast]).
:- spec_post(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),var],
                                             [list(ast),list(ast),list(ast),ast]).
exclude_transition_constraint(Input,Output,CurrentVars,b(negation(EqualityConjunction),pred,[])) :-
    b_get_machine_variables(MachineVars) ,
    maplist(get_varname_from_id_node,CurrentVars,CurrentIDs) ,
    create_equality_nodes_from_examples(CurrentIDs,Input,Output,MachineVars,EqualityNodes) ,
    conjunct_predicates(EqualityNodes,EqualityConjunction).

sequence_well_definedness_constraint(SeqIDNode,WDConstraint) :-
    % well-defined sequence: domain enumerated from 1..n
    WDConstraint = b(equal(b(domain(SeqIDNode),set(integer),[]),
        b(interval(b(integer(1),integer,[]),b(size(SeqIDNode),integer,[])),set(integer),[])),pred,[]).

is_ast(b(_,_,_)).

delete_numbers_from_atom(NumberAtom,Atom) :-
    atom_codes(NumberAtom,Codes) ,
    delete_numbers_from_codelist(Codes,NewCodes) ,
    atom_codes(Atom,NewCodes).
delete_numbers_from_codelist([],[]).
delete_numbers_from_codelist([Code|R],[Code|NR]) :-
    (Code > 57 ; Code < 48) , ! ,
    delete_numbers_from_codelist(R,NR).
delete_numbers_from_codelist([_|R],NR) :-
    delete_numbers_from_codelist(R,NR).

% Returns a list of all guards or 'any' if none is given.
:- spec_pre(get_guard_list_from_operation/2,[atom,var]).
:- spec_invariant(get_guard_list_from_operation/2,[atom,one_of([ast,atom])]).
:- spec_post(get_guard_list_from_operation/2,[atom,var],[atom,one_of([ast,atom])]).
get_guard_list_from_operation(Operation,GuardList) :-
    b_get_machine_operation(Operation,_,_,OpBody) ,
    get_guard_list_from_operation_aux(Operation,OpBody,Guard) ,
    conjunction_to_list(Guard,GuardList).
% No given guards.
get_guard_list_from_operation(_,any).

get_guard_list_from_operation_aux(_,b(precondition(Guard,_),_,_),Guard).
get_guard_list_from_operation_aux(Operation,b(rlevent(Operation,_,_,_,Guard,_,_,_,_,_,_),_,_),Guard).
get_guard_list_from_operation_aux(_,b(any(_,Guard,_),subst,_),Guard).

:- spec_pre(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,var]).
:- spec_invariant(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,ast]).
:- spec_post(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,var],
                                                                 [atom,one_of([atom,ast,list(ast)]),ast,ast]).
conjunct_previous_invariant_or_guard_if_necessary(invariant,ValidInvariants,ProgramAST,SolutionAST) :-
    ! , conjunct_predicates([ProgramAST|ValidInvariants],SolutionAST).
conjunct_previous_invariant_or_guard_if_necessary(guard,Guards,ProgramAST,SolutionAST) :-
    Guards \= any , ! ,
    % search for valid guards that are already given by the machine ( NewPreCond => OldPreCond )
    find_typed_identifier_uses(ProgramAST,[],ProgIDs) ,
    findall(Pre,(member(Pre,Guards) , find_typed_identifier_uses(Pre,PreIDs) ,
                 append(ProgIDs,PreIDs,TIDs) , remove_dups(TIDs,IDs) ,
                 b_test_boolean_expression_wf(
                    b(forall(IDs,ProgramAST,Pre),pred,[used_ids([])]),[],[])),InvalidPreConds) ,
    sets:subtract(Guards,InvalidPreConds,ValidPreConds) ,
    conjunct_predicates([ProgramAST|ValidPreConds],SolutionAST).
conjunct_previous_invariant_or_guard_if_necessary(action,Guard,ProgramAST,SolutionAST) :-
   \+member(Guard,[any,none]) ,  ! ,
   SolutionAST = b(precondition(Guard,ProgramAST),subst,[]).
conjunct_previous_invariant_or_guard_if_necessary(_,_,ProgramAST,ProgramAST).

% Replace equal by assign_single_id and conjunctions by parallel substitution.
:- spec_pre(replace_preds_by_subst_if_neccessary/3,[atom,ast,var]).
:- spec_invariant(replace_preds_by_subst_if_neccessary/3,[atom,ast,ast]).
:- spec_post(replace_preds_by_subst_if_neccessary/3,[atom,ast,var],[atom,ast,ast]).
replace_preds_by_subst_if_neccessary(action,TSolutionAST,SolutionAST) :-
    conjunction_to_list(TSolutionAST,EqualityList) ,
    maplist(equality_to_assignment,EqualityList,AssignmentList) ,
    SolutionAST = b(parallel(AssignmentList),subst,[]).
replace_preds_by_subst_if_neccessary(_,SolutionAST,SolutionAST).

equality_to_assignment(b(equal(Primed,Assignment),pred,Info),b(assign_single_id(Primed,Assignment),subst,Info)).

% Replace parallel substitution by conjunction predicates and assignments by equalities.
:- spec_pre(parallel_to_conjunction/2,[ast,var]).
:- spec_invariant(parallel_to_conjunction/2,[ast,ast]).
:- spec_post(parallel_to_conjunction/2,[ast,var],[ast,ast]).
parallel_to_conjunction(b(Node,Type,Info),b(Node,Type,Info)) :-
    Type \= subst.
parallel_to_conjunction(b(Node,subst,_),Conjunction) :-
    Node =.. [parallel,Assignments] ,
    findall(Equality,(member(Assignment,Assignments) , equality_to_assignment(Equality,Assignment)),EqualityNodes) ,
    conjunct_predicates(EqualityNodes,Conjunction).
%%

% Adjust the distinguishing state to the current variables (synthesis type is guard or invariant).
:- spec_pre(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,var]).
:- spec_invariant(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,solution_bindings]).
:- spec_post(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,var],
                                                   [list(ast),solution_bindings,solution_bindings]).
adjust_distinguishing_state_binding(CurrentVars,Distinguishing,Distinguishing) :-
    length(CurrentVars,VarAmount) ,
    length(Distinguishing,VarAmount).
% Shrink distinguishing state.
adjust_distinguishing_state_binding(CurrentVars,Distinguishing,ShrunkenInput) :-
    length(CurrentVars,VarAmount) ,
    length(Distinguishing,DistAmount) ,
    VarAmount < DistAmount ,
    findall(Binding,(member(Binding,Distinguishing) , Binding = binding(VarName,_,_) ,
                     member(b(identifier(VarName),_,_),CurrentVars)),ShrunkenInput).
% Expand distinguishing state.
adjust_distinguishing_state_binding(CurrentVars,Distinguishing,ShrunkenInput) :-
    length(CurrentVars,VarAmount) ,
    length(Distinguishing,DistAmount) ,
    VarAmount > DistAmount ,
    findall(Binding,(member(b(identifier(VarName),_,_),CurrentVars) ,
                     \+member(Binding,Distinguishing) , Binding = binding(VarName,_,_)),ShrunkenInput).

% Get the variables either involved in an invariant violation or those that should be considered
% when synthesizing a guard.
get_involved_variables(Guards,PosInput,NegInput,VarNames) :-
    get_involved_variables_aux(Guards,PosInput,NegInput,VarNames) ,
    synthesis_log(involved_variables(VarNames)).
get_involved_variables_aux(Guards,PosInput,NegInput,VarNames) :-
    Guards \= [] ,
    append(PosInput,NegInput,Input) ,
    check_states_on_guards(Input,Guards,ViolatedGuards) ,
    findall(UsedIDs,(member(Guard,ViolatedGuards) , find_identifier_uses(Guard,[],UsedIDs)),NestedList) ,
    append(NestedList,UsedIDs) ,
    UsedIDs \= [] ,
    remove_dups(UsedIDs,VarNames).
get_involved_variables_aux([],_,NegInput,VarNames) :-
    get_violated_and_valid_machine_invariants(NegInput,_,ViolatingVars,_) ,
    ViolatingVars \= [] , ! ,
    findall(Name,member(b(identifier(Name),_,_),ViolatingVars),VarNames).
get_involved_variables_aux(_,PosInput,_,VarNames) :-
    PosInput = [Example|_] ,
    maplist(get_varname_from_node_info,Example,VarNames).

check_states_on_guards(PosInput,Guards,ViolatedGuards) :-
    maplist(create_equality_nodes_from_example,PosInput,InputEqualities) ,
    findall(Guard,(member(Guard,Guards) , check_states_on_guard(InputEqualities,Guard)),ValidGuards) ,
    subtract(Guards,ValidGuards,ViolatedGuards).
check_states_on_guard([],_).
check_states_on_guard([InputEquality|T],Guard) :-
    conjunct_predicates([Guard|InputEquality],CheckGuard) ,
    b_test_boolean_expression_wf(CheckGuard,_,[]) , ! ,
    check_states_on_guard(T,Guard).

:- spec_pre(get_violated_and_valid_machine_invariants/4,[list(ast),var,var,var]).
:- spec_invariant(get_violated_and_valid_machine_invariants/4,[list(ast),list(ast),list(ast),list(ast)]).
:- spec_post(get_violated_and_valid_machine_invariants/4,[list(ast),var,var,var],
                                                         [list(ast),list(ast),list(ast),list(ast)]).
get_violated_and_valid_machine_invariants(Input,ViolatedInvariants,ViolatingVars,ValidInvariants) :-
    % according to specific operation
    b_get_machine_variables(TypedIDs) ,
    b_get_invariant_from_machine(MachineInvariant) ,
    conjunction_to_list(MachineInvariant,InvariantList) ,
%% TODO: maybe rewrite this more elegant
    maplist(get_violated_and_valid_machine_invariants_aux(TypedIDs,InvariantList),Input,TempViolatedInvariants) ,
    tools:flatten(TempViolatedInvariants,TempViolatedInvariants2) ,
    remove_dups(TempViolatedInvariants2,ViolatedInvariants) ,
%%
    conjunct_predicates(ViolatedInvariants,TempPred) ,
    find_typed_identifier_uses_with_info(TempPred,ViolatingVars) ,
    subtract(InvariantList,ViolatedInvariants,ValidInvariants).

get_violated_and_valid_machine_invariants_aux(TypedIDs,InvariantList,Input,ViolatedInvariants) :-
    findall(Equality,(member(IDNode,TypedIDs) , member(ValueNode,Input) ,
                      nth0(I,TypedIDs,IDNode) , nth0(I,Input,ValueNode) ,
                      Equality = b(equal(IDNode,ValueNode),pred,[])),EqualityList) ,
    conjunct_predicates(EqualityList,EqualityNodes) ,
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,InvariantList,ViolatedInvariants).

get_violated_and_valid_machine_invariants_aux2(_,[],[]).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[SingleInv|T],[SingleInv|NT]) :-
    conjunct_predicates([EqualityNodes,SingleInv],InvPred) ,
    find_identifier_uses(InvPred,[],UsedIDs) ,
    findall(Binding,(member(ID,UsedIDs) , Binding = bind(ID,_)),Bindings) ,
    \+ b_test_boolean_expression_wf(InvPred,Bindings,[]) , ! ,
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[_|T],NT) :-
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).

% TODO: support set-expressions
% Get another transition that violates the invariant, exclude given input/output example.
:- spec_pre(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),var]).
:- spec_invariant(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),pair(ast,ast)]).
:- spec_post(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),var],
                                                  [atom,list(ast),list(ast),list(ast),list(ast),pair(ast,ast)]).
get_invariant_violating_transition(Operation,UsedVars,ViolatedInvariants,Input,Output,Counter) :-
    b_get_machine_operation(Operation,_,_,TOpBody) ,
    % don't create contradiction in case the guard is equivalent
    % to an invariant which we will negate in the following
    remove_pre_condition_from_body(TOpBody,OpBody) ,
    before_after_predicate_with_equalities(OpBody,UsedVars,BABody) ,
    %find_identifier_uses(OpBody,[],BodyIDs) ,
    findall(V,(member(b(identifier(V),_,_),UsedVars)),BodyIDs) ,
    b_get_machine_variables(MachineVars) ,
    create_equality_nodes_from_examples(BodyIDs,Input,Output,MachineVars,TNodeList) ,
    % negate equalities to exclude this transition
    findall(Inequality,(create_texpr(negation(Equality),pred,[],Inequality) ,
        member(Equality,TNodeList)),NodeList) ,
    conjunct_predicates(ViolatedInvariants,Invariant) ,
    prime_or_unprime_variables_in_ast(prime,Invariant,PrimedInvariant) ,
    create_texpr(negation(Invariant),pred,[],NInvariant) ,
    create_texpr(negation(PrimedInvariant),pred,[],NPrimedInvariant) ,
    conjunct_predicates([BABody,NInvariant,NPrimedInvariant|NodeList],BAExclusion) ,
    solve_predicate(BAExclusion,_,Solution) ,
    get_invariant_violating_transition_aux(Solution,Counter).

get_invariant_violating_transition_aux(Solution,none) :-
    Solution \= solution(_).
get_invariant_violating_transition_aux(solution(Solution),(CounterInput,CounterOutput)) :-
    % split by input and output state
    get_input_output_nodes_from_bindings(Solution,CounterInput,CounterOutput).

get_invariant_violating_example(_,[],[],_) :- fail.
get_invariant_violating_example(ViolatedInvariants,[PosInput|TIn],[PosOutput|TOut],Example) :-
  maplist(check_invariant_on_state(PosInput,PosOutput),ViolatedInvariants) , ! ,
  % transition doesn't violate an invariant
  get_invariant_violating_example(ViolatedInvariants,TIn,TOut,Example).
% invariant violating transition found
get_invariant_violating_example(_,[PosInput|_],[PosOutput|_],(PosInput,PosOutput)).

% Is true if the given invariant is true for the pair of input/output states.
check_invariant_on_state(Input,Output,InvariantAst) :-
  create_equality_nodes_from_examples(Input,Output,EqualityNodes) ,
  conjunct_predicates([InvariantAst|EqualityNodes],InvariantTransitionPred) ,
  b_test_boolean_expression_wf(InvariantTransitionPred,[],_).

%% Get the input and/or output nodes from a solution of the distinguishing constraint.
:- spec_pre(get_input_output_nodes_from_bindings/3,[solution_bindings,var,var]).
:- spec_invariant(get_input_output_nodes_from_bindings/3,[solution_bindings,list(ast),list(ast)]).
:- spec_post(get_input_output_nodes_from_bindings/3,[solution_bindings,var,var],[solution_bindings,list(ast),list(ast)]).
get_input_output_nodes_from_bindings(Solution,Input,Output) :-
    get_input_nodes_from_bindings(Solution,Input) ,
    get_output_nodes_from_bindings(Solution,Output).

:- spec_pre(get_input_nodes_from_bindings/2,[solution_bindings,var]).
:- spec_invariant(get_input_nodes_from_bindings/2,[solution_bindings,list(ast)]).
:- spec_post(get_input_nodes_from_bindings/2,[solution_bindings,var],[solution_bindings,list(ast)]).
get_input_nodes_from_bindings(Solution,Input) :-
    findall(Binding,(member(Binding,Solution) , get_varname_from_binding(Binding,VarName) ,
        \+is_primed_id(VarName)),Bindings) ,
    example_bindings_to_ast_nodes(Bindings,Input).

:- spec_pre(get_output_nodes_from_bindings/2,[solution_bindings,var]).
:- spec_invariant(get_output_nodes_from_bindings/2,[solution_bindings,list(ast)]).
:- spec_post(get_output_nodes_from_bindings/2,[solution_bindings,var],[solution_bindings,list(ast)]).
get_output_nodes_from_bindings(Solution,Output) :-
    get_output_nodes_from_bindings_aux(Solution,NewSolution) ,
    example_bindings_to_ast_nodes(NewSolution,Output).
get_output_nodes_from_bindings_aux([],[]).
get_output_nodes_from_bindings_aux([Binding|T],[Binding|NT]) :-
    get_varname_from_binding(Binding,VarName) ,
    is_primed_id(VarName) , ! ,
    get_output_nodes_from_bindings_aux(T,NT).
get_output_nodes_from_bindings_aux([_|T],NT) :-
    get_output_nodes_from_bindings_aux(T,NT).

get_varname_from_binding(binding(VarName,_,_),VarName).
get_varname_from_binding(bind(VarName,_),VarName).

% rlevent/11
remove_pre_condition_from_body(b(Node,_,_),Body) :-
    Node =.. [_,_,_,_,_,_,_,Actions|_] , ! ,
    create_texpr(parallel(Actions),subst,[],OpBody) ,
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(b(any(_,_,TActions),subst,_),Body) :-
    (is_list(TActions) -> Actions = TActions ; Actions = [TActions]) , ! ,
    create_texpr(parallel(Actions),subst,[],OpBody) ,
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(PreBody,Body) :-
    PreBody = b(precondition(_,Body),_,_) , !.
remove_pre_condition_from_body(Body,Body).

% prime/unprime all identifiers within an ast
:- spec_pre(prime_or_unprime_variables_in_ast/3,[atom,ast,var]).
:- spec_invariant(prime_or_unprime_variables_in_ast/3,[atom,ast,ast]).
:- spec_post(prime_or_unprime_variables_in_ast/3,[atom,ast,var],[atom,ast,ast]).
prime_or_unprime_variables_in_ast(prime,b(identifier(ID),Type,Info),PrimedID) :- ! ,
    prime_variable(b(identifier(ID),Type,Info),PrimedID).
prime_or_unprime_variables_in_ast(unprime,b(identifier(ID),Type,Info),UnprimedID) :- ! ,
    (is_primed_id(ID)
    ->  remove_prime(b(identifier(ID),Type,Info),UnprimedID)
    ;   UnprimedID = b(identifier(ID),Type,Info)).
prime_or_unprime_variables_in_ast(_,b(Node,Type,Info),b(Node,Type,Info)) :-
    (Node = integer(_) ; Node = value(_) ; Node = integer_set(_) ; Node = string(_) ; Node =.. [_]).
prime_or_unprime_variables_in_ast(PrimeType,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [parallel,Assignments] ,
    maplist(prime_or_unprime_variables_in_ast(PrimeType),Assignments,NewAssignments) ,
    NewNode =.. [parallel,NewAssignments].
prime_or_unprime_variables_in_ast(PrimeType,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Op,In] ,
    prime_or_unprime_variables_in_ast(PrimeType,In,PrimedIn) ,
    NewNode =.. [Op,PrimedIn].
prime_or_unprime_variables_in_ast(PrimeType,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Op,In1,In2] ,
    prime_or_unprime_variables_in_ast(PrimeType,In1,PrimedIn1) ,
    prime_or_unprime_variables_in_ast(PrimeType,In2,PrimedIn2) ,
    NewNode =.. [Op,PrimedIn1,PrimedIn2].

%% User adds a missing state transition.
% Search for a machine operation that completely supports the missing state
% or one where we just have to synthesize a relaxed guard.
% VarNames refers to the variables used in the added states.
check_transitions_against_machine_operations(VarNodes,InputExamples,OutputExamples,Result) :-
    % check each example againt the existing machine operations
    maplist(check_transition_against_machine_operations(VarNodes),InputExamples,OutputExamples,SingleResults) ,
    check_transitions_against_machine_operations_aux(SingleResults,Result).
check_transitions_against_machine_operations_aux(SingleResults,Result) :-
    % an operation satisifies the transition but its guard needs to be relaxed
    member((OpName,guard),SingleResults) ,
    % and all other example transitions are also satisfied by this operation,
    maplist(operation_match(OpName),SingleResults) ,
    % therefore we synthesize a relaxed guard
    Result = (OpName,guard).
check_transitions_against_machine_operations_aux(SingleResults,Result) :-
    % an operation satisifies all transitions, and thus, we don't have to synthesize anything
    member((OpName,done),SingleResults) ,
    maplist('='((OpName,done)),SingleResults) ,
    Result = (OpName,done).
check_transitions_against_machine_operations_aux(_,no_match).

operation_match(OpName,(OpName,_)).

:- spec_pre(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),var]).
:- spec_invariant(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),
                                                                 one_of([atom,pair(atom,atom)])]).
:- spec_post(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),var],
                                                           [list(ast),list(ast),list(ast),
                                                            one_of([atom,pair(atom,atom)])]).
% Check a single transition against the existing machine operations.
check_transition_against_machine_operations(VarNodes,Input,Output,Result) :-
    findall((Op,Res,Par,Body),b_get_machine_operation(Op,Res,Par,Body),Operations) ,
    check_transition_against_machine_operations(Operations,Input,Output,VarNodes,Result).
check_transition_against_machine_operations([],_,_,_,no_match).
check_transition_against_machine_operations([(_,_,_,b(skip,subst,_))|T],Input,Output,VarNodes,Result) :-
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
% Skip if the desired action's and the given ones' variables doesn't match.
check_transition_against_machine_operations([(_,_,_,OpBody)|T],Input,Output,VarNodes,Result) :-
    find_identifier_uses(OpBody,[],BodyIDs) ,
    findall(Name,member(b(identifier(Name),_,_),VarNodes),VarNames) ,
    % find distinguishing variables
    findall(DistName,(member(DistName,BodyIDs) , \+member(DistName,VarNames)),DistNames) ,
    DistNames \= [] , ! ,
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
check_transition_against_machine_operations([(OpName,_,_,OpBody)|T],Input,Output,VarNodes,Result) :-
    b_get_machine_variables(MachineVars) ,
    before_after_predicate_with_equalities(OpBody,MachineVars,BABody) ,
    find_identifier_uses(OpBody,[],BodyIDs) ,
    create_equality_nodes_from_examples(BodyIDs,Input,Output,VarNodes,NodeList) ,
    conjunct_predicates([BABody|NodeList],BABodyPred) ,
    % double length because of primed variables
    length(VarNodes,Length) , NLength is Length * 2 , length(BindList,NLength) ,
    findall(E,(member(E,BindList) , E = bind(_,_)),BindList) ,
    check_transition_against_machine_operations_aux(BABodyPred,BindList,NodeList,[(OpName,_,_,OpBody)|T],
                                                    Input,Output,VarNodes,Result).

% Found operation that already executes the transition.
check_transition_against_machine_operations_aux(BABodyPred,BindList,_,[(OpName,_,_,_)|_],_,_,_,(OpName,done)) :-
    b_test_boolean_expression_wf(BABodyPred,BindList,[]) , !.
% Only the guard of an operation does not fit.
check_transition_against_machine_operations_aux(_,BindList,NodeList,[(OpName,_,_,OpBody)|_],_,_,_,(OpName,guard)) :-
    b_get_machine_variables(MachineVars) ,
    % remove guard and try if action itself supports the missing state
    OpBody = b(precondition(_PreCondition,Action),_,_) ,
    before_after_predicate_with_equalities(Action,MachineVars,BAAction) ,
    conjunct_predicates([BAAction|NodeList],BAActionPred) ,
    b_test_boolean_expression_wf(BAActionPred,BindList,[]) , !.
% The action also does not match so that we search for another operation.
check_transition_against_machine_operations_aux(_,_,_,[_|T],Input,Output,VarNodes,Result) :-
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
%%

get_valid_transitions_for_operation(Amount,OperationName,BeforeAcc,AfterAcc,BeforeStates,AfterStates) :-
    before_after_predicates:before_after_predicate_for_operation(OperationName,BeforeAfterPredicate) ,
    b_get_invariant_from_machine(Invariant) ,
    prime_or_unprime_variables_in_ast(prime,Invariant,PrimedInvariant) ,
    create_texpr(conjunct(BeforeAfterPredicate,PrimedInvariant),pred,[],BeforeAfterPredicateWithInvariant) ,
    b_get_machine_variables(MachineVars) ,
    get_valid_transitions_for_operation_aux(Amount,MachineVars,BeforeAfterPredicateWithInvariant,BeforeAcc,
                                            AfterAcc,BeforeStates,AfterStates).

get_valid_transitions_for_operation_aux(0,_,_,BeforeAcc,AfterAcc,BeforeAcc,AfterAcc).
get_valid_transitions_for_operation_aux(Amount,MachineVars,BeforeAfterPredicateWithInvariant,BeforeAcc,
                                        AfterAcc,BeforeStates,AfterStates) :-
    exclude_accumulated_before_after_states(MachineVars,BeforeAcc,[],TExclusion) ,
    (TExclusion = [] -> Exclusion = b(truth,pred,[]) ; Exclusion = TExclusion) ,
    create_texpr(conjunct(Exclusion,BeforeAfterPredicateWithInvariant),pred,[],Predicate) ,
    solve_predicate(Predicate,_,Solution) ,
    Solution = solution(SolutionBindings) ,
    findall(Primed,(member(Primed,SolutionBindings) , Primed = binding(Name,_,_) ,
                    bmachine_eventb:is_primed_id(Name)),AfterBindings) ,
    sets:subtract(SolutionBindings,AfterBindings,BeforeBindings) ,
    example_bindings_to_ast_nodes(BeforeBindings,Before) ,
    example_bindings_to_ast_nodes(AfterBindings,After) ,
    NewAmount is Amount - 1 ,
    get_valid_transitions_for_operation_aux(NewAmount,MachineVars,BeforeAfterPredicateWithInvariant,
                                            [Before|BeforeAcc],[After|AfterAcc],BeforeStates,AfterStates).

exclude_accumulated_before_after_states(_,[],Acc,Exclusion) :-
    conjunct_predicates(Acc,Exclusion).
exclude_accumulated_before_after_states(MachineVars,[BeforeState|T],Acc,Exclusion) :-
    maplist(exclude_variable_state(MachineVars),BeforeState,ExclusionList) ,
    conjunct_predicates(ExclusionList,Inequalities) ,
    exclude_accumulated_before_after_states(MachineVars,T,[Inequalities|Acc],Exclusion).

exclude_variable_state(MachineVars,b(Node,Type,Info),Inequality) :-
    member(synthesis(machinevar,VarName),Info) ,
    member(b(identifier(VarName),MachineVarType,MachineVarInfo),MachineVars) ,
    create_texpr(not_equal(b(Node,Type,Info),b(identifier(VarName),MachineVarType,MachineVarInfo)),pred,[],Inequality).

% it is probably faster to use:  length(L,N),random_subseq(List, L, _Rest).
%get_n_random_elements(_,[],[]).
%get_n_random_elements(1,[H|_],[H]).
%get_n_random_elements(C,List,[E|Elements]) :-
%    C1 is C - 1 ,
%    random:random_select(E,List,Rest) ,
%    get_n_random_elements(C1,Rest,Elements).

 % Split examples from model checking (check if invariant violated or valid) to
 % give assumptions to the user.
:- spec_pre(pre_select_transitions/3,[atom,var,var]).
:- spec_invariant(pre_select_transitions/3,[atom,list(ast),list(ast)]).
:- spec_post(pre_select_transitions/3,[atom,var,var],[atom,list(ast),list(ast)]).
pre_select_transitions(Operation,ExamplesAssumingYes,ExamplesAssumingNo) :-
     findall((From,To),
         (transition(From,Operation,_,To) , \+ invariant_not_yet_checked(To)),Transitions) ,
     findall((SF,ST),
         (member((F,T),Transitions) , visited_expression(F,SF) , visited_expression(T,ST) ,
             invariant_violated(T)),ExamplesAssumingNo) ,
     findall((SF,ST),
         (member((F,T),Transitions) , visited_expression(F,SF) , visited_expression(T,ST) ,
             \+invariant_violated(T)),TExamplesAssumingYes), ! ,
     pre_select_transitions_aux(TExamplesAssumingYes,ExamplesAssumingYes).

 % use at most 5 positive examples
pre_select_transitions_aux(TExamplesAssumingYes,ExamplesAssumingYes) :-
     length(TExamplesAssumingYes,L) ,
     L > 5 , ! ,
     sublist(TExamplesAssumingYes,ExamplesAssumingYes,_,5,_).
pre_select_transitions_aux(TExamplesAssumingYes,TExamplesAssumingYes).

% Can be used in both directions: Either VarName or Value has to be ground.
:- spec_pre(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
:- spec_invariant(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
:- spec_post(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])],
                                             [solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
get_binding_for_name_or_value(SolutionBindings,VarName,Value) :-
    member(binding(VarName,Value,_),SolutionBindings) ;
    member(bind(VarName,Value),SolutionBindings).


% search identifier nodes within an ast and keep each nodes information
find_typed_identifier_uses_with_info(AST,UsedIDs) :-
    find_typed_identifier_uses_with_info2(AST,TempUsedIDs) ,
    remove_dups(TempUsedIDs,UsedIDs).

find_typed_identifier_uses_with_info2(List,IDs) :-
    is_list(List) ,
    findall(ID,(member(ID,List) , ID = b(identifier(_),_,_)),IDs).
find_typed_identifier_uses_with_info2(b(identifier(IDName),Type,Info),[b(identifier(IDName),Type,Info)]).
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :-
    Node =.. [_,Arg] ,
    find_typed_identifier_uses_with_info2(Arg,UsedIDs).
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :-
    Node =.. [_,Arg1,Arg2] ,
    find_typed_identifier_uses_with_info2(Arg1,UsedIDs1) ,
    find_typed_identifier_uses_with_info2(Arg2,UsedIDs2) ,
    append(UsedIDs1,UsedIDs2,UsedIDs).
% rlevent
find_typed_identifier_uses_with_info2(b(Node,_,_),UsedIDs) :-
    Node =.. [_,_,_,_,_,Body|_] ,
    find_typed_identifier_uses_with_info2(Body,UsedIDs).
find_typed_identifier_uses_with_info2(_,[]).

check_if_any_action_fits_example([],_,_,none) :- !.
check_if_any_action_fits_example(Actions,Input,Output,MatchingAction) :-
    % TODO: maybe better to set the bindings of b_test_boolean_expression_wf/3 ?
    create_equality_nodes_from_examples(Input,Output,EqualityNodes) ,
    check_if_any_action_fits_example_aux(Actions,Input,Output,EqualityNodes,MatchingAction) , !.
check_if_any_action_fits_example_aux([(Action,_)|T],Input,Output,EqualityNodes,MatchingAction) :-
    remove_pre_condition_from_body(Action,Body) ,
    parallel_to_conjunction(Body,LogicalBody) ,
    conjunct_predicates([LogicalBody|EqualityNodes],Conjunction) ,
    (b_test_boolean_expression_wf(Conjunction,[],_)
    -> MatchingAction = Action
    ;  check_if_any_action_fits_example(T,Input,Output,MatchingAction)).

check_if_guard_fits_example(Action,_,yes) :-
    % no given guard
    remove_pre_condition_from_body(Action,Body) ,
    Action = Body , !.
check_if_guard_fits_example(b(precondition(Guard,_),subst,_),Input,Out) :-
    create_equality_nodes_from_examples(Input,Input,EqualityNodes) ,
    conjunct_predicates([Guard|EqualityNodes],Conjunction) ,
    (b_test_boolean_expression_wf(Conjunction,[],_) -> Out = yes ; Out = no) , !.

% Get current identifier names from an example before creating equality nodes to the example.
:- spec_pre(create_equality_nodes_from_examples/3,[ast_example,ast_example,var]).
:- spec_invariant(create_equality_nodes_from_examples/3,[ast_example,ast_example,ast_example]).
:- spec_post(create_equality_nodes_from_examples/3,[ast_example,ast_example,var],
                                                   [ast_example,ast_example,ast_example]).
create_equality_nodes_from_examples(Input,Output,EqualityNodes) :-
    b_get_machine_variables(MachineVars) ,
    maplist(get_varname_from_node_info,Input,CurrentIDs) ,
    create_equality_nodes_from_examples(CurrentIDs,Input,Output,MachineVars,EqualityNodes).
% We need to set variables to values from example by creating equality nodes
% that we can add to the before after predicate.
create_equality_nodes_from_examples([],_,_,_,[]).
create_equality_nodes_from_examples([VarName|T],Input,Output,VarNodes,[Node1,Node2|NodeList]) :-
    member(VarNode,VarNodes) ,
    VarNode = b(identifier(VarName),_,_) ,
    member(InputNode,Input) , get_varname_from_node_info(InputNode,VarName) ,
    member(OutputNode,Output) , get_varname_from_node_info(OutputNode,VarName) ,
    Node1 = b(equal(VarNode,InputNode),pred,[]) ,
    prime_variable(VarNode,PrimedVarNode) ,
    Node2 = b(equal(PrimedVarNode,OutputNode),pred,[]) ,
    create_equality_nodes_from_examples(T,Input,Output,VarNodes,NodeList).

create_equality_nodes_from_example(ExampleState,EqualityNodes) :-
  create_equality_nodes_from_example(ExampleState,[],EqualityNodes).
create_equality_nodes_from_example([],Acc,Acc).
create_equality_nodes_from_example([ValueAst|T],Acc,EqualityNodes) :-
    get_varname_from_node_info(ValueAst,VarName) ,
    get_machinevars_by_name([VarName],[MachineVar]) ,
    Equality = b(equal(MachineVar,ValueAst),pred,[]) ,
    create_equality_nodes_from_example(T,[Equality|Acc],EqualityNodes).

% Get valid states for synthesis of invariants.
:- spec_pre(get_valid_states/2,[int,var]).
:- spec_invariant(get_valid_states/2,[int,list(ast)]).
:- spec_post(get_valid_states/2,[int,var],[int,list(ast)]).
get_valid_states(C,ValidStates) :-
    b_get_invariant_from_machine(Invariant) ,
    solve_predicate(Invariant,_,solution(Solution)) ,
    get_valid_states(C,Solution,Invariant,ValidStates).

:- spec_pre(get_valid_states/4,[int,solution_bindings,ast,var]).
:- spec_invariant(get_valid_states/4,[int,solution_bindings,ast,list(ast)]).
:- spec_post(get_valid_states/4,[int,solution_bindings,ast,var],[int,solution_bindings,ast,list(ast)]).
get_valid_states(1,Solution,_,[ValidState]) :-
    example_bindings_to_ast_nodes(Solution,ValidState).
get_valid_states(C,Solution,Invariant,[ValidState|NT]) :-
    C1 is C - 1 ,
    b_get_machine_variables(MachineVars) ,
    findall(Exclusion,(member(Var,MachineVars) , Var = b(identifier(Name),_,_) ,
                       member(binding(Name,Value,_),Solution) ,
                       value_to_ast(Value,ValueNode) ,
                       Exclusion = b(not_equal(Var,ValueNode),pred,[])),ExclusionList) ,
    conjunct_predicates([Invariant|ExclusionList],NewPred) ,
    solve_predicate(NewPred,_,solution(NewSolution)) ,
    example_bindings_to_ast_nodes(Solution,ValidState) ,
    get_valid_states(C1,NewSolution,NewPred,NT).

%% Only for synthesis of guards or invariants:
% In case we synthesize guard/invariant it is possible that a state is graduated as positive and negative
% because we replace output states with TRUE/FALSE. Thus, the constraint would have a contradiction.
% E.g. synthesizing an invariant: If we graduate [0]-->[-1] as valid and [-1]-->[-2] as invalid,
% we don't want to assume [-1] to be a negative state from the second example.
% Synthesizing a guard it is the opposite, i.e., we don't want to assume [-1] to be a positive state.
get_distinct_states(invariant,Positive,TempNegative,Positive,Negative) :-
    get_distinct_states(Positive,TempNegative,Negative).
get_distinct_states(guard,TempPositive,Negative,Positive,Negative) :-
    get_distinct_states(Negative,TempPositive,Positive).

% Duplicates are subtracted from the examples given in the second argument.
:- spec_pre(get_distinct_states/3,[ast_examples,ast_examples,var]).
:- spec_invariant(get_distinct_states/3,[ast_examples,ast_examples,ast_examples]).
:- spec_post(get_distinct_states/3,[ast_examples,ast_examples,var],[ast_examples,ast_examples,ast_examples]).
get_distinct_states(Positive,TempNegative,Negative) :-
    findall(Neg,(member(Neg,TempNegative) , \+contains_equivalent_state(Neg,Positive,_)),Negative).

% Check if a set of examples contains a specific example. Remove the example ast node infos to be able to use member
% and return the asts without infos.
%contains_example(Example,Examples,ExampleWithoutInfo) :-
%  findall(NodeWithoutInfo,(member(b(Node,Type,_),Example) , NodeWithoutInfo = b(Node,Type,_)),ExampleWithoutInfo) ,
%  member(ExampleWithoutInfo,Examples).

% Mark initial examples to be able to restrict constant domains when defining the synthesis
% constraint. Otherwise we would expand the constant domains for each new example derived from
% distinguishing interactions.
mark_examples_as_initial(TempPositive,TempNegative,Positive,Negative) :-
    maplist(mark_examples_as_initial_aux,TempPositive,Positive) ,
    maplist(mark_examples_as_initial_aux,TempNegative,Negative).
mark_examples_as_initial_aux([],[]).
mark_examples_as_initial_aux([ValueNode|T],[b(Node,Type,NewInfo)|NT]) :-
    ValueNode = b(Node,Type,Info) ,
    mark_examples_as_initial_aux2(Info,NewInfo) ,
    mark_examples_as_initial_aux(T,NT).
mark_examples_as_initial_aux2(Info,Info) :-
    member(synthesis(initial_example),Info).
mark_examples_as_initial_aux2(Info,[synthesis(initial_example)|Info]) :-
    \+member(synthesis(initial_example),Info).

% InputList contains all machine variables, but for synthesizing an invariant we only need to
% observe variables from violated invariants.
% Furthermore we accept hints from the user on variables that are probably involved in an action's
% guard so we also need to reduce examples there.
reduce_input_to_violating_vars(ViolatingVars,InputExample,ReducedExample) :-
    % find values of matching variable names located in ast nodes info term synthesis/2
    findall(Value,(member(Value,InputExample) , get_varname_from_node_info(Value, VarName) ,
                   member(VarName,ViolatingVars)),ReducedExample).

% Get the current variables from an example (info contains synthesis/2 with var name).
get_current_vars_from_example(Example,CurrentVars) :-
    b_get_machine_variables(MachineVars) ,
    findall(VarName,(member(Node,Example) , get_varname_from_node_info(Node,VarName)),CurrentVarNames) ,
    findall(VarNode,(member(VarNode,MachineVars) , VarNode = b(identifier(VarName),_,_) ,
                     member(VarName,CurrentVarNames)),CurrentVars).

pretty_example_to_ast_nodes([],[],[]).
pretty_example_to_ast_nodes([Value|T],[VarName|VT],[Node|NT]) :-
    value_to_ast(Value,TempNode) ,
    add_texpr_infos(TempNode,[synthesis(machinevar,VarName)],TNode) ,
    set_type_if_necessary(TNode,Node) ,
    pretty_example_to_ast_nodes(T,VT,NT).

value_to_ast(b(Node,Type,Info),b(Node,Type,Info)) :- !.
value_to_ast(Value,AST) :-
    value_to_ast_aux(Value,AST).

value_to_ast_aux((Value1,Value2),b(couple(Node1,Node2),couple(Type1,Type2),[])) :-
    value_to_ast_aux(Value1,Node1) ,
    value_to_ast_aux(Value2,Node2) ,
    Node1 = b(_,Type1,_) , Node2 = b(_,Type2,_).
value_to_ast_aux(Value,b(integer(Value),integer,[])) :-
    integer(Value).
value_to_ast_aux(int(Value),b(integer(Value),integer,[])).
value_to_ast_aux(Bool,b(boolean_true,boolean,[])) :-
    Bool = 'TRUE' ; Bool = pred_true.
value_to_ast_aux(Bool,b(boolean_false,boolean,[])) :-
    Bool = 'FALSE' ; Bool = pred_false.
value_to_ast_aux(string(S),b(string(S),string,[])).
value_to_ast_aux(Atom,b(string(Atom),string,[])) :-
    Atom \= [] , atom(Atom).
value_to_ast_aux(ListOrSet,b(value([]),set(_),[])) :-
    (ListOrSet = [] ; ListOrSet = {}).
value_to_ast_aux(avl_set(Tree),b(value(avl_set(Tree)),set(Type),[])) :-
    Tree = node(_,_,_,_,_) ,
    avl_to_list(Tree,[Value-_|_]) ,
    value_to_ast(Value,b(_,Type,_)).
value_to_ast_aux(List,b(set_extension(NodeList),set(Type),[])) :-
    is_list(List) ,
    maplist(value_to_ast,List,NodeList) ,
    NodeList = [b(_,Type,_)|_].
% ...

get_varname_from_node_info(b(_,_,Info),VarName) :-
    member(synthesis(machinevar,VarName),Info).

get_varname_from_id_node(b(identifier(VarName),_,_),VarName).

% translate prob values to asts node and add the machine variable name to each node's info
example_bindings_to_ast_nodes([],[]).
example_bindings_to_ast_nodes([Binding|T],[Ast|NT]) :-
    (Binding = binding(TVarName,In,_) ; Binding = bind(TVarName,In)) ,
    value_to_ast(In,TAst) ,
    remove_prime(TVarName,VarName) ,
    add_texpr_infos(TAst,[synthesis(machinevar,VarName)],TTAst) ,
    set_type_if_necessary(TTAst,Ast) ,
    example_bindings_to_ast_nodes(T,NT).

%:- spec_pre(get_pretty_value_list/2,[solution_bindings,var]).
%:- spec_invariant(get_pretty_value_list/2,[solution_bindings,list(compound)]).
%:- spec_post(get_pretty_value_list/2,[solution_bindings,var],[solution_bindings,list(compound)]).
%get_pretty_value_list(Solution,PrettyValues) :-
%    findall(Value,member(bind(_,Value),Solution),Values) ,
%    maplist(translate_bvalue,Values,PrettyValues).

remove_prime(b(identifier(VarName),Type,Info),b(identifier(UnPrimed),Type,Info)) :-
    remove_prime(VarName,UnPrimed).
remove_prime(VarName,UnPrimed) :-
    atom_concat(UnPrimed,'\'',VarName).
remove_prime(VarName,VarName).

% we have to set the types of untyped ast nodes like empty sets
set_type_if_necessary(b(Node,Type,Info),b(Node,GroundType,Info)) :-
    untyped(Type) ,
    b_get_machine_variables(MachineVars) ,
    member(synthesis(machinevar,VarName),Info) ,
    member(b(identifier(VarName),GroundType,_),MachineVars).
set_type_if_necessary(b(Node,Type,Info),b(Node,Type,Info)).

untyped(Type) :-
    var(Type).
untyped(set(Type)) :-
    untyped(Type).
untyped(seq(Type)) :-
    untyped(Type).

% Zip two lists of the same size, i.e., create tuples of elements having the same index respectively.
zip(L1,L2,Zipped) :-
    zip_aux(L1,L2,[],Zipped).
zip_aux([],[],Acc,Acc).
zip_aux([Elm1|T1],[Elm2|T2],Acc,Zipped) :-
    zip_aux(T1,T2,[(Elm1,Elm2)|Acc],Zipped).
