% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(version, [ version/4, version_str/1,
                     revision/1,
                     lastchangeddate/1,
                     compare_current_version/2
                   ]).

:- use_module(module_information,[module_info/2]).

:- module_info(group,infrastructure).
:- module_info(description,'This module contains the current version number of ProB together with automatically updated version control infos.').


version(1,7,0,'final').

% example call: compare_current_version([1,7],Res).
compare_current_version(NrList,Result) :- version(A,B,C,_),
   compare_aux(NrList,[A,B,C],Result).

compare_aux([],_,current_matches).
compare_aux([Requested|RT],[Cur|CT],Res) :- Requested = Cur,!, compare_aux(RT,CT,Res).
compare_aux([Requested|_],[Cur|_],Res) :- Requested > Cur,!, Res = current_older.
compare_aux(_,_,current_newer).

:- use_module(library(lists),[append/2]).
version_str(Str) :- version(V1,V2,V3,Suffix),
    number_codes(V1,C1),
    number_codes(V2,C2),
    number_codes(V3,C3),
    atom_codes(Suffix,S),
    append([C1,".",C2,".",C3,"-",S],C),
    atom_codes(Str,C).

% this file needs to be touched and committed for information below to be updated!

get_revision(R) :- catch(read_revision(R),_,R='no revision found').
get_lastchanged(R) :- catch(read_lastchanged(R),_,R='no lastchanged found').


% Tries to read the revision file generated by make.
% If it does not come from svn, the fallback revision number is used.
read_revision(A) :-
    open('revision', read, ID),
    read_line(ID,R), close(ID),
    atom_codes(A,R).

read_lastchanged(A) :-
    open('lastchanged', read, ID),
    read_line(ID,R), close(ID),
    atom_codes(A,R).

:- dynamic  revision/1, lastchangeddate/1.
:- get_revision(X), assert(revision(X)).
:- get_lastchanged(X), assert(lastchangeddate(X)).
