% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

%:- prolog_flag(source_info,_,on). % comment in if you want source location, e.g., for exceptions,...

portray_message(informational, _).

% patch for SICStus 4.3.3 performance issue
% sprm_14972_patch.pl BEGIN
:- if((current_prolog_flag(dialect, sicstus),
      current_prolog_flag(version_data, sicstus(4,3,3,_,_)))).
prolog:wf_call_like_arg(A, B, C, D, E, F, G, H, I, _) :-
  prolog:wellformed_body_iso(A, B, C, D, E, F, G, H, I, quiet),
  !.
prolog:wf_call_like_arg(A, B, C, A, D, _, E, _, _, _) :-
  F=E:A,
  prolog:condense_layout(B, G),
  prolog:comp_layout2(B, G, B, H),
  C=call(F),
  prolog:comp_layout1(B, H, D).
:- endif.
% sprm_14972_patch.pl END

%% sp_4_1_3_nested_exception_patch.pl BEGIN
%% Compile this to patch a bug in 4.1.3 affecting nested exception handlers.
%% This source code is a no-op on other versions of sicstus than 4.1.3.
:- if(current_prolog_flag(dialect,sicstus)).
:- if(current_prolog_flag(version_data,sicstus(4,1,3,0,_))).

prolog:(matching_exception(internal, Ball, Templ) :- !,
        Templ = Ball).
prolog:(matching_exception(prolog, '$reserved'(_), _) :- !,
        fail).
prolog:(matching_exception(prolog, time_out, _) :- !,
        fail).
prolog:(matching_exception(_, Ball, Template) :-
        compound(Ball),
        sysball_iso_equiv(Ball,Iso1Exc), !,
        Template = error(Iso1Exc, Ball)).
prolog:(matching_exception(_, Ball, Ball)).

:- endif.                       % sicstus 4.1.3
:- endif.                       % sicstus
%% sp_4_1_3_nested_exception_patch.pl END

%:- include('self_check_off.pl').

% use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(prob_cli,group,infrastructure).
:- module_info(prob_cli,description,'ProB start file in cli mode.').

:- use_module(debugging_calls).
:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
:- disable_debugging_calls.

:- use_module(pathes,[set_search_pathes/0]). % called first to set_compile_time_search_pathes
:- use_module(tools,[string_concatenate/3,arg_is_number/2]).
:- use_module(tools_printing,[print_error/1]).


% :- use_module(extension('timer/assert_profiler.pl')).


:- compile(tcltk_interface).
%:- compile(gui_tcltk).

:- use_module(eclipse_interface).
:- use_module(prob2_interface,[start_animation/0, is_initialised_state/1,
                               set_eclipse_preference/2, update_preferences_from_spec/0,
                               load_cspm_spec_from_cspm_file/1]).


:- use_module(prob_socketserver).

:- use_module(coverage_statistics,[compute_the_coverage/5]).
:- use_module(value_persistance, [set_storage_directory/1,unset_storage_directory/0]).

do_one_gui_event.
tk_do_all_events.

:- dynamic junit_mode/1.

:- use_module(b_trace_checking,[check_default_trace_for_specfile/1, tcltk_check_state_sequence_from_file/1]).
%:- use_module(cspParserHs,'cia/haskell/java_api/cspParserHs.pl',
%               [initCspParser/0]).

:- use_module(library(file_systems),[file_exists/1]).
:- use_module(debug).
:- use_module(preferences).
:- use_module(probltlsrc(ltl),[ltl_check_assertions/2,ltl_model_check/4]).
:- use_module(probltlsrc(ctl),[ctl_model_check/4]).
:- use_module(test_typechecker,[run_typecheck_testcase/2]).

:- use_module(basic_unit_tests). % basic unit tests
:- use_module(self_check,[disable_interaction_on_errors/0]).
:- use_module(error_manager).

:- use_module(extension('zmq/master/master')).
:- use_module(extension('zmq/worker/worker')).
:- use_module(extension('ltsmin/ltsmin')).

:- use_module(self_check,[perform_self_check/1,turn_off_run_time_type_checks/0,turn_on_run_time_type_checks/0]).

:- use_module(library(system)).
:- use_module(library(codesio)).

:- use_module(logger).
%:- use_module(log_analyser, [analyse_log/3, check_log/0]).
:- use_module(specfile).
:- use_module(bmachine,[b_write_machine_representation_to_file/2,full_b_machine/1, b_write_eventb_machine_to_classicalb_to_file/1]).

:- use_module(xtl_interface,[set_cspm_main_process/1]).
:- use_module(meta_interface,[is_dot_command/1, call_dot_command/3,
                              is_dot_command_for_expr/1,call_dot_command_for_expr/4,
                              command_description/3]).
:- use_module('kodkod/kodkod_test').
:- use_module(predicate_analysis,[test_predicate_analysis/0]).
:- use_module(b_show_history,[write_history_to_file/2,write_values_to_file/1,write_all_values_to_dir/1,write_history_to_user_output/1]).
:- use_module(external_functions,[check_if_predicates_were_evaluated/1]).
:- use_module(sap, [write_all_deadlocking_paths_to_xml/1, test_generation_by_xml_description/1]).

:- use_module(smtlib_solver(smtlib2_cli)).
:- use_module(disprover_test_runner, [run_disprover_on_all_pos/1, load_po_file/1,print_disprover_stats/0, set_disprover_timeout/1, set_disprover_options/1, reset_disprover_timeout/0]).
:- use_module(latex_processor, [process_latex_file/2]).

:- meta_predicate if_option_set(-,0).
:- meta_predicate if_option_set(-,0,0).
:- meta_predicate if_options_set(-,0).
:- meta_predicate if_option_set_loaded(-,-,0).
:- meta_predicate ifm_option_set(-,0).
:- meta_predicate ifm_option_set_loaded(-,-,0).

start_probcli_timer(timer(T1,WT1)) :-
                   statistics(runtime,[T1,_]),
                   statistics(walltime,[WT1,_]).
stop_probcli_debug_timer(Timer,Msg) :- (debug_mode(on) -> stop_probcli_timer(Timer,Msg) ; true).
stop_probcli_timer(timer(T1,WT1),Msg) :- stop_probcli_timer(timer(T1,WT1),Msg,_).
stop_probcli_timer(timer(T1,WT1),Msg,WTotTime) :-
        statistics(runtime,[T2,_]), TotTime is T2-T1,
        statistics(walltime,[WT2,_]), WTotTime is WT2-WT1,
        format('~w ~w ms walltime (~w ms runtime), since start: ~w ms~n',[Msg,WTotTime,TotTime,WT2]),
        !.
stop_probcli_timer(Timer,Msg,_) :- add_internal_error('Illegal timer call: ',stop_probcli_timer(Timer,Msg)).
print_total_probcli_timer :-
        statistics(runtime,[T2,_]),
        statistics(walltime,[WT2,_]),
        format('Since start of probcli: ~w ms walltime (~w ms runtime)~n',[WT2,T2]).


:- use_module(tools_meta,[safe_time_out/3]).
:- meta_predicate timeout_call(0,-).
timeout_call(Call,NOW) :- option(timeout(TO)),!,
   statistics(runtime,[T1,_]),
   safe_time_out(Call,TO,Res),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls,
   (Res=time_out -> print('*** Timeout occurred: '), print(TO),nl,
                    print('*** Call: '), print(Call),nl,
                    nl,
                    writeln_log(timeout_occurred(NOW,Call))
                 ;  true).
timeout_call(Call,_NOW) :-
   statistics(runtime,[T1,_]),
   call(Call),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls.

:- use_module(junit_tests,[set_junit_dir/1, print_junit/2, create_junit_result/6]).
set_junit_mode(X) :- set_junit_dir(X), statistics(runtime,[Start,_]), assert(junit_mode(Start)).


go_proxy :-
    catch( go_cli(['-s','8888']), halt(ExitCode),
           ( nl,write('CLI halt prevented, exit code '),write(ExitCode),nl) ).

go_cli :-
    prolog_flag(argv,ArgV),
    go_cli_with_junit_check(ArgV).

initialise_cli :- counter_init,new_counter(cli_execute_inits),new_counter(cli_errors), new_counter(cli_warnings).
reset_cli :-
  reset_errors,
  reset_expected_error_occurred,
  reset_optional_errors,
  reset_counter(cli_errors), reset_counter(cli_warnings),
  preferences:reset_to_defaults,
  reset_logger,
  unset_storage_directory,
  % now done via event handling: clear_dynamic_predicates_for_POR,
  %     retractall(refinement_checker: generated_predeterministic_refinement(_,_,_)),
  %     retractall(refinement_checker: determinism_check_div_found(_)),
  %     reset_cspm_main_process,
  retractall(accumulated_infos(_,_,_)),
  (file_loaded(_)
   -> clear_loaded_machines,
      retractall(file_loaded(_,_)),
      retractall(loaded_main_file(_,_))
   ;  true).

go_cli_with_junit_check(ArgV) :-
    catch( go_cli(ArgV),
           halt(ExitCode),
           ( ExitCode = 0 ->
               true
           ; junit_mode(S) ->
               statistics(runtime,[E,_]), T is E - S,
	       create_junit_result(ArgV,'Integration Tests','Integration_Test',T,error([ExitCode]),Result),
               print_junit([Result],'Integration_Test')
           ; throw(halt(ExitCode)))).

go_cli(ArgV) :-
     (go_cli2(ArgV) -> true
     ; print_error('INTERNAL ERROR OCCURRED !'),nl,
       print_error(' (go_cli failed)'),nl, halt_exception(1)).

no_command_issued :- \+ command_option(_).
command_option(X) :- option(X), \+ not_command_option(X).
not_command_option(verbose).
not_command_option(very_verbose).
not_command_option(profiling_on).
not_command_option(set_pref(_,_)).
not_command_option(set_card(_,_)).
not_command_option(set_argv(_)).
not_command_option(silent).


%:- use_module(extension('myheap/myheap')).
go_cli2(ArgV) :-
    %print_total_probcli_timer,
    set_search_pathes,
    %myheap_init,
    announce_event(startup_prob),
    initialise_cli,
    init_eclipse_preferences, % startup_prob will already call init_preferences; TO DO: simplify
    external_functions:reset_argv,
    cli_init_options(ArgV,RemArgV),
    %% cli_init_options(['-ll'|ArgV],RemArgV),  %% comment in to have an automatically logging probcli
    %% cli_init_options(['-vv','-version'|ArgV],RemArgV),  %% comment in to have an automatically verbose probcli
    !,
    ((RemArgV=[File],no_command_issued,
      get_filename_extension(File,Ext), \+ do_not_execute_automatically(Ext))
       % then assume we want to do -execute_all:
      -> assert_option(execute(2147483647,false,current_state))
       ; true),
    if_option_set(test_mode,set_random_seed_to_deterministic_start_seed, set_new_random_seed),
    if_option_set(verbose,
                  verbose, %tcltk_turn_debugging_on(19),
                  tcltk_turn_debugging_off),
    if_option_set(very_verbose,
                  very_verbose), %tcltk_turn_debugging_on(5)),
    if_option_set(profiling_on,profiling_on),
    debug_print(9,'Command Line Arguments: '),debug_println(9,ArgV),
    if_option_set(very_verbose,
                  print_options),
    debug_print(6,'Command Line File Args: '),debug_println(6,RemArgV),
    if_option_set(cache_storage(StorageDir), set_storage_directory(StorageDir)),
    if_option_set(parsercp(ParserLoc),
                  assert(parsercall:parser_location(ParserLoc))),
    cli_show_help(ArgV,RemArgV),
    generate_time_stamp(Datime,NOW),
    if_option_set(log(F,Mode),
                  cli_start_logging(F,Mode,NOW,Datime,RemArgV)),
    if_option_set(runtimechecking,
                  turn_on_run_time_type_checks,
                  turn_off_run_time_type_checks),
    if_option_set(junit(X),
                  set_junit_mode(X)),
    set_prefs,
    set_optional_errors,
    if_option_set(set_argv(F),
                  set_argv(F)),
    if_option_set(selfcheck(_),
                  cli_start_selfcheck),
    if_option_set(typechecker_test(Filename),
                  (run_typecheck_testcase(Filename,typesok) -> halt_prob(NOW); halt1_prob(NOW))),
    if_option_set(print_version, print_version),
    if_option_set(print_short_version, print_short_version),
    if_option_set(check_java_version, check_java_version),
    if_option_set(very_verbose,
                  preferences:print_preferences),
    if_option_set(zmq_worker, zmq_start_worker(localhost,5000,0,NOW)),
    if_option_set(zmq_worker2(MasterIP, Port, Recover), zmq_start_worker(MasterIP,Port,Recover,NOW)),
    debug_println(6,loading(RemArgV)),
    cli_load_files(RemArgV,NOW), % all CLI arguments which are not understood are supposed to be files to be treated
    debug_println(6,finished_loading(RemArgV)),
    %if_options_set(analyse_log(LF,RowID,ForGnu),
    %              analyse_log(LF,RowID,ForGnu)),
    %if_options_set(check_log(LF),
    %              check_log(LF)),
    if_option_set(socket(Port),
                  cli_start_socketserver(Port)),
    % check_all_expected_errors_occurred(NOW),  % is now checked for each file; socket_server should not generate errors ?
    debug_println(20,'% ProB Finished'),
    cli_print_junit_results(ArgV),
    get_counter(cli_errors,CErrs), get_counter(cli_warnings,CWarns),
    writeln_log_time(prob_finished(NOW,CErrs,CWarns)),
    write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns]),
    ((CErrs>0 ; CWarns>0) -> format(user_error,'! Total Errors: ~w, Warnings:~w~n',[CErrs,CWarns]) ; true),
    stop_xml_group_in_log('probcli-run').

cli_init_options(ArgV,RemArgV) :- %print(argv(ArgV)),nl,
    append(ProBArgV,['--'|BArgV],ArgV),!, % pass arguments after -- to B via external_functions
    cli_init_options2(ProBArgV,RemArgV),
    debug_println(20,set_argv_from_list(BArgV)),
    external_functions:set_argv_from_list(BArgV).
cli_init_options(ArgV,RemArgV) :- cli_init_options2(ArgV,RemArgV).
cli_init_options2(ArgV,RemArgV) :-
    retractall(option(_)),
    %%assert(option(log('/tmp/ProBLog.log'))), print('LOGGING'),nl, %% coment in to build a version of probcli that automatically logs
    ( get_options(ArgV,recognised_cli_option,Options,RemArgV) ->
        assert_all_options(Options)
    ; otherwise ->
        print_error(get_options_failed(ArgV)),definite_error_occurred).
cli_show_help(ArgV,RemArgV) :-
    ( (option(help) ; ArgV=[]) ->
        print_help, (RemArgV=[] -> halt_exception ; true)
    ; true).
cli_start_logging(F,Mode,NOW,Datime,RemArgV) :-
    debug_print(20,'%logging to: '), debug_println(20,F),
    set_log_file(F), set_logging_mode(Mode),
    start_xml_group_in_log('probcli-run'),
    writeln_log(start_logging(NOW,F)),
    version(V1,V2,V3,Suffix), revision(Rev),  lastchangeddate(LCD),
    writeln_log(version(NOW,V1,V2,V3,Suffix,Rev,LCD)), % still used by log_analyser
    prolog_flag(version,PV),
    write_xml_element_to_log(version,[major/V1,minor/V2,patch/V3,suffix/Suffix,revision/Rev,lastchanged/LCD,prolog/PV]),
    findall(Opt, option(Opt), Options),
    write_prolog_term_as_xml_to_log(options(NOW,Options)),
    write_prolog_term_as_xml_to_log(files(NOW,RemArgV)),
    datime(Datime,DateRecord),
    writeln_log(date(NOW,DateRecord)),
    (DateRecord=datime(Yr,Mon,Day,Hr,Min,Sec)
     -> write_xml_element_to_log(date,[year/Yr,month/Mon,day/Day,hour/Hr,minutes/Min,seconds/Sec]) ; true).

:- use_module(eventhandling,[store_virtual_event/1]).
cli_start_selfcheck :-
      %clear_loaded_machines_wo_errors,
      b_set_initial_machine,
      set_animation_mode(b),
      store_virtual_event(clear_specification), % TO DO: try and get rid of the need for this
      start_animation,
      option(selfcheck(Module)),
      (perform_self_check(Module) -> true ; true),
      (test_error_occurred(_) -> error_occurred(selfcheck) ; true),
      fail.
cli_start_selfcheck.


:- dynamic file_loaded/2.
file_loaded(Status) :- file_loaded(Status,_File).

%cli_load_files(RemArgV,NOW) :-
%    if_options_set(l_cspm_pl_file(PlFile),cli_l_cspm_pl_file(PlFile)),!,
%    assert(file_loaded(true,PlFile)),
%    cli_process_loaded_file(NOW,PlFile),
%(RemArgV=[] -> true ; cli_load_files2(RemArgV,NOW)).
cli_load_files([],NOW) :- % no files are provided
   !,
   ( options_allow_start_without_file -> cli_set_empty_machine
     ; we_did_something -> true
     ; otherwise -> print('No file to process'),nl),
   writeln_log_time(start_processing_empty_machine(NOW)),
   cli_process_loaded_file(NOW,'$EMPTY_MACHINE'),
   check_all_expected_errors_occurred(NOW). % check that all expected errors occurred; below they will be checked for each file
cli_load_files(RemArgV,NOW) :- cli_load_files2(RemArgV,NOW,0).


cli_set_empty_machine :- % TO DO: do this more properly here and for initialise_required
%       announce_event(clear_specification),
       set_animation_mode(b),
%       announce_event(start_initialising_specification),
       bmachine:b_set_empty_machine,
       assert(file_loaded(true,empty_machine)).
       %announce_event(specification_initialised).


options_allow_start_without_file :- option(eval_repl(_)).
options_allow_start_without_file :- option(eval_string_or_file(_,_,_,_)).
options_allow_start_without_file :- option(check_log(_)).
options_allow_start_without_file :- option(process_latex_file(_,_)).

we_did_something :- option(print_version).
we_did_something :- option(print_short_version).
we_did_something :- option(check_java_version).

option_only_works_for_single_file(zmq_assertion(_PortStart,_Logfile)).
option_only_works_for_single_file(zmq_master2(_N,_Max,_PortStart,_Workers,_Logfile)).
option_only_works_for_single_file(zmq_master(_Workers,_Logfile)).

clear_loaded_files :- (file_loaded(_) -> clear_loaded_machines_wo_errors ; true).
cli_load_files2([],_,NrFilesProcessed) :- !, print_accumulated_infos(NrFilesProcessed).
cli_load_files2([F1,F2|T],NOW,_NrFilesProcessed) :-
    option(Option),
    option_only_works_for_single_file(Option),!,
    add_error(probcli,'The following option can only be used for a single file: ',Option),
    add_error(probcli,'Multiple files provided: ',[F1,F2|T]),
    halt_prob(NOW).
cli_load_files2(RemArgV,NOW,NrFilesProcessed) :-
    %print_total_probcli_timer,
    clear_loaded_files,
    retractall(file_loaded(_,_)),
    RemArgV = [MainFile0|Rest],!,
    N1 is NrFilesProcessed+1,
    cli_load_files3(MainFile0,Rest,NOW,N1).
cli_load_files3(MainFile0,Rest,NOW,NrOfFile) :-
    safe_absolute_file_name(MainFile0,MainFile,[access(none)]), % converts Windows slash into Unix slash,...
    if_option_set(file_info,print_file_info(MainFile)),
    ((Rest=[_|_] ; NrOfFile>1)
     -> length(Rest,RLen), Tot is NrOfFile+RLen,
        format('~n~n% Processing file ~w/~w: ~w~n',[NrOfFile,Tot,MainFile]) % was formatsilent
      ; true),
    start_xml_feature(process_file,filename,MainFile,FINFO),
    ( file_exists(MainFile) ->
        debug_println(6,file_exists(MainFile)),
        ( load_main_file(MainFile,NOW,Already_FullyProcessed) ->
          (Already_FullyProcessed==true -> true
           ; assert(file_loaded(true,MainFile)),
             writeln_log_time(start_processing(NOW)),
             start_probcli_timer(Timer),
             (cli_process_loaded_file(NOW,MainFile)
               -> stop_probcli_debug_timer(Timer,'% Finished processing file after')
               ;  print_error('### Processing file failed: '), print_error(MainFile)
             ),
             writeln_log_time(finished_processing(NOW))
           )
        ; otherwise ->
            assert(file_loaded(error,MainFile)),
            print_error('### Loading Machine Failed'),
            writeln_log_time(loading_failed(NOW,MainFile)),
            error_occurred(load_main_file)
        ),nl
    ; otherwise ->
        nl, assert(file_loaded(error,MainFile)),
        (number(MainFile0)
            -> add_error(load_main_file,'Command-line argument is a number which is not associated with a command and does not exist as file: ',MainFile0)
         ; atom_codes(MainFile0,[45|_]) % starts with a dash - : probably an illegal command-line option
             -> add_error(load_main_file,'Specified Option or File does not exist: ',MainFile0)
          ; add_error(load_main_file,'Specified File does not exist: ',MainFile)
        )
    ),
    check_all_expected_errors_occurred(NOW),
	stop_xml_feature(process_file,FINFO),

    reset_expected_error_occurred, % reset for next file
    NOW1 is NOW+1, update_time_stamp(NOW1),
    cli_load_files2(Rest,NOW1,NrOfFile).

print_file_info(F) :-
   print('Specification_File('), print(F), print(')'),nl.

% ---------------------

% process all the commands for a loaded file:
cli_process_loaded_file(NOW,MainFile) :-
    (real_error_occurred -> print_error('% *** Errors occurred while loading ! ***'),nl,nl ; true),
    get_errors, reset_errors,
    if_option_set(kodkod_performance(KPFile,Iterations),
                  compare_kodkod_performance1(KPFile,Iterations,NOW)),
    if_option_set(kodkod_comparision(MaxResiduePreds),
                  test_kodkod_and_exit(MaxResiduePreds,NOW)),
    % if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)), %% moved to later to ensure B machine is precompiled; allows e.g. type_check_csp_and_b to run

    if_option_set(csp_main(MAINPROC),
                  set_cspm_main_process(MAINPROC)),

    if_option_set(zmq_master2(N,Max,PortStart,Recover,Logfile), zmq_start_master(invariant,N,Max,PortStart,Recover,Logfile)),
    if_option_set(zmq_master(Logfile), zmq_start_master(invariant,200,-1,5000,0,Logfile)),

    % STARTING ANIMATION/MODEL CHECKING
    cli_start_animation(NOW),


    if_option_set_loaded(trace_check,trace_check,
                         cli_start_trace_check(MainFile)),

    cli_process_loaded_file_afer_start_animation(NOW).

cli_process_loaded_file_afer_start_animation(NOW) :-
    ifm_option_set(cli_print_machine_statistics,
                  cli_print_machine_statistics),
    ifm_option_set(pretty_print_internal_rep(PPFILE1),
                  b_write_machine_representation_to_file(needed,PPFILE1)),
    ifm_option_set(pretty_print_all_types(PPFILE2),
                  b_write_machine_representation_to_file(all,PPFILE2)),
    ifm_option_set(pretty_print_internal_rep_to_B(PPFILE3),
                  b_write_eventb_machine_to_classicalb_to_file(PPFILE3)),

    if_option_set_loaded(state_trace(TraceFile),state_trace,
                         cli_start_trace_state_check(TraceFile)),

    if_option_set(evaldot(EvalDotF),
                  set_eval_dot_file(EvalDotF)),

    (initialise_required
      -> check_loaded(initialise),
         cli_start_initialisation(NOW),
         writeln_log_time(initialised(NOW))
      ;  true),

    if_option_set(zmq_assertion(PortStart, ZMQLogfile),
                  zmq_start_master(assertion,PortStart, ZMQLogfile)),

    if_option_set(cli_start_mc_with_tlc,cli_start_mc_with_tlc),
    if_option_set(cli_start_sym_mc_with_lts(LType),cli_start_sym_mc_with_lts(LType)),

    if_option_set(cli_symbolic_model_check(Algorithm),cli_symbolic_model_check(Algorithm)),

    if_option_set_loaded(cli_check_properties,check_properties,
                         cli_check_properties(NOW)),
    ifm_option_set_loaded(cli_check_assertions(ALL,ReqInfos),check_assertions,
                         cli_check_assertions(ALL,ReqInfos,NOW)),
    if_option_set(set_goal(GOAL),
                  cli_set_goal(GOAL)),
    if_option_set(set_searchscope(SCOPE),
                  cli_set_searchscope(SCOPE)),
    ifm_option_set_loaded(cli_mc(Nr),model_check,
                         cli_start_model_check(Nr,NOW)),
    ifm_option_set_loaded(cli_random_animate(Steps,ErrOnDeadlock),animate,
                         cli_random_animate(NOW,Steps,ErrOnDeadlock)),
    ifm_option_set_loaded(execute(ESteps,ErrOnDeadlock,From),execute,
                         cli_execute(ESteps,ErrOnDeadlock,From)),
    ifm_option_set_loaded(logxml_write_vars(Prefix),logxml_write_vars,
                         logxml_write_vars(Prefix)),
    if_option_set_loaded(pa_check,predicate_analysis,
                         test_predicate_analysis),

    cbc_check(NOW),
    if_options_set(generate_read_write_matrix_csv(RWCsvFile),
                   generate_read_write_matrix(RWCsvFile)),
    if_options_set(feasibility_analysis_csv(TimeOut,EnablingCsvFile),
                   do_feasibility_analysis(TimeOut,EnablingCsvFile)),
    ifm_option_set_loaded(mcm_tests(ADepth1,AMaxS,ATarget1,Output1),mcm_test_cases,
                         mcm_test_case_generation(ADepth1,AMaxS,ATarget1,Output1)),
    ifm_option_set_loaded(all_deadlocking_paths(File),all_deadlocking_paths,
                         write_all_deadlocking_paths_to_xml(File)),
    ifm_option_set_loaded(cbc_tests(ADepth2,ATarget2,Output2),cb_test_cases,
                         cbc_test_case_generation(ADepth2,ATarget2,Output2)),
    ifm_option_set_loaded(test_description(TestDescFile),cb_test_cases,
                         test_generation_by_xml_description(TestDescFile)),
    if_options_set(csp_in_situ_refinement_check(RP,RType,RQ),
                  cli_csp_in_situ_refinement_check(RP,RType,RQ,NOW)),
    if_options_set(csp_checkAssertion(Proc,Model,AssertionType),
                   cli_checkAssertion(Proc,Model,AssertionType,NOW)),
    if_options_set(check_csp_assertion(Assertion),
                   cli_check_csp_assertion(Assertion,NOW)),
    if_options_set(refinement_check(RefFile,PerformSingleFailures,RefNrNodes),
                  cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW)),
    if_options_set(ctl_forumla_model_check(Formula,Expected),cli_ctl_model_check(Formula,Expected,_)),
    % TO DO print ctl/ltl statistics
    if_options_set(csp_get_assertions,cli_csp_get_assertions),
    if_options_set(eval_csp_expression(CspExpr),cli_eval_csp_expression(CspExpr)),
    if_options_set(csp_add_unicode(UnicodeFile1), cli_csp_add_unicode(UnicodeFile1)),
    if_options_set(csp_remove_unicode(UnicodeFile2), cli_csp_remove_unicode(UnicodeFile2)),
    if_options_set(csp_translate_to_file(PlFile),cli_csp_translate_to_file(PlFile)),
    if_options_set(get_min_max_coverage(MinMxFileName),cli_get_min_max_coverage(MinMxFileName)),
    if_options_set(get_coverage_information(CovFileName),cli_get_coverage_information(CovFileName)),
    if_options_set(vacuity_check,cli_vacuity_check),
    if_option_set_loaded(check_goal,check_goal,cli_check_goal),
    if_option_set_loaded(animate,animate,
                         (animate_machine -> true ; true)),
    if_option_set(ltsmin, start_ltsmin_srv('/tmp/ltsmin.probz', NOW)),
    if_option_set(ltsmin2(EndpointPath), start_ltsmin_srv(EndpointPath, NOW)),
    if_option_set(ltsmin_ltl_output(Path), ltsmin_ltl_output(Path, NOW)),
    evaluate_from_commandline,
    if_option_set_loaded(ltl_assertions,check_ltl_assertions,
                         (timeout_call(ltl_check_assertions,NOW) -> true; true)),
    if_option_set_loaded(ltl_formula_model_check(Formula,Expected),check_ltl_assertions,
                         (option(cli_start_sym_mc_with_lts(_))-> true % we request LTSMin, do not start prob model check
                          ; timeout_call(cli_ltl_model_check(Formula,Expected,_),NOW) -> true; true)),
    ifm_option_set_loaded(ltl_file(LtlFilename),check_ltl_file,
                         (ltl_check_file(LtlFilename) -> true; true)),
    ifm_option_set_loaded(history(HistoryFilename),history,
                         cli_print_history(HistoryFilename)),
    ifm_option_set_loaded(print_values(ValuesFilename),values,
                         cli_print_values(ValuesFilename)),
    ifm_option_set_loaded(print_all_values(ValuesDirname),values,
                         cli_print_all_values(ValuesDirname)),
    if_options_set(save_state_for_refinement(SaveRefF),
                  refinement_checker:tcltk_save_specification_state_for_refinement(SaveRefF)),
    if_options_set(dot_command(DCommand,DotFile),
                  call_dot_command(DCommand,DotFile,[])),
    if_options_set(dot_command_for_expr(DECommand,Expr,DotFile,Opts),
                  call_dot_command_for_expr(DECommand,Expr,DotFile,Opts)),
    if_options_set(evaluate_expression_over_history_to_csv_file(HistExpr,HistDotFile),
                  user:evaluate_expression_over_history_to_csv_file(HistExpr,HistDotFile)),
    if_options_set(enabling_analysis_csv(EnablingCsvFile),
                   do_enabling_analysis_csv(EnablingCsvFile,NOW)),
    if_options_set(process_latex_file(LatexF1,LatexF2),
                   process_latex_file(LatexF1,LatexF2)),
    if_options_set(expect_evaluation(EEvals),
                   check_if_predicates_were_evaluated(EEvals)),
    ifm_option_set(coverage(Nodes,Operations,ShowEnabledInfo),
                  cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW)),
    ifm_option_set(check_statespace_hash(ExpectedHash,Kind),
                  cli_check_statespace_hash(ExpectedHash,Kind)),
    ifm_option_set(variable_coverage,
                  cli_show_variable_coverage(NOW)),
    ifm_option_set(coverage(ShowEnabledInfo),
                  cli_show_coverage(ShowEnabledInfo,NOW)),
    if_option_set(save_state(StateFile),
                  (debug_println(20,'% Saving state space to file'),
                   state_space:tcltk_save_state(StateFile))),
    if_option_set(prob_profile,
                  prob_profile),
    if_option_set(statistics,
                  cli_print_statistics),
    if_option_set(profile_statistics,
                  profile_statistics),
    if_option_set(show_cache,
                  show_cache),
    if_option_set(check_complete, check_complete),
    if_option_set(check_complete_operation_coverage, check_complete_operation_coverage),
    if_option_set(check_scc_for_ltl_formula(LtlFormula,SCC),cli_check_scc_for_ltl_formula(LtlFormula,SCC)).

:- use_module(value_persistance,[show_cache_file_contents/0,show_cache_file_constants/0]).
show_cache :- option(verbose),!,show_cache_file_contents,nl.
show_cache :- show_cache_file_constants,nl.

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%cli_print_statistics :- pen,nl,garbage_collect,statistics,nl,state_space:state_space_initialise_with_stats.

% old profiler
:- use_module(hit_profiler).
:- use_module(b_operation_cache,[print_op_cache/0]).
cli_print_statistics :- profile_statistics,
    nl,garbage_collect,statistics,nl,
    print_op_cache,
    state_space:state_space_initialise_with_stats.

:- use_module(state_space,[not_all_transitions_added/1, not_invariant_checked/1]).
check_complete :-
  (tcltk_find_max_reached_node(Node1) ->
     add_error(check_complete,'Maximum number of transitions reached for at least one state: ',Node1) ; true),
  (not_all_transitions_added(Node2) ->
     add_error(check_complete,'At least one state was not examined: ',Node2) ; true),
  (not_invariant_checked(Node3) ->
     add_error(check_complete,'The invariant was not checked for at least one state: ',Node3) ; true).

check_complete_operation_coverage :-
  (state_space: operation_name_not_yet_covered(OpName) ->
     add_error(check_complete_operation_coverage,'At least one operation is not covered: ', OpName)
  ; true).

show_operation_coverage_summary(NOW) :-
   findall(ON, specfile:get_possible_event(ON), Possible), length(Possible,Nr),
   findall(OpName, state_space: operation_name_not_yet_covered(OpName), Uncovered),
   writeln_log(uncovered_info(NOW,Nr,Uncovered)),
   (Uncovered=[] -> format(' All ~w possible operations have been covered',[Nr]),nl
     ; format(' The following operations (out of ~w) were not covered:~n   ~w~n',[Nr,Uncovered])).
show_initialisation_summary(NOW) :-
   findall(ID,state_space:is_concrete_constants_state_id(ID),L),
   length(L,Nr), N1 is Nr+1, % for root
   writeln_log(uninitialised_states(NOW,N1)),
   format(' Uninitialised states: ~w (root and constants only)~n',[N1]).

% ---------------------

animation_mode_does_not_support_animation(File) :-
    loaded_main_file(smt2,File).
cli_start_animation(NOW) :-
    file_loaded(true,LoadedFile),
    \+ animation_mode_does_not_support_animation(LoadedFile),
    !,
    debug_println(20,'% Starting Animation'),
    writeln_log_time(start_animation(NOW)),
    start_probcli_timer(Timer1),
    start_animation_without_computing,
    stop_probcli_debug_timer(Timer1,'% Finished Starting Animation'),
    if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)),

    xml_log_machine_statistics,
    getAllOperations(Ops),
    debug_print(20,'Operations: '), debug_println(20,Ops),

    (we_need_only_static_assertions(ALL)
      -> debug_println(20,'% Projecting on static ASSERTIONS'),
         print('% PROJECTING ON STATIC ASSERTIONS'),nl,
         b_interpreter:set_projection_on_static_assertions(ALL) ; true),

    (option(load_state(File))
      -> debug_println(20,'% Loading stored state from file'),
         state_space:tcltk_load_state(File)
       ; computeOperations_for_root_required ->
         debug_println(20,'% Searching for valid initial states'),
         start_probcli_timer(Timer2),
         cli_computeOperations(EO),
         stop_probcli_debug_timer(Timer2,'% Finished searching for valid initial states'),
         debug_println(10,EO)
       ; debug_println(20,'% No initialisation required')
    ).
cli_start_animation(_NOW).

start_animation_without_computing :-
    update_preferences_from_spec,
    set_prefs, % override SET_PREF in DEFINITIONS with values from command-line;
    start_animation,
    ifm_option_set(add_additional_property(PROP),
                   cli_add_additional_property(PROP)),
    if_option_set(plugin(Plugin),
            (plugins:initialise_plugins,plugins:start_plugin_for_current_mode(Plugin))),
    get_errors.

:- use_module(state_space,[current_expression/2]).
:- use_module(extension('counter/counter')).
:- use_module(tools_lists,[count_occurences/2]).
% an execution engine with minimal overhead: states are not stored in visited_expression database, only first enabled operation is taken
cli_execute(Steps,ErrorOnDeadlock,FromWhere) :-
   FromWhere=from_all_initial_states,!,
  % try out all initial states and from each of those perform deterministic execution
    start_ms_timer(Start),
    format('Running execute from all initial states~n',[]),
    reset_counter(cli_execute_inits),
	findall(Result,
	 (cli_trans(root,Action,CurState,0), %print(Action),nl,
	     (\+ functor(Action,'$setup_constants',_)
		   -> inc_counter(cli_execute_inits,Nr),
		      format('~nExecuting model from initial state ~w~n',[Nr]),
			  (option(animate_stats) -> print_state_silent(CurState) ; true),
			  (cli_execute_from(CurState,Steps,ErrorOnDeadlock,1,Result) -> true)
		   ;  format('~nInitialising state~n',[]), % we need to execute initialise_machine
			  cli_trans(CurState,_ActionName,NewState,0),
			  inc_counter(cli_execute_inits,Nr),
			  format('~nExecuting model from initial state ~w with constants~n',[Nr]),
			  (option(animate_stats) -> print_state_silent(NewState) ; true),
			  (cli_execute_from(NewState,Steps,ErrorOnDeadlock,2,Result) -> true)
	  )), Results),
    get_counter(cli_execute_inits,Nr),
    format('---------~nTotal runtime for all ~w executions:',[Nr]),nl,
    stop_ms_timer(Start),
    count_occurences(Results,Occs), format('Results: ~w~n',[Occs]).
cli_execute(Steps,ErrorOnDeadlock,_) :-
    current_expression(_ID,CurState),
    start_xml_feature(execute,max_steps,Steps,FINFO),
    cli_execute_from(CurState,Steps,ErrorOnDeadlock,1,_Result),
    stop_xml_feature(execute,FINFO).
% TO DO: add transition from root to executed state ?

cli_execute_from(CurState,Steps,ErrorOnDeadlock,FirstStepNr,Result) :-
    start_ms_timer(Start),
    (b_top_level_operation(_) -> temporary_set_preference(filter_unused_constants,true,CHNG) ; true), % we can filter out unused constants, unless there are no operations in which case the user probably wants to see the constant values
    cli_execute_aux(FirstStepNr,Steps,CurState,ErrorOnDeadlock,'$NO_OPERATION',Result),
    (b_top_level_operation(_) -> reset_temporary_preference(filter_unused_constants,CHNG) ; true),
    stop_ms_timer(Start).

:- use_module(external_functions,[reset_side_effect_occurred/0, side_effect_occurred/1]).
cli_execute_aux(Nr,Steps,CurState,_ErrorOnDeadlock,_LastActionName,Result) :- Nr>Steps,!,
    formatsilent('Stopping execution after ~w steps~n',[Steps]), Result = stopped,
    print_state_silent(CurState),
    cli_execute_add_virtual_transition(Steps,CurState,Result).
cli_execute_aux(Nr,Steps,CurState,ErrorOnDeadlock,LastActionName,Result) :-
    ((Nr mod 50000 =:= 0,\+option(animate_stats),\+option(silent))
     -> format('Step ~w (after ~w)~n',[Nr,LastActionName]),
        (option(verbose) -> print_state_silent(CurState) ; true),
        print('Memory used: '),tools:print_memory_used_wo_gc,flush_output,nl ; true),
    (\+ cli_invariant_ok(CurState,LastActionName) -> % also recognises no_inv command
           N1 is Nr-1,
           tools:ajoin(['INVARIANT VIOLATED after ',N1,' steps (after ',LastActionName,').'],ErrMsg),
           format('~w~n',[ErrMsg]),!,
           print_state_silent(CurState),
           error_occurred_with_msg(invariant_violation,ErrMsg),
           Result=invariant_violation,
           cli_execute_add_virtual_transition(N1,CurState,Result)
    ; \+ cli_assertions_ok(CurState,LastActionName) -> % also recognises no_inv command
           N1 is Nr-1,
           format('ASSERTION VIOLATED after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState),
           error_occurred(assertion_violation), Result=assertion_violation,
           cli_execute_add_virtual_transition(N1,CurState,Result)
     ; cli_goal_found(CurState) -> % also recognizes no_goal command
           N1 is Nr-1,
           format('GOAL FOUND after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState), Result=goal_found,
           cli_execute_add_virtual_transition(N1,CurState,Result)
    ;      reset_side_effect_occurred,
           cli_trans(CurState,ActionName,NewState,Nr),
           !,
           N1 is Nr+1,
           ((NewState=CurState, \+ side_effect_occurred(file))
             -> formatsilent('Infinite loop reached after ~w steps (looping on ~w).~n',[N1,ActionName]),
                Result=loop
              ; cli_execute_aux(N1,Steps,NewState,ErrorOnDeadlock,ActionName,Result))
    ).
cli_execute_aux(Nr,_Steps,CurState,ErrorOnDeadlock,LastActionName,Result) :- N1 is Nr-1,
    formatsilent('Deadlock reached after ~w steps (after ~w).~n',[N1,LastActionName]),
    Result=deadlock,
    ((ErrorOnDeadlock=true,\+ option(no_deadlocks) ) -> error_occurred(deadlock) ; true),
    print_state_silent(CurState),
    cli_execute_add_virtual_transition(N1,CurState,Result).

cli_execute_add_virtual_transition(Steps,CurState,Result) :-
    getCurrentStateID(CurID),
    write_xml_element_to_log(executed,[steps/Steps,result/Result]),
    (Steps>0 -> user:tcltk_add_new_transition(CurID,'$execute'(Steps),ToID,CurState,[]),
                debug_println(added_transition_for_execute(Steps,ToID)),
                tcltk_goto_state('$execute'(Steps),ToID)
             ; true).

% will be used to compute a single successor
% b_operation_cannot_modify_state
cli_trans(CurState,ActionName,NewState,Nr) :- option(animate_stats),!, % provide statistics about the animation
    start_probcli_timer(Timer),
    cli_trans_aux(CurState,ActionName,Act,NewState),
    translate:translate_event(Act,Str),
    truncate_animate_action(Str,TStr),
    format('~w~5|: ~w~n',[Nr,TStr]),
    format(' ~5|',[]),stop_probcli_timer(Timer,' ').
cli_trans(CurState,ActionName,NewState,_Nr) :-
    cli_trans_aux(CurState,ActionName,_,NewState).

cli_trans_aux(CurState,ActionName,Act,NewState) :-
     catch_enumeration_warning_exceptions(
        (throw_enumeration_warnings_in_current_scope,
         cli_execute_trans(CurState,ActionName,Act,NewState), % no time-out !
         (error_occurred_in_error_scope -> ErrorEvent=true ; ErrorEvent=false)
         ),
             (error_occurred(virtual_time_out_execute),
              ActionName = '*** VIRTUAL_TIME_OUT ***', Act=ActionName,
              CurState=NewState) % this forces loop detection above; not very elegant way of signalling
              ),
     ((ErrorEvent==true,option(strict_raise_error))
       -> print('*** ERROR OCCURED DURING EXECUTE ***'),nl, error_occurred(execute),fail
        ; true).

cli_execute_trans(CurState,ActionName,Act,NewState) :-
   specfile_possible_trans_name_for_successors(CurState,ActionName), % to provide feedback
   on_exception(user_interrupt_signal,
      specfile_trans_or_partial_trans(CurState,ActionName,Act,NewState,_TransInfo,Residue), % no time-out !
      (format('~n*** User-Interrupt (CTRL-C) while executing operation "~w"~n~n',[ActionName]),
       throw(user_interrupt_signal))),
    (Residue=[] -> true ; error_occurred(cli_execute_residue(ActionName,Residue,Act))).

print_state_silent(_) :- option(silent),!.
print_state_silent(CurState) :- (option(verbose);\+ b_machine_has_constants),!,
   translate:print_state(CurState),nl.
print_state_silent(CurState) :- remove_constants(CurState,VarState),
   % only print variables
   format('VARIABLES (use -v to see constants or -silent to suppress output):~n',[]),
   translate:print_state(VarState),nl.

is_constant_binding(bind(C,_)) :- b_is_constant(C).
remove_constants(const_and_vars(_,Vars),Res) :- !,Res=Vars.
remove_constants([H|T],Res) :- !,exclude(is_constant_binding,[H|T],Res).
remove_constants(X,X).

% write vars to xml log file if they start with a given prefix
logxml_write_vars(Prefix) :- current_expression(_,CurState),
    remove_constants(CurState,VarState),
    % TO DO: store also final state in xml_log
    write_bstate_to_log(VarState,Prefix).

cli_invariant_ok(_,_) :- option(no_invariant_violations),!. % user asks not to check it
cli_invariant_ok(_,_) :- get_preference(do_invariant_checking,false),!. % user asks not to check it via preference
cli_invariant_ok(CurState,LastActionName) :-
    state_corresponds_to_initialised_b_machine(CurState,BState),!,
    start_probcli_timer(InvTimer),
    (b_specialized_invariant_for_op(LastActionName,Invariant) -> true
       %, print('Specialized invariant: '),translate:print_bexpr(Invariant),nl
      ; b_get_invariant_from_machine(Invariant)),
    cli_test_pred(BState,Invariant),
    stop_probcli_debug_timer(InvTimer,'Finished Invariant Checking').
cli_invariant_ok(_,_). % not yet initialised

cli_test_pred(BState,Pred) :- b_interpreter:b_test_boolean_expression_for_ground_state(Pred,[],BState ),!.
cli_test_pred(_,Pred) :- bsyntaxtree:get_texpr_label(Pred,Label),
    format('Label = ~w~n',[Label]),fail.
cli_test_pred(_,Pred) :- bsyntaxtree:get_texpr_description(Pred,Desc),
    format('Description = ~w~n',[Desc]),fail.
cli_test_pred(_,Pred) :- bsyntaxtree:get_texpr_pos(Pred,Pos), Pos \= none, translate_span(Pos,Str),
    format('Location = ~w~n',[Str]),fail.

:- use_module(bmachine,[get_assertions_from_machine/2]).
% TO DO: also check static assertions
cli_assertions_ok(_,_) :- option(no_assertion_violations),!. % user asks not to check it
cli_assertions_ok(CurState,_LastActionName) :-
    state_corresponds_to_initialised_b_machine(CurState,BState),
    get_assertions_from_machine(dynamic,Assertions), % TO DO: do something similar to b_specialized_invariant_for_op
    !,
    start_probcli_timer(AssTimer),
    %nl,nl,print(check),nl,maplist(translate:print_bexpr,Assertions),nl,
    maplist(cli_test_pred(BState),Assertions),
    stop_probcli_debug_timer(AssTimer,'Finished Checking Assertions').
cli_assertions_ok(_,_). % not yet initialised or no assertions


cli_goal_found(_):- option(no_goal),!,fail.
cli_goal_found(CurState) :-
   b_get_machine_goal(Goal),
   state_corresponds_to_initialised_b_machine(CurState,BState),
   b_interpreter:b_test_boolean_expression_for_ground_state(Goal,[],BState ).


% random animation
:- public cli_random_animate/2. % for repl to use -animate command
cli_random_animate(Steps,Err) :- probcli_time_stamp(NOW),
    cli_random_animate(NOW,Steps,Err).
cli_random_animate(_NOW,Steps,ErrorOnDeadlock) :-
    start_xml_feature(random_animate,max_steps,Steps,FINFO),
    start_ms_timer(S),
    perform_random_steps(Steps,ErrorOnDeadlock),
    stop_ms_timer(S,Res),
    %tcltk_save_history_as_trace_file(user),
    printsilent(finished_random_animate(Steps,Res)),nls,
    stop_xml_feature(random_animate,FINFO).

we_need_only_static_assertions(ALL) :- specfile:b_or_z_mode,
  (option(cli_check_assertions(main,_)) -> ALL=main
     ; option(cli_check_assertions(ALL,_))),
   % we do an assertion check
   \+ ((option(A), option_requires_all_properties(A))),
   b_get_assertions(ALL,dynamic,[]). % the assertions do not reference variables

% do we need all properties/constants of the machine, or only certain ones (e.g., static)
option_requires_all_properties(cli_mc(_)).
option_requires_all_properties(cli_check_properties(_)). % we may want to check all properties
option_requires_all_properties(cli_random_animate(_)).
option_requires_all_properties(trace_check).
option_requires_all_properties(state_trace(_)).
option_requires_all_properties(mcm_tests(_,_,_,_)).
option_requires_all_properties(cbc_tests(_,_,_)).
option_requires_all_properties(animate).
option_requires_all_properties(initialise). % INITIALISATION may access constants
option_requires_all_properties(eval_repl(_)).
option_requires_all_properties(eval_string_or_file(_,_,_,_)).
option_requires_all_properties(ltl_assertions).
option_requires_all_properties(ltl_file(_)).
option_requires_all_properties(refinement_check(_,_,_)).
option_requires_all_properties(cli_random_animate(_)).
option_requires_all_properties(cli_start_mc_with_tlc).
option_requires_all_properties(cli_symbolic_model_check(_)).

cli_start_trace_check(MainFile) :-
     debug_println(20,'% Starting Trace Check: '),
     ( check_default_trace_for_specfile(MainFile) -> true
     ; otherwise -> error_occurred(trace_check)).
cli_start_trace_state_check(File) :-
     debug_println(20,'% Starting Trace Check: '),
     ( tcltk_check_state_sequence_from_file(File) -> true
     ; otherwise -> error_occurred(state_trace)).

% is it necessary to compute enabled operations for root state
computeOperations_for_root_required :- initialise_required.
computeOperations_for_root_required :- option(trace_check).
computeOperations_for_root_required :- option(state_trace(_)).
computeOperations_for_root_required :- option(ltl_assertions).
computeOperations_for_root_required :- option(cli_random_animate(_,_)).
computeOperations_for_root_required :- option(socket(_)).
computeOperations_for_root_required :- option(cli_mc(_)).
computeOperations_for_root_required :- option(ltl_file(_)).
computeOperations_for_root_required :- option(ltl_formula_model_check(_,_)).
computeOperations_for_root_required :- option(ctl_forumla_model_check(_,_)).
computeOperations_for_root_required :- option(refinement_check(_,_,_)).
computeOperations_for_root_required :- option(csp_in_situ_refinement_check(_,_)).
computeOperations_for_root_required :- option(csp_checkAssertion(_,_)).
computeOperations_for_root_required :- option(mcm_tests(_,_,_,_)).
computeOperations_for_root_required :- option(mcm_cover(_)).

% is an initialisation mandatory:
initialise_required :- option(initialise), \+ file_loaded(true,empty_machine).
initialise_required :- \+ option(trace_check), \+ option(state_trace(_)),
                       \+ option(load_state(_)),
                       \+ file_loaded(true,empty_machine), init_req2.
init_req2 :- option(cli_check_properties).
init_req2 :- option(zmq_assertion(_,_)).
init_req2 :- option(cli_check_assertions(_,_)).
init_req2 :- option(eval_string_or_file(_,_,_,_)). % ensure that we initialise/precompile empty machine in case no main file specified; currently no longer required

initialise :- probcli_time_stamp(NOW),cli_start_initialisation(NOW). % for REPL

cli_start_initialisation(NOW) :-
    debug_println(20,'% Performing INITIALISATION: '),
    (perform_random_initialisation -> true ;
      writeln_log_time(cli_start_initialisation_failed(NOW)),
      fail).

:- use_module(state_space,[current_state_corresponds_to_setup_constants_b_machine/0]).
:- public cli_check_properties/0. % for REPL
cli_check_properties :- probcli_time_stamp(NOW),
    cli_check_properties(NOW).
cli_check_properties(NOW) :-
    print('% Checking PROPERTIES: '),nl,
    writeln_log_time(starting_check_properties(NOW)),
    ( current_state_corresponds_to_setup_constants_b_machine ->
        set_analyse_hook('_P'),
        predicate_evaluator:tcltk_analyse_properties(_PROPRES,PROPInfos),
        unset_analyse_hook,
        print(PROPInfos),nl,   % ex: [total/33,true/29,false/0,unknown/4,timeout/4,runtime/49950]
        accumulate_infos(properties,PROPInfos),
        write_important_xml_element_to_log(check_properties,PROPInfos),
        (predicate_evaluator:check_summary_all_true(PROPInfos) -> true
          ; print_error('### Not all PROPERTIES true'), error_occurred(check_properties))
    ; otherwise ->
        (tcltk_debug_properties(PROPRES,false,Satisfiable) ->
         print(PROPRES),nl,
         print(Satisfiable),nl
         ; error_occurred(debug_properties_failed))
    ),
    writeln_log_time(finished_check_properties(NOW,PROPInfos)),
    loaded_root_filename(RootName),
    format('% Finished checking PROPERTIES of ~w~n',[RootName]).
:- public cli_check_assertions/2. % for REPL
cli_check_assertions(ALL,RI) :-
    probcli_time_stamp(NOW),
    cli_check_assertions(ALL,RI,NOW).
cli_check_assertions(ALL,ReqInfos,NOW) :-
    print('% Checking ASSERTIONS: '),nl,
    writeln_log_time(starting_check_assertions(NOW)),
    set_analyse_hook('_A'), % for dot output, in case users wants to generate dot files for assertions
    predicate_evaluator:tcltk_analyse_assertions(ALL,_ASSRES,Infos),
    unset_analyse_hook,
    print(Infos),nl,
    accumulate_infos(assertions,Infos),
    write_important_xml_element_to_log(check_assertions,Infos),
    check_required_infos(ReqInfos,Infos,check_assertions),
    writeln_log_time(finished_check_assertions(NOW,Infos)),
    loaded_root_filename(RootName),
    format('% Finished checking ASSERTIONS of ~w~n',[RootName]),!.
cli_check_assertions(ALL,ReqInfos,NOW) :-
    add_internal_error('Analyse ASSERTIONS unexpectedly failed',cli_check_assertions(ALL,ReqInfos,NOW)),
    error_occurred(internal_error).
cli_set_goal(GOAL) :-
    debug_println(20,set_goal(GOAL)), %print(set_goal(GOAL)), nl,
    bmachine:b_set_machine_goal(GOAL).
cli_add_additional_property(PROP) :-
    debug_println(20,add_additional_property(PROP)),
    bmachine:add_additional_property(PROP).
cli_set_searchscope(GOAL) :-
    debug_println(20,set_searchscope(GOAL)),
    format('Setting SCOPE for verification: ~w~n (Only states satsifying this predicate will be examined)~n',[GOAL]),
    bmachine:b_set_machine_searchscope(GOAL).
cli_check_goal :- \+ b_get_machine_goal(_),!,
    add_error(cli_check_goal,'No GOAL DEFINITION found'),
    error_occurred(cli_check_goal).
cli_check_goal :-
    print('% Checking GOAL predicate: '),nl,
    tcltk_analyse_goal(_List,Summary),
    debug_println(20,Summary),
    accumulate_infos(check_goal,Summary),
    write_important_xml_element_to_log(check_goal,Summary),
    check_required_infos([false/0,unknown/0],Summary,check_goal).
:- public cli_mc/1. % for REPL
cli_mc(Nr) :- probcli_time_stamp(NOW), cli_start_model_check(Nr,NOW).
cli_start_model_check(Nr,NOW) :-
    start_xml_feature(model_check,max_states,Nr,FINFO),
    model_check(Nr,Time,MCRes,NOW),
    %nl,
    stop_xml_feature(model_check,FINFO),
    writeln_log_time(model_check(NOW,Nr,Time,MCRes)).
cli_start_mc_with_tlc :-
    ((animation_mode(b), \+ animation_minor_mode(eventb)) -> true; error_manager: add_error_and_fail(mc_with_tlc,'TLC4B tool can be used only for classical B models.')),
    % TO DO: use b_write_eventb_machine_to_classicalb_to_file to do conversion
    on_exception(error(E,_),
                 safe_absolute_file_name(prob_lib('TLC4B.jar'),TLC4BTool),
                 error_manager: add_error_fail(get_tlc_command,'Could not find TLC4B.jar file.',E)),
    start_xml_feature(model_check_with_tlc,tlc4bjar,TLC4BTool,FINFO),
    construct_and_execute_tlc_command(TLC4BTool),
    stop_xml_feature(model_check_with_tlc,FINFO).

:- use_module(system_call,[system_call/4]).
construct_and_execute_tlc_command(TLC4BTool) :-
    parsercall: get_java_command_path(JavaCmd),
    loaded_main_file(File),
    ((get_preference(tlc_number_of_workers,TLCWorkers), TLCWorkers>1)
     -> number_codes(TLCWorkers,CC), atom_codes(TLAWA,CC), WW = ['-workers',TLAWA]
      ; WW=[]),
    (option(no_assertion_violations) -> WA = ['-noass'] ; WA=[]),
    (option(no_deadlocks) -> WD = ['-nodead'] ; WD=[]),
    (option(no_invariant_violations) -> WI = ['-noinv'] ; WI=[]),
    (option(no_goal) -> WG = ['-nogoal'] ; WG=[]),
    (option(no_ltl) -> WL = ['-noltl'] ; WL=[]),
    append([WW,WA,WD,WI,WG,WL],XtraTLCArgs),
    append(XtraTLCArgs,[File],TLCArgs),
    print(tlc_args(TLCArgs)),nl,
    statistics(walltime,[W1,_]),
    system_call: system_call(JavaCmd, ['-jar', TLC4BTool | TLCArgs], Text,JExit),
    statistics(walltime,[W2,_]),
    WTime is W2-W1,
    format('exit : ~w  walltime: ~w ms~n',[JExit,WTime]),
    (JExit=exit(0)
     ->  true
     ;   error_manager: add_error(construct_and_execute_tlc_command,'Error while model checking with TLC: ',TLC4BTool/File),
         name(T,Text),
         error_manager: add_error_fail(construct_and_execute_tlc_command,'Std error: ',T)
    ).

% SymbolicOrSequential = symbolic or sequential
cli_start_sym_mc_with_lts(SymbolicOrSequential) :-
    (option(no_deadlocks) -> NoDead = true ; NoDead = false),
    (option(no_invariant_violations) -> NoInv = true ; NoInv = false), % does LTSMin support goal checking
    findall(Option,option(ltsmin_option(Option)),MoreFlags1),
    findall(ltl_formula(LTLF),option(ltl_formula_model_check(LTLF,_)),MoreFlags2),
    append(MoreFlags1,MoreFlags2,MoreFlags),
    (NoDead = false, NoInv = false ->
        print('ERROR: cannot start LTSmin with both deadlock and invariant checking'),nl,flush_output
      ; formatsilent('starting prob2lts-sym/seq (flags nodead=~w, noinv=~w, moreflags=~w)~n',[NoDead,NoInv,MoreFlags])),
    start_ltsmin(SymbolicOrSequential, [NoDead, NoInv], MoreFlags).
    % TO DO: start lts-sym + start start_ltsmin_srv('/tmp/ltsmin.probz', NOW) + print output

:- use_module(symbolic_model_checker(ic3), [ic3_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(ctigar), [ctigar_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(kinduction), [kinduction_symbolic_model_check/1,
                                                   tinduction_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(bmc), [bmc_symbolic_model_check/1]).
cli_symbolic_model_check(Algorithm) :-
    debug_format(20,'% Starting Symbolic Model Check. Using ~w Algorithm', [Algorithm]),
    start_xml_feature(model_check,algorithm,Algorithm,FINFO),
    (animation_mode(b)
     -> true
     ;  error_manager:add_error_and_fail(cli_symbolic_model_check,'Symbolic Model Checking is currently only available for B and Event-B.')),
    perform_symbolic_model_checking(Algorithm,Result),
    handle_symbolic_model_check_result(Result),
    stop_xml_feature(model_check,FINFO).

perform_symbolic_model_checking(ic3,Result) :- !, ic3:ic3_symbolic_model_check(Result).
perform_symbolic_model_checking(ctigar,Result) :- !, ctigar:ctigar_symbolic_model_check(Result).
perform_symbolic_model_checking(kinduction,Result) :- !, kinduction:kinduction_symbolic_model_check(Result).
perform_symbolic_model_checking(tinduction,Result) :- !, kinduction:tinduction_symbolic_model_check(Result).
perform_symbolic_model_checking(bmc,Result) :- !, bmc:bmc_symbolic_model_check(Result).
perform_symbolic_model_checking(Alg,_) :- add_error_fail(cli_symbolic_model_check,'Invalid symbolic model checking algorithm: ',Alg).

handle_symbolic_model_check_result(counterexample_found) :- !, error_occurred(invariant_violation).
handle_symbolic_model_check_result(property_holds) :- !,
    format('Model checking complete, invariant holds~n',[]).
handle_symbolic_model_check_result(solver_and_provers_too_weak) :- !,
    format('Model checking incomplete because a constraint could not be solved in time~n',[]),
    error_occurred(model_check_incomplete).
handle_symbolic_model_check_result(limit_reached) :- !,
    format('Model checking incomplete because an iteration limit was reached~n',[]),
    error_occurred(model_check_incomplete).

zmq_start_master(invariant,N,Max,PortStart,Workers,Logfile) :-
    start_animation_without_computing,
    zmq_get_initialisation_term(InitTerm),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    start_master(InitTerm,N,Max,PortStart,Workers,Strict,Logfile),
    halt.
zmq_start_master(assertion,PortStart,Logfile) :-
    current_state_corresponds_to_setup_constants_b_machine,
    animation_mode(b),
    full_b_machine(Machine),
    b_get_assertions(_,static,SAss),
    b_get_assertions(_,dynamic,DAss),
    append(SAss,DAss,Ass),
    count_assertions(Ass,0,N),
    assert(master:assertion_count(N)),
    current_expression(_,State1),
    specfile:state_corresponds_to_set_up_constants(State1,State),
    zmq_get_important_options(Options),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    start_master(assertions(classical_b(Machine,Options),State,Ass),2,-1,PortStart,0,Strict,Logfile),
    halt.

zmq_get_initialisation_term(Term) :-
    (animation_mode(b) ; animation_mode(csp_and_b)), % CSP file not yet added when ZMQ master starts working
    \+ animation_minor_mode(eventb),
    option(add_csp_guide(CspGuide)),
    !, % Classical B + CSP
    debug_println(20,'ZMQ: Transferring CSP || B model'),
    full_b_machine(Machine),
    % TO DO: extract CSP Term rather than file name: will not work for distribution on other file-systems
    zmq_get_important_options(Options),
    Term = classical_b_with_csp(Machine,CspGuide,Options).
zmq_get_initialisation_term(Term) :-
    debug_println(20,'Generating ZMQ Worker Initialisation'),
    animation_mode(b), \+ animation_minor_mode(eventb), !, % Classical B
    debug_println(20,'ZMQ: Transferring Classical-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = classical_b(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(b), animation_minor_mode(eventb), !, % Event-B
    debug_println(20,'ZMQ: Transferring Event-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = eventb(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(cspm),
    loaded_main_file(MainCSPFile),!, % TO DO: pass CSP Prolog term rather than file name (for distribution)
    zmq_get_important_options(Options),
    debug_println(20,'ZMQ: Transferring CSP specification'),
    Term = csp_specification(MainCSPFile,Options).
zmq_get_initialisation_term(_Term) :-
    \+ real_error_occurred, % otherwise error occured while loading
    animation_mode(Mode),
    add_internal_error('Unsupported formalism for ZMQ', Mode),
    fail.

zmq_get_initialisation_term(filename(FN)) :-
    loaded_main_file(FN).

% get important command-line options to be transmitted to probcli worker
zmq_get_important_options(Options) :- findall(O, (option(O), zmq_important_option(O)), Options),
   debug_println(20,transferring_zmq_options_to_workers(Options)).
zmq_important_option(coverage(_)).
zmq_important_option(expect_error(_)).
zmq_important_option(optional_error(_)).
zmq_important_option(file_info).
zmq_important_option(log(_)).
zmq_important_option(print_version).
zmq_important_option(print_short_version).
zmq_important_option(profiling_on).
zmq_important_option(set_card(_,_)).
zmq_important_option(set_pref(_,_)).
zmq_important_option(set_preference_group(_,_)).
zmq_important_option(verbose).
zmq_important_option(statistics).
zmq_important_option(variable_coverage).
zmq_important_option(very_verbose).
% we could consider also supporting: -argv, -cache, -prefs FILE  csp_main(ProcessName) profiling_on prob_profile runtimechecking

% set options received by a zmq worker
zmq_set_important_options(Options) :- debug_println(20,setting_zmq_options(Options)),
     maplist(zmq_set_option,Options).
zmq_set_option(file_info) :- !, file_info.
zmq_set_option(log(F)) :- !,
    generate_time_stamp(Datime,NOW),
    cli_start_logging(F,ascii,NOW,Datime,[zmq_worker]).
zmq_set_option(print_version) :- !, print_version.
zmq_set_option(print_short_version) :- !, print_short_version.
zmq_set_option(profiling_on) :- !, profiling_on.
zmq_set_option(set_card(P,V)) :- !, set_eclipse_card(P,V).
zmq_set_option(set_pref(P,V)) :- !, set_pref(P,V).
zmq_set_option(set_preference_group(P,V)) :- !, set_preference_group(P,V).
zmq_set_option(verbose) :- !, verbose.
zmq_set_option(very_verbose) :- !, very_verbose.
zmq_set_option(O) :- zmq_delayed_option(O),!, assert_option(O). % DO IT LATER
zmq_set_option(O) :- add_internal_error('Unsupported option for ZMQ worker: ',zmq_set_option(O)).

zmq_delayed_option(coverage(_)).
zmq_delayed_option(expect_error(_)).
zmq_delayed_option(expect_error_pos(_,_,_)).
zmq_delayed_option(optional_error(_)).
zmq_delayed_option(statistics).
zmq_delayed_option(variable_coverage).

ltsmin_ltl_output(Filename, NOW) :-
    if_option_set(ltl_formula_model_check(Formula, _),true),
    ltsmin:ltsmin_generate_ltlfile(Formula, Filename),
    halt_prob(NOW). % if we additionally specify -ltsformula, we do not want to model check it


start_ltsmin_srv(X, NOW) :-
    nls,printsilent('Starting LTSMin Server...'),nls,
    if_option_set(ltl_formula_model_check(Formula, _),true),
    ltsmin:ltsmin_init(X, Zocket, Formula),
    ltsmin:ltsmin_loop(Zocket),
    ltsmin:ltsmin_teardown(Zocket, X),
    nls,printsilent('Stopped LTSMin Server.'),nls,
    halt_prob(NOW). % if we additionally specify -ltsformula, we do not want to model check it

zmq_start_worker(Master, Port, Recover, NOW) :-
    start_worker(Master,Port,Recover,zmq_worker_load_model),
    formatsilent('ZMQ worker finished (Master: ~w, Port:~w)~n',[Master,Port]),
    cli_process_loaded_file_afer_start_animation(NOW),
    printsilent('Exiting probcli worker'),nls,
    halt_prob(NOW).

zmq_worker_load_model(classical_b(Machine,Options)) :-
    debug_println(20,'ZMQ WORKER: Loading classical B model'),
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_machine(Machine),
    set_animation_mode(b),
    prob2_interface:start_animation, cli_computeOperations(_)
    .
zmq_worker_load_model(classical_b_with_csp(Machine,CspGuide,Options)) :-
    debug_println(20,'ZMQ WORKER: Loading CSP || B model'),
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_machine(Machine),
    set_animation_mode(b),
    prob2_interface:start_animation,
    tcltk_add_csp_file(CspGuide), % TO DO: use CSP Prolog term rather than filename <----------------
    prob2_interface:start_animation,
    cli_computeOperations(_)
    .
zmq_worker_load_model(eventb(Machine,Options)) :-
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_machine(Machine),
    set_animation_mode(b), set_animation_minor_mode(eventb),
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(csp_specification(CSPFile,Options)) :-
    zmq_set_important_options(Options),
    load_cspm_spec_from_cspm_file(CSPFile), % TO DO: pass CSP Prolog term rather than filename
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(filename(FN)) :-
    printsilent('loading file by filename\n'),flush_output,
    ( is_eventb_b(FN) ->
        eclipse_interface:load_eventb_file(FN)
    ; otherwise ->
        bmachine:b_load_machine_probfile(FN)),
    prob2_interface:start_animation.
zmq_worker_load_model(assertions(Machine,State,Assertions)) :-
   assert(assertion_counter(-1)),
   printsilent(loaded_model_for_assertion_checking),nl,
   zmq_worker_load_model(Machine),
   assert(worker:assertion_state(State)),
   make_assertionwps(Assertions).
   % assert current state
zmq_worker_load_model(Other) :-
    add_internal_error('ZMQ worker: Unexpected machine description', zmq_worker_load_model(Other)),
    fail.


:-dynamic assertion_counter/1.

count_assertions([],A,A).
count_assertions([H|T],A,R) :- bsyntaxtree:conjunction_to_list(H,HL),
                               length(HL,N1),
                               NN is A + N1,
                               count_assertions(T,NN,R).

make_assertionwps([]).
make_assertionwps([H|T]) :- bsyntaxtree:conjunction_to_list(H,HL),
                            sort_assertions(HL,SL),
                            append_assertion(SL),
                            make_assertionwps(T).

append_assertion([]).
append_assertion([H|T]) :- assertion_counter(N),
                           retractall(assertion_counter(_)),
                           N1 is N + 1,
                           assert(assertion_counter(N1)),
                           assert(worker:assertion_task(N1,H)),
                           append_assertion(T).

%assertions_order(A,B) :- term_size(A,NA),term_size(B,NB), NA > NB.
sort_assertions(X,X).
% :- samsort(assertions_order,X,Y).



is_eventb_b(FN) :- append(_,FN,".eventb").
 % load_model(Initialisation)


load_model(classical_b(M)) :- !,eclipse_interface:load_classical_b(M).
load_model(load_event_b_project(A,B,C,D)) :- !, prob2_interface:load_event_b_project(A,B,C,D), start_animation.
load_model(X) :- print('*** Unknown model: '), print(X),nl.


:- use_module(predicate_evaluator).
:- use_module(bmachine,[b_machine_name/1]).
set_analyse_hook(AddPrefix) :- % set a hook to write false/unknown expressions into a dot file
   reset_dot_file_number,
   if_options_set(dot_analyse_output_prefix(_Path),
                  (set_dot_file_prefix_if_option_set(AddPrefix),
                   register_conjunct_error_hook(user:pred_eval_hook))).
unset_analyse_hook :- predicate_evaluator:reset_conjunct_error_hook.

loaded_root_filename(RootName) :- loaded_main_file(MainFile),
                   get_modulename_filename(MainFile,RootName).

set_dot_file_prefix_if_option_set(AddPrefix) :-
   if_options_set(dot_analyse_output_prefix(Path),
                  (loaded_root_filename(RootName),
                   % we could also use b_machine_hierarchy:main_machine_name(RootName)
                   string_concatenate(Path,RootName,P1),
                   string_concatenate(P1,AddPrefix,FullPath),
                   set_dot_file_prefix(FullPath),
                   debug_println(9,dot_file_prefix(FullPath)))).

% Status: true, false, unknown
pred_eval_hook(_Conjunct,true,_EnumWarning,_IsExpanded, _CS) :-
  \+ option(dot_generate_for_all_formulas),!. % don't generate .dot for true formulas, unless explicitly requested
pred_eval_hook(Conjunct,Status,_EnumWarning,_IsExpanded, CS) :-
   printsilent('Generating dotfile for: '),printsilent(CS),nls,
   (write_dot_graph_to_new_file(Status,Conjunct) -> true
     ; add_error(dot_output,'Writing dot to file failed: ',CS)).


:- dynamic dot_file_prefix/1.
:- dynamic dot_file_number/1.

dot_file_prefix('~/Desktop/dot').
set_dot_file_prefix(F) :- retractall(dot_file_prefix(_)), assert(dot_file_prefix(F)).
dot_file_number(0).
reset_dot_file_number :- retractall(dot_file_number(_)), assert(dot_file_number(0)).
get_next_nr(GNr) :- retract(dot_file_number(Nr)), N1 is Nr+1,
  assert(dot_file_number(N1)), GNr = Nr.
write_dot_graph_to_new_file(Status,BExpr) :-
    dot_file_prefix(Dir),get_next_nr(Nr),
    string_concatenate('_',Status,Str1),
    string_concatenate(Nr,Str1,NS),
    string_concatenate(Dir,NS,F1),
    atom_concat(F1,'.dot',FileName),
    user:write_dot_file_for_pred(BExpr,FileName).

% get dot file name if dot_output has been set
get_dot_file(Type,FileName) :- option(dot_analyse_output_prefix(_)),
    set_dot_file_prefix_if_option_set(Type),
    dot_file_prefix(Dir),
    string_concatenate('_',Type,Str1),
    string_concatenate(Dir,Str1,F1),
    atom_concat(F1,'.dot',FileName).

:- use_module(refinement_checker).
cli_csp_in_situ_refinement_check(P,Type,Q,NOW) :-
    debug_println(20,'% Starting CSP Refinement Check'),
    loaded_main_file(CSPFile),
    ajoin_with_sep(['assert',P,Type,Q], ' ',Assertion),
    start_xml_feature(csp_refinement_check,assertion,Assertion,FINFO),
    ( timeout_call(user:tcltk_check_csp_assertion(Assertion,CSPFile,'False',_PlTerm,RefTrace),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_refinement_check,FINFO).
cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW) :-
    start_xml_feature(refinement_check,file,RefFile,FINFO),
    refinement_checker:tcltk_load_refine_spec_file(RefFile),
    ( timeout_call(tcltk_refinement_search(RefTrace,PerformSingleFailures,RefNrNodes),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    stop_xml_feature(refinement_check,FINFO).
check_ref_result(RefTrace) :-
   ( RefTrace==no_counter_example ->
            print('==> Refinement Check Successful'),nl
        ; otherwise ->
            print('*** Refinement Check Counter-Example: ***'),nl, print(RefTrace),nl,
            print('*** Refinement Check Failed ***'),nl,
            error_occurred(refinement_check_fails)).
cli_checkAssertion(Proc,Model,AssertionType,_NOW) :-
    loaded_main_file(CSPFile),
    ajoin(['assert ',Proc,' :[ ',AssertionType,'[',Model,']',' ]'],Assertion),
    start_xml_feature(csp_deadlock_check,assertion,Assertion,FINFO),
    ( /*timeout_call(*/user:tcltk_check_csp_assertion(Assertion,CSPFile,'False',_PlTerm,ResTrace)/*,NOW)*/
      -> check_model_result(Assertion,ResTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_deadlock_check,FINFO).
cli_check_csp_assertion(Assertion,NOW) :-
    start_xml_feature(csp_assertion_check,assertion,Assertion,FINFO),
    loaded_main_file(CSPFile),
    ajoin(['assert ',Assertion],AssertionFull),
    ( timeout_call(user:tcltk_check_csp_assertion(AssertionFull,CSPFile,_Negated,PlTerm,ResTrace),NOW)
      -> check_model_result(PlTerm,ResTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_assertion_check,FINFO).



check_model_result(AssertionPlTerm,ResTrace) :-
  ( ResTrace==no_counter_example ->
    print('==> Model Check Successful'),nl
  ;otherwise ->
    (functor(AssertionPlTerm,assertRef,_Arity) ->
      print('*** Refinement Check Counter-Example: ***'),nl, print(ResTrace),nl,
      print('*** Refinement Check Failed ***'),nl,
      error_occurred(refinement_check_fails)
    ; otherwise ->
      print('*** Model Check Counter-Example: ***'),nl,print(ResTrace),nl,
      print('*** Model Check Failed ***'),nl,
      error_occurred(model_check_fails))
  ).
cli_csp_get_assertions :-
    loaded_main_file(CSPFile),
    get_csp_assertions_as_string(CSPFile,String),
    print('*** Assertions in File (separated by $) ***'),nl,print(String),nl.
cli_eval_csp_expression(E) :-
    (loaded_main_file(CSPFile) ->
	      haskell_csp: evaluate_csp_expression(E, CSPFile, Res)
	    ; haskell_csp: evaluate_csp_expression(E,Res)
    ), print('Evaluated Expression: '),nl,print(Res),nl.
cli_csp_add_unicode(File) :-
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--addUnicode=').
cli_csp_remove_unicode(File) :-
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--removeUnicode=').
cli_csp_translate_to_file(PlFile) :-
    loaded_main_file(CSPFile),parse_and_load_cspm_file_into_specific_pl_file(CSPFile,PlFile).
cli_check_scc_for_ltl_formula(LtlFormula,SCC) :-
    ltl_fairness: check_scc_ce(LtlFormula,SCC).
:- use_module(tools_files,[write_lines_to_file/2]).
cli_get_min_max_coverage(Filename) :-
    coverage_statistics:tcltk_compute_min_max(list(R)),
    tools:filter(tools: is_a_comment('/*','*/'),R,_Comments,Res),
    write_lines_to_file(Filename,Res).
%    pretty_print_min_max_coverage_to_file(FileName).
:- use_module(coverage_statistics,[pretty_print_coverage_information_to_file/1]).
cli_get_coverage_information(FileName) :-
    pretty_print_coverage_information_to_file(FileName).
cli_vacuity_check :-
    eclipse_interface:get_vacuous_invariants(L),
    (L=[] -> print('No vacuous invariants'),nl
     ; maplist(add_vacuous_invariant,L)).
add_vacuous_invariant(Inv) :-
    translate:translate_bexpression(Inv,TI),
    add_error(vacuity_check,'Vacuous invariant: ',TI).
cli_start_socketserver(Port) :-
    printsilent('Starting Socket Server'),nls,
    safe_absolute_file_name(prob_home('.'),AppDir),
    printsilent('Application Path: '),printsilent(AppDir),nls,
    disable_interaction_on_errors,
    ( start(Port) -> true
    ; otherwise ->
        print('Starting socket server failed, Port: '), print(Port),nl),
    printsilent('Finished Socket Server'),nls.
cli_check_statespace_hash(Expected,Kind) :-
    print('Computing hash of entire statespace: '),
    state_space:compute_full_state_space_hash(Hash),
    print(Hash),nl, % TO DO: maybe also compute hash for transitions and check that
    (Hash=Expected -> true
      ; Kind=='64bit', \+ platform_is_64_bit -> format('Hash does not match ~w (but was computed on 64-bit system)~n',[Expected])
      ; Kind=='32bit', platform_is_64_bit -> format('Hash does not match ~w (but was computed on 32-bit system)~n',[Expected])
      ; add_error(hash,'Expected Statespace Hash to be: ',Expected)).
cli_show_coverage(ShowEnabledInfo,NOW) :-
    cli_show_coverage(_Nodes,_Operations,ShowEnabledInfo,NOW).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo==just_check_stats,!, % no printing of individual transition coverage
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo== just_summary,
    !, % no printing of detailed transition coverage (avoid traversing state space)
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    format('Coverage:~n States: ~w~n Transitions: ~w~n',[TotalNodeSum,TotalTransSum]),
    show_initialisation_summary(NOW),
    show_operation_coverage_summary(NOW),
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    print('Coverage:'),nl,
    compute_the_coverage(Res,TotalNodeSum,TotalTransSum,ShowEnabledInfo,false),
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    print(Res),nl,
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum) :-
    ( Nodes=TotalNodeSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of nodes: ',TotalNodeSum),
        add_error(probcli,'Expected: ',Nodes),error_occurred(coverage)),
    ( Operations=TotalTransSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of transitions: ',TotalTransSum),
        add_error(probcli,'Expected: ',Operations),error_occurred(coverage)).
cli_show_variable_coverage(NOW) :-
    print('Variable (min/max) coverage: '),nl,
    coverage_statistics:tcltk_compute_min_max(list(R)),
    print_list(R),nl,
    writeln_log(computed_variable_coverage(NOW,R)).

cli_print_machine_statistics :-
    b_machine_name(Name),
    %(b_get_main_filename(File) -> true ; File=unknown),
    format('Machine statistics for ~w~n',[Name]),
    findall(Key/Nr,b_machine_statistics(Key,Nr),L),
    maplist(print_keynr,L),!.
cli_print_machine_statistics :- add_error(machine_stats,'Could not obtain machine statistics').
print_keynr(Key/Nr) :- format(' ~w : ~w~n',[Key,Nr]).

xml_log_machine_statistics :-
    animation_mode(Major),(animation_minor_mode(Minor) -> true ; Minor=none),
    write_xml_element_to_log(animation_mode,[major/Major,minor/Minor]),
    ((b_or_z_mode, b_machine_name(Main))
     ->  findall(Key/Nr,b_machine_statistics(Key,Nr),BMachStats),
         (b_get_main_filename(MainFile) -> get_tail_filename(MainFile,TailFile) ; TailFile = unknown),
         write_xml_element_to_log(b_machine_statistics,[machine_name/Main, tail_filename/TailFile|BMachStats])
       ; true).

cli_print_junit_results(ArgV) :-
    junit_mode(S),!,
    statistics(runtime,[E,_]),
    T is E - S,
    create_junit_result(ArgV,'Integration Tests','Integration_Test',T,pass,Result),
    print_junit([Result],'Integration_Test').
cli_print_junit_results(_).

cli_print_history(HistFile) :-
    findall( O, option(history_option(O)), Options),
    debug_println(9,writing_history_to_file(HistFile)),
    (Options=[trace_file] -> tcltk_save_history_as_trace_file(HistFile)  % save as Prolog trace file for replay with -t
     ; write_history_to_file(HistFile,Options) -> true
     ; add_error(history,'Writing history to file failed: ',HistFile)).

cli_print_values(ValuesFilename) :-
    (write_values_to_file(ValuesFilename) -> true ; add_error(sptxt,'Writing values to file failed: ',ValuesFilename)).
cli_print_all_values(ValuesDirname) :-
    (write_all_values_to_dir(ValuesDirname) -> true ; add_error(sstxt,'Writing all values to directory failed: ',ValuesDirname)).

:- use_module(library(terms),[term_hash/2]).
:- dynamic probcli_time_stamp/1.
generate_time_stamp(NOW,TS) :- retractall(probcli_time_stamp(_)),
  now(NOW),
  prolog_flag(argv,ArgV),term_hash(ArgV,Hash),
  Rnd is Hash mod 1000,
  % random(0,1000,Rnd),  always returns 216 % TO DO: try to get milliseconds from some library function
  TS is (NOW*1000)+Rnd,
  assert(probcli_time_stamp(TS)).
update_time_stamp(NOW1) :- retractall(probcli_time_stamp(_)),
  assert(probcli_time_stamp(NOW1)).

%get_errors :- \+ real_error_occurred,!, (get_error(_Source,_Msg) -> print('*** Warnings occurred'),nl ; true), reset_errors.
get_errors :-
   (get_preference(view_probcli_errors_using_bbresults,true)
     -> tools_commands:show_errors_with_bb_results([current]) ; true),
   get_error_sources.

get_error_sources :- get_error(ErrSource,Msg), !, error_occurred_with_msg(ErrSource,Msg),
      findall(1,get_error(ErrSource,_),L), length(L,Nr),
      (Nr>0 -> N1 is Nr+1, get_error_category_and_type(ErrSource,Cat,Type),
               (Type=error -> print_error('*** Occurences of this error: ')
                 ; print_error('*** Occurences of this warning: ')),
               print_error(N1),
               write_xml_element_to_log(multiple_errors_occurred,[category/Cat,type/Type,number/N1])
       ; true),
      get_error_sources.
get_error_sources.

:- use_module(state_space,[state_error/3, invariant_violated/1, time_out_for_invariant/1, time_out_for_assertions/1, time_out_for_node/3]).
get_state_space_errors :- option(strict_raise_error),
      !,
	  ((\+ option(no_invariant_violations),invariant_violated(ID))
		  -> print(invariant_violation(ID)),nl,
			  error_occurred(invariant_violation)
		 ; true),
	  (state_error(_,_,abort_error(TYPE,Msg,_,Span)) -> error_occurred(TYPE,error,Span,Msg) ; true),
	  get_state_errors(_).
get_state_space_errors.

get_state_errors(ID) :- state_error(ID,_,X), X\=invariant_violated, X\=abort_error(_,_,_,_),
  create_state_error_description(X,Msg),error_occurred(Msg),fail.
get_state_errors(ID) :- time_out_for_invariant(ID),error_occurred(time_out_for_invariant),fail.
get_state_errors(ID) :- time_out_for_assertions(ID),error_occurred(time_out_for_assertions),fail.
get_state_errors(ID) :- time_out_for_node(ID,_,time_out),error_occurred(time_out),fail.
get_state_errors(ID) :- time_out_for_node(ID,_,virtual_time_out(_)),error_occurred(virtual_time_out),fail.
get_state_errors(_).


create_state_error_description(eventerror(Event,Error,_),Description) :- !,
    functor(Error,Functor,_),
    ajoin(['event_error:',Event,':',Functor],Description).
create_state_error_description(StateError,Description) :-
    functor(StateError,Functor,_),
    atom_concat('state_error:',Functor,Description).

check_loaded(Action) :-
    ( file_loaded(true)  -> true
    ; file_loaded(error) -> fail /* we have already generated error message */
    ; otherwise ->
        add_error(probcli,'No file specified; cannot perform action: ',Action),
        error_occurred(loading),fail).

:- dynamic loaded_main_file/2.
loaded_main_file(File) :- loaded_main_file(_Ext,File).

load_main_file(MainFile,NOW,Already_FullyProcessed) :- retractall(loaded_main_file(_,_)),
               debug_print(20,'% Loading: '), debug_println(20,MainFile),
               writeln_log_time(loading(NOW,MainFile)),
               get_filename_extension(MainFile,Ext),
               debug_println(6,file_extension(Ext)),
               start_probcli_timer(Timer),
               load_spec_file(Ext,MainFile,Already_FullyProcessed),
               stop_probcli_debug_timer(Timer,'% Finished loading'),
               (Already_FullyProcessed==true -> true
                ; assert(loaded_main_file(Ext,MainFile))).
%load_spec_file('pl',MainFile) :- !, load_cspm_spec_from_pl_file(MainFile). % no longer needed ?
load_spec_file('csp',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('cspm',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('P',MainFile) :- !, load_xtl_file(MainFile).
load_spec_file('p',MainFile) :- !, load_xtl_file(MainFile). % sometimes windows is confused about the upper case letter....
load_spec_file('eventb',MainFile) :- !, load_eventb_file(MainFile).
load_spec_file('v',MainFile) :- !,
    print('---------------------------------'),nl,
    print('Loading: '),print(MainFile),nl,
    load_b_file(MainFile). % Siemens Rule File; maybe in future use -eval_rule_file
load_spec_file('prob',MainFile) :- !,load_prob_file(MainFile).
load_spec_file('mch',MainFile) :- !,load_b_file(MainFile).
load_spec_file('sys',MainFile) :- !,load_b_file(MainFile).
load_spec_file('ref',MainFile) :- !,load_b_file(MainFile).
load_spec_file('imp',MainFile) :- !,load_b_file(MainFile).
load_spec_file('rmch',MainFile) :- !,load_b_file(MainFile).
load_spec_file('fuzz',MainFile) :- !,tcltk_open_z_file(MainFile).
load_spec_file('tla',MainFile) :- !, load_tla_file(MainFile).
load_spec_file('eval',File) :- !, cli_set_empty_machine, cli_start_animation(eval_file), eval_string_or_file(file,File,exists,_).
load_spec_file('pb',File) :- !, cli_set_empty_machine, cli_start_animation(pb_file), eval_string_or_file(file,File,exists,_).
load_spec_file('pml',MainFile) :- !,parsercall:call_promela_parser(MainFile),
   parsercall:promela_prolog_filename(MainFile,PrologFile),
   printsilent(consulting(PrologFile)),nls,
   tcltk_open_promela_file(PrologFile).
load_spec_file(EXT,MainFile) :- print_error('### Unknown file extension, assuming B machine:'),
                                 print_error(EXT),
                                 load_b_file(MainFile).

load_spec_file('pl',MainFile, Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing PO file: '),printsilent(MainFile),nls,
   load_po_file(MainFile),
   (option(timeout(TO)) -> set_disprover_timeout(TO) ; reset_disprover_timeout),
   (option(disprover_options(L)) -> set_disprover_options(L) ; set_disprover_options([])),
   printsilent('Running ProB Disprover'), nls,
   run_disprover_on_all_pos(Summary),
   print_disprover_stats,
   accumulate_infos(disprover,[po_files-1|Summary]),
   get_errors,
   (option(cli_check_disprover_result(Infos)) -> check_required_infos(Infos,Summary,load_po_file)
    ; option(strict_raise_error) -> check_required_infos([false-0,unknown-0,failure-0],Summary,load_po_file)
    % TO DO: provide way for user to specify expected info
     ; true),
     clear_loaded_machines.
load_spec_file('smt2',MainFile,Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing SMT file: '),printsilent(MainFile),nls,
   smtlib2_file(MainFile).
load_spec_file(EXT,F,false) :- load_spec_file(EXT,F).


% do not perform -execute_all if no parameters provided
do_not_execute_automatically('pl').
do_not_execute_automatically('smt2').

test_kodkod_and_exit(MaxResiduePreds,NOW) :-
    start_animation_without_computing,
    test_kodkod(MaxResiduePreds),
    halt_prob(NOW).

compare_kodkod_performance1(KPFile,Iterations,NOW) :-
    start_animation_without_computing,
    compare_kodkod_performance(KPFile,Iterations),
    halt_prob(NOW).

:- use_module(parsercall,[check_java_version/2,get_parser_version/1, ensure_console_parser_launched/0]).
check_java_version :- check_java_version(V,Result),
    format('Result of checking Java version:~n  ~w~n',[V]),
    (Result=compatible -> check_parser_version
      ; add_error(check_java_version,V)).

check_parser_version :- get_parser_version(PV),!,
    format('  ProB B Java Parser available in version: ~w.~n ',[PV]). % will also launch parser
check_parser_version :- add_error(check_parser_version,'Cannot start Java B Parser to obtain version number').

:- use_module(version).
print_short_version :-
    version(V1,V2,V3,Suffix),revision(Rev),
    format('VERSION ~p.~p.~p-~p (~p)~N',[V1,V2,V3,Suffix,Rev]).

:- use_module(parsercall,[get_parser_version/1, get_java_command_path/1, get_java_fullversion/1]).
print_version :-
    print('ProB Command Line Interface'),nl,
    version(V1,V2,V3,Suffix),
    revision(Rev),  lastchangeddate(LCD), prolog_flag(version,PV),
    format('  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV]),
    ( option_verbose ->
        safe_absolute_file_name(prob_home('.'),AppDir),
        format('  Application Path: ~p~N', [AppDir]),
        (get_java_command_path(JavaPath)
          -> format('  Java Runtime: ~p~N', [JavaPath]),
             (get_java_fullversion(JavaVersion)
               ->  format('  Java Version: ~s~N', [JavaVersion])
               ;   format('  Java Version: *** not available ***~N',[])
             ),
             get_parser_version(ParserVersion),
             format('  Java Parser: ~p~N', [ParserVersion])
          ;  format('  Java Runtime: *** not available ***~N',[])
        ),
        (option(very_verbose) -> print_prolog_flags ; true)
     ;  true
    ), print_compile_time_flags.

print_prolog_flags :-
   findall(Flag/Val, (relevant_flag(Flag),current_prolog_flag(Flag,Val)), Flags),
   format('  PROLOG FLAGS: ~w~N',[Flags]).
relevant_flag(dialect).
relevant_flag(version_data).
relevant_flag(profiling).

print_compile_time_flags :-
   findall(Flag, compile_time_flag(Flag), Flags),
   (Flags=[] -> true ; format('  COMPILE TIME FLAGS: ~w~N',[Flags])).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
compile_time_flag(prob_safe_mode).
:- endif.
:- if(environ(prob_data_validation_mode,true)).
compile_time_flag(prob_data_validation_mode).
:- endif.
:- if(environ(partially_evaluate,true)).
compile_time_flag(partially_evaluate).
:- endif.
:- if(environ(partially_evaluate_compile,true)).
compile_time_flag(partially_evaluate_compile).
:- endif.
:- if(environ(no_wd_checking,true)).
compile_time_flag(no_wd_checking).
:- endif.
:- if(environ(use_pe_cache,true)).
compile_time_flag(use_pe_cache).
:- endif.
:- if(environ('SP_JIT',disabled)).
compile_time_flag('SP_JIT'(disabled)).
:- endif.
:- if(environ('SP_JIT',yes)).
compile_time_flag('SP_JIT'(yes)). % this is the default
:- endif.
:- if(environ('SP_JIT',no)).
compile_time_flag('SP_JIT'(no)).
:- endif.
:- if(environ('SP_JIT_COUNTER_LIMIT',_)).
compile_time_flag('SP_JIT_COUNTER_LIMIT_SET'). % default is 0
:- endif.
:- if(environ('SP_JIT_CLAUSE_LIMIT',_)).
compile_time_flag('SP_JIT_CLAUSE_LIMIT_SET'). % default is 1024
:- endif.
:- if(environ('SP_SPTI_PATH',verbose)).
compile_time_flag('SP_SPTI_PATH'(verbose)). % can be set to verbose
:- endif.

compile_time_flag(_) :- fail.

print_help :- print_version,
          print('Usage: probcli FILE [OPTIONS]'),nl,
          print('  OPTIONS are: '),nl,
          print('     -mc Nr       model check; checking at most Nr states'),nl,
          print('     -model_check model check without limit on states explored'),nl,
          print('     -noXXX       XXX=dead,inv,goal,ass (for model check)'),nl, % -nodead, -noinv, -nogoal, -noass
          print('     -bf          proceed breadth-first'),nl,
          print('     -df          proceed depth-first'),nl,
          print('     -mc_mode M   M=hash,heuristic,random,breadth-first,depth-first,mixed'),nl,
          print('     -timeout N   total timeout in ms for model and refinement checking'),nl, % also for disprover
          print('     -disable_timeout disable timeouts for operations, invariants,....'),nl, % speeds up mc
          print('     -t           trace check (associated .trace file must exist)'),nl,
          print('     -init        initialise specification'),nl,
          print('     -cbc OPNAME      constraint-based invariant checking for an operation'),nl,
          print('                      (you can also use OPNAME=all)'),nl,
          print('     -cbc_deadlock    constraint-based deadlock checking'),nl,
          print('                      (you can also use -cbc_deadlock_pred PRED)'),nl,
          print('     -cbc_assertions  constraint-based static assertion checking'),nl,
          print('     -cbc_refinement  constraint-based static refinement checking'),nl,
          print('     -cbc_sequence S  constraint-based search for sequence of operations'),nl,
          print('     -strict          raise error if model-checking finds counter example'),nl,
          print('                      or trace checking fails or any error state found'),nl,
          print('     -expcterr ERR    expect error to occur (ERR=cbc,mc,ltl,...)'),nl,
          print('     -animate Nr      random animation (max. Nr steps)'),nl,
          print('     -animate_all     random animation until a deadlock is reached'),nl,
          print('     -animate_stats   provide feedback which operations are executed'),nl,
          print('     -his File        write history to File'),nl,
          print('     -his_option O    additional option when writing a history (show_init or show_states)'),nl,
          print('     -sptxt File      save constants and variable values of last discovered state to File'),nl,
          print('     -sstxt Dir       save constants and variable values of all discovered states to files in Dir'),nl,
          print('     -cache Directory automatically save constants to files and avoid recomputation'),nl,
          print('     -det_check       check if animation steps are deterministic'),nl,
          print('     -det_constants   only check if SETUP_CONSTANTS step is deterministic'),nl,
          ( \+ option_verbose -> true ;
          print('     -i               interactive animation. Only for interactive sessions,'),nl,
          print('                      the output can arbitrarily change in future versions. '),nl,
          print('                      Do not build automatic tools using the interactive mode'),nl
          ),
          print('     -repl            start interactive read-eval-loop'),nl,
          print('     -eval "E"        evaluate expression or predicate'),nl,
          print('     -eval_file FILE  evaluate expression or predicate from file'),nl,
          print('     -c               print coverage statistics'),nl,
          print('     -cc Nr Nr        print and check coverage statistics'),nl,
          print('     -vacuity_check   look for vacuous implications in invariant'),nl,
          print('     -cbc_redundant_invariants Nr  cbc search for redundant invariants'),nl,
          print('     -statistics      print memory and other statistics at the end'),nl,
          print('     -p PREF Val  set Preference to Value'),nl,
          print('     -prefs FILE  set Preferences from Prolog File'),nl,
          print('     -card GS Val set scope of B deferred set'),nl,
          print('     -goal "PRED" set GOAL predicate for model checker'),nl,
          print('     -check_goal  check GOAL (after -mc, -t, or -animate)'),nl,
          print('     -scope "PRED"set scope predicate for model checker'),nl,
          print('     -s Port      start socket server on given port'),nl,
          print('     -ss          start socket server on port 9000'),nl,
          print('     -sf          start socket server on some free port'),nl,
          print('     -l LogFile   log activities in LogFile'),nl,
          print('     -ll          log activities in /tmp/prob_cli_debug.log'),nl,
          print('     -lg LogFile  analyse logfile using gnuplot'),nl,
          print('     -pp FILE     pretty-print internal representation to file'), nl,
          print('     -ppf FILE    like -pp, but force printing of all type infos'),nl,
          print('     -ppB FILE    pretty-print internal representation to file in valid B syntax'),nl,
          print('     -v           verbose'),nl,
          print('     -mc_with_tlc        model check using the model checker TLC'),nl,
          print('     -mc_with_lts_sym    model check using the model checker LTSmin (symbolic)'),nl,
          print('     -mc_with_lts_seq    model check using the model checker LTSmin (sequential)'),nl,
 %%         print('     -vv          very verbose'),nl,  % no effect when debugging calls disabled
          print('     -version            print version information (-svers for short info)'),nl,
          print('     -check_java_version check that Java version compatible with ProB parser'),nl,
          print('     -assertions         check ASSERTIONS'),nl,
          print('     -main_assertions    check ASSERTIONS from main file only'),nl,
          print('     -properties         check PROPERTIES'),nl,
          print('     -cache Dir          use directory "Dir" to cache constants and variables'),nl,
          print('     -ltlfile F          check LTL formulas in file F'),nl,
          print('     -ltlassertions      check LTL assertions (in DEFINITIONS)'),nl,
          print('     -ltllimit L         explore at most L states when model-checking LTL or CTL'),nl,
          print('     -ltlformula \"F\"   check the LTL formula F'),nl,
          print('     -ctlformula \"F\"   check the CTL formula F'),nl,
          print('     -save File          save state space for later refinement check'),nl,
          print('     -refchk File        refinement check against previous saved state space'),nl,
          print('     -mcm_tests Depth MaxStates EndPredicate File'),nl,
          print('                  generate test cases with maximum length Depth, explore'),nl,
          print('                  maximally MaxStates, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -mcm_cover Operation'),nl,
          print('                  when generating MCM test cases, Operation should be covered'),nl,
          print('     -cbc_tests Depth EndPredicate File'),nl,
          print('                  generate test cases by constraint solving with maximum'),nl,
          print('                  length Depth, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -cbc_cover Operation'),nl,
          print('                  when generating CBC test cases, Operation should be covered'),nl,
%          print('     -cbc_cover_all   try and cover all operations'),nl, % is now default if no cbc_cover provided
          print('     -test_description File'),nl,
          print('                      read information for test generation from File'),nl,
          print('     -dot CMD File         write a graph to a dot file, with CMD being one of:'),nl,
          (is_dot_command(Cmd),command_description(Cmd,_,Desc),
           format('          ~w : ~w~n',[Cmd,Desc]),fail
            ; true),
          print('     -dotexpr CMD Expr File write a graph for Expr to a dot file, with CMD:'),nl,
          (is_dot_command_for_expr(Cmd),command_description(Cmd,_,Desc),
           format('          ~w : ~w~n',[Cmd,Desc]),fail
            ; true),
          print('     -dot_output      Path generate dot files for false assertions/properties'),nl,
          print('     -dot_all         also generate dot files for true assertions/properties'),nl,
          print('     -csvhist E File  evaluate expression over history and generate CSV file.'),nl,
          print('     -load_state File load state from a saved state (generated by PrB Tcl/Tk or -save_state)'),nl,
       % For Eclipse Version only
       %%  print('     -parsercp  CP class path of the B Parser, this has to be a valid Java class path'),nl,
       %%   print('     -cspm         load CSP-M .csp file rather than B Machine .mch/.ref/.imp File'),nl,
       %%   print('     -csp         load CSP-M .pl file rather than B Machine File'),nl,

          /* Options -cspref, -cspdeadlock, -cspdeterministic, and -csplivelock are deprecated, should be excluded in favor of -csp_assertion */
          print('     -cspref Spec [m= Impl File'),nl,
          print('                  checks a refinement statement,'),nl,
          print('                  where Spec and Impl are processes from File, and \'m\' the type of the refinement:'),nl,
          print('                  \'T\' for traces, \'F\' for failures, or \'FD\' for failures-divergences.'),nl,
          print('     -cspdeadlock P m File'),nl,
          print('                  checks a process for deadlock,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -cspdeterministic P m File'),nl,
          print('                  checks a process for determinism,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -csplivelock P File'),nl,
          print('                  checks a process for divergence,'),nl,
          print('                  where \'P\' is a process from File.'),nl,
          /* Options -cspref, -cspdeadlock, -cspdeterministic, and -csplivelock are deprecated, should be excluded in favor of -csp_assertion */

          print('     -csp_assertion \"A\" File'),nl,
          print('                  checks the CSP assertion \'A\' on file \'File\''),nl,
	        print('     -csp_eval "E" evaluate CSP-M expression.'),nl,
          print('     -csp-guide File   CSP||B: Use the CSP File to control the B machine'),nl,
          print('     '),nl,
          ( \+ option_verbose -> true
           ;
	      print('     -test_mode       set random seed to the Prolog\'s current random state'),nl,
          print('     -expcteval Evals expect that the given predicates have been evaluated or not (see external predicate IS_EVALUATED)'),nl,
          print('     -rc          runtime checking of types/pre-/post-conditions'),nl,
          print('     -state_trace File      read a file of B predicates (one per line) and try find a matching trace.'),nl

          ),
          print(' FILE extensions are: '),nl,
          print('      .mch         for B abstract machines'),nl,
          print('      .ref         for B refinement machines'),nl,
          print('      .imp         for B implementation machines'),nl,
          print('      .csp, .cspm  for CSP-M files, same format as FDR'),nl,
          print('      .eventb      for Event-B packages exported from Rodin ProB Plugin'),nl,
          print('      .tla         for TLA+ models'),nl,
          ( option_verbose ->
              print(' Preferences PREF are: '),nl,
              print_eclipse_prefs
          ; otherwise ->
              print(' Use --help -v to print avalaible preferences PREF'),nl
          ),
          print(' More info at: http://www.stups.hhu.de/ProB/'),nl,
          print('       and at: http://www.prob2.de/probcli/'),nl,
          % http://www.stups.hhu.de/ProB/index.php5/Using_the_Command-Line_Version_of_ProB
          nl.


set_argv(V) :-
   debug_println(20,set_argv(V)),
   external_functions:set_argv_from_atom(V).

:- use_module(state_space_exploration_modes,[set_depth_breadth_first_mode/1]).
set_prefs :-
   if_option_set(socket(_), % then we may need the event(.) transition_info for the Java API
       preferences:set_preference(store_event_transinfo,true)),
   option(set_prefs_from_file(File)),
   debug_println(20,load_preferences(File)),
   preferences:load_preferences(File),
   fail.
set_prefs :-
   option(set_preference_group(P,V)),
   debug_println(20,set_preference_group(P,V)),
   set_preference_group(P,V),
   fail.
% eclipse preference or 'normal preference'
set_prefs :-
   option(set_pref(P,V)),
   set_pref(P,V),
   fail.
set_prefs :- option(set_card(P,V)),
   debug_println(20,set_card(P,V)),
   set_eclipse_card(P,V),
   fail.
set_prefs :-
    ( option(breadth_first) -> set_depth_breadth_first_mode(breadth_first)
    ; option(depth_first)   -> set_depth_breadth_first_mode(depth_first)
    ; option(depth_breadth_first_mode(M)) -> set_depth_breadth_first_mode(M)
    ; otherwise -> true
    ).
set_pref(P,V) :-
   debug_println(20,set_pref(P,V)),
   ( eclipse_preference(P,_)
     -> set_eclipse_preference(P,V)
    ; obsolete_eclipse_preference(P) -> add_warning(probcli,'Obsolete preference: ',P)
     ; % might be a term if its a plugin preference
     atom_codes(P,Codes),
     append(Codes,".",Codes2), % to make term readable by read_from_codes
     read_from_codes(Codes2,Preference),
     ((nonvar(Preference),preference_val_type(Preference,_))
        -> prob2_interface: convert_cli_arg(V,Value),
           set_preference(Preference,Value)
        ;  add_error(probcli,'Illegal preference: ',P)
     )
   ).

set_optional_errors :- % register optional/expected errors in the error_manager; avoid printing on stderr
   reset_optional_errors,
   (option(optional_error(Type)) ; option(expect_error(Type)) ; option(expect_error_pos(Type,_Line,_Col))),
   register_optional_error(Type),
   fail.
set_optional_errors.

model_check(Nr,Time,ErrRes,NOW) :-
   option(timeout(TO)),!,
   safe_time_out(model_check(Nr,Time,ErrRes),TO,Res),
   (Res=time_out -> print('*** Timeout occurred while Model Checking: '), print(TO),nl,nl,
                    writeln_log(timeout_occurred(NOW,model_check(Nr,Time,ErrRes))),
                    Time=time_out(TO), ErrRes=time_out
                 ; true).
model_check(Nr,Time,ErrRes,_NOW) :- model_check(Nr,Time,ErrRes).

:- use_module(model_checker,[model_checking_is_incomplete/6]).

add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations) :-
   %print(check(model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term))),nl,
   model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term),
   nl,add_warning(model_check_incomplete,Msg,Term),
   fail.
add_model_checking_warnings(_,_,_,_).

model_check(Nr,Time,ErrRes) :-
   statistics(runtime,[T1,_]),
   statistics(walltime,[W1,_]),
   (option(no_deadlocks) -> FindDeadlocks=0 ; FindDeadlocks=1),
   (option(no_invariant_violations) -> FindInvViolations=0 ; FindInvViolations=1),
   (option(no_goal) -> FindGoal=0 ; FindGoal=1),
   (option(no_assertion_violations)
     -> FindAssViolations=0
      ; FindAssViolations=1
    ),
   get_preference(por,POR),
   get_preference(pge,PGE),
   StopAtFullCoverage=0,
   %STOPMCAFTER =  86400000, /* 86400000 = 1 day timeout */
   STOPMCAFTER = 1152921504606846975, /* equals 13,343,998,895 days */
   InspectExistingNodes = 1,
   write_xml_element_to_log(model_checking_options,[find_deadlocks/FindDeadlocks,
        find_invariant_violations/FindInvViolations, find_goal/FindGoal,
        find_assertion_violations/FindAssViolations,
        partial_order_reduction/POR,
        partial_guard_evaluation/PGE,
        inspect_existing_nodes/InspectExistingNodes]),
   (user:do_model_check(Nr,NodesAnalysed,STOPMCAFTER,ErrRes,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,POR,PGE, InspectExistingNodes)
      -> (statistics(runtime,[T2,_]), statistics(walltime,[W2,_]),
           Time1 is T2-T1, WTime is W2-W1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           format('Model Checking Time: ~w ms (~w ms walltime)~n',[Time,WTime]),
           format('States analysed: ~w~n',[NodesAnalysed]),
           state_space: get_state_space_stats(_,NrTransitions,_),
           print('Transitions fired: '),print(NrTransitions),nl,
           write_xml_element_to_log(model_checking_statistics,
                   [result/ErrRes,runtime/Time,walltime/WTime,states/NodesAnalysed,transitions/NrTransitions]),
          (ErrRes = no
           -> print('No Counter Example Found, not all nodes visited')
           ;  (ErrRes=all
                -> (tcltk_find_max_reached_node
                     -> print('No Counter Example found. However, not all transitions were computed !')
                     ;  print('No Counter Example found. ALL nodes visited.')
                   ),
                   add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations)
                ;   print('*** COUNTER EXAMPLE FOUND ***'),nl,
                    debug_println(20,ErrRes),nl,
                    user:translate_error_for_tclk(ErrRes,TclTkRes),
                    print(TclTkRes),nl,
                    (option(silent) -> true
                     ; option(no_counter_examples) -> true
                      ; print('*** TRACE: '),nl,
                        user:tcltk_get_history(list(Hist)),
                        reverse(Hist,Trace),
                        print_list(Trace,1)
                    ),
                    error_occurred(TclTkRes)
              )
          )
         )
      ;  ( statistics(runtime,[T2,_]),
           Time1 is T2-T1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           print('Model Checking Time: '), print(Time), print(' ms'),nl,
           print_error('*** Model Checking FAILED '),nl,
           ErrRes=fail,
           definite_error_occurred)
   ).
cbc_check(_NOW) :-
     option(cbc_deadlock_check(DeadlockGoalPred)),
     print('% Starting Constraint-Based DEADLOCK check '),nl,
     start_xml_feature(cbc_deadlock_check,FINFO),
     (tcltk_constraint_find_deadlock_state_with_goal(DeadlockGoalPred,false,Res)
       -> write_result_to_file(Res),
         (Res=time_out ->
            print('*** TIME_OUT occurred ***'),nl,
            error_occurred(cbc_deadlock_check_time_out)
          ; print('*** DEADLOCK state found ***'),nl,
            error_occurred(cbc_deadlock_check),
            print('*** STATE = '),nl,
            current_b_expression(DBState), translate:print_bstate(DBState),nl,
            print('*** END DEADLOCKING STATE '),nl
         )
       ;  write_result_to_file(no_deadlock_found), Res=no_deadlock_found,
          print('No DEADLOCK state found'),nl
     ),
     write_xml_element_to_log(cbc_check_result,[result/Res]),
     stop_xml_feature(cbc_deadlock_check,FINFO),
     fail.
cbc_check(_NOW) :-
     option(constraint_based_check(OpName)),OpName\=all,
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for Operation: '), print(OpName),nl,
     start_xml_feature(cbc_operation_check,operation,OpName,FINFO),
	  (tcltk_constraint_based_check_with_timeout(OpName,Result)
		->  print('% Constraint-Based Check Result: '),nl,
			print(Result),nl,
			write_result_to_file(Result),
			(Result=time_out
			  ->  print('*** TIMEOUT OCCURRED ***'),nl, error_occurred(cbc)
			  ;  (Result=ok -> print('NO ERRORS FOUND'),nl)
				   ; print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl,error_occurred(cbc) )
		;  write_result_to_file(constraint_based_check_failed), Result=internal_error,
		   add_error(probcli,'ConstraintBasedCheck unexpectedly failed'),
		   (b_get_machine_operation(OpName,_,_,_) -> true
		     ; add_error(probcli,'Unknown Operation Name: ',OpName)),
		   definite_error_occurred
	  ),nl,
      write_xml_element_to_log(cbc_check_result,[result/Result]),
      stop_xml_feature(cbc_operation_check,FINFO),
	  fail.
cbc_check(NOW) :-
     option(constraint_based_check(all)),
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for all Operations: '),nl,
     start_xml_feature(cbc_operation_check,all_operations,true,FINFO),
          (tcltk_constraint_based_check(list(Result),ErrorsWereFound)
            ->  print('% Constraint-Based Check Result: '),nl,
                print(Result),nl,
                write_result_to_file(Result),
                (ErrorsWereFound=true
                  ->  print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl, error_occurred(cbc)
                  ;  (ErrorsWereFound=false -> print('NO ERRORS FOUND'),nl)
                       ; print('*** TIMEOUT OCCURRED ***'),nl,error_occurred(cbc)
                )
            ;  write_result_to_file(cbc_check_failed), Result=internal_error, ErrorsWereFound=false,
               add_internal_error('ConstraintBasedCheck unexpectedly failed. ',cbc_check(NOW)),definite_error_occurred
          ),nl,
      write_xml_element_to_log(cbc_check_result,[result/Result,errors_were_found/ErrorsWereFound]),
      stop_xml_feature(cbc_operation_check,FINFO),
      fail.
cbc_check(_NOW) :-
	 option(cbc_assertions(AllowEnumWarning,Options)),
	 findall(OPT,option(cbc_option(OPT)),FullOptions,Options),
     check_loaded(cbc_assertions),
     print('% Starting Constraint-Based static ASSERTIONS check '),nl,
     start_xml_feature(cbc_assertion_check,allow_enumeration_warning,AllowEnumWarning,FINFO),
     write_prolog_term_as_xml_to_log(options(Options)),
     (cbc_constraint_find_static_assertion_violation(Res,FullOptions)
       -> process_cbc_assertion_result(Res,AllowEnumWarning)
       ;  write_result_to_file(cbc_assertions_failed), Res=internal_error,
          print('CBC Check failed'),nl,
          error_occurred(cbc_assertions_failure)
     ),
      write_xml_element_to_log(cbc_check_result,[result/Res]),
      stop_xml_feature(cbc_assertion_check,FINFO),
      fail.
%cbc_check(NOW) :-
%     option(cbc_pred(TargetPredString)),
%     check_loaded(cbc_sequence),
%     print('% Starting Constraint-Based Check for Predicate: '), print(TargetPredString),nl,
%     b_set_up_valid_state_with_pred(...)  TO DO: add this feature
cbc_check(_NOW) :-
     option(cbc_sequence(Sequence,TargetPredString,Findall)),
     check_loaded(cbc_sequence),
     print('% Starting Constraint-Based Check for Sequence: '), print(Sequence),
     start_xml_feature(cbc_sequence_check,sequence,Sequence,FINFO),
     (TargetPredString='' -> true ; print(' with target: '), print(TargetPredString)),
     nl,
     write_xml_element_to_log(options,[target_predicate/TargetPredString]),
     (tcltk_cbc_find_trace(Sequence,TargetPredString,Findall,Res)
       -> (Res=ok -> print('Sequence found and executed'),nl
           ; Res=time_out -> error_occurred(cbc_sequence_time_out)
           ; Res=no_solution_found -> print('*** NO SOLUTION FOUND '),error_occurred(cbc_sequence_no_solution_found)
           ; Res=nr_cbc_sols(NrSols) -> print('*** # SOLUTIONS FOUND: '),print(NrSols),nl
           ; print('*** Unknown result: '), print(Res),nl,
             error_occurred(cbc_sequence)
          )
       ;  print('*** Internal error: Check failed '), error_occurred(cbc_sequence), Res=internal_error
     ),
      write_xml_element_to_log(cbc_check_result,[result/Res]),
      stop_xml_feature(cbc_sequence_check,FINFO),
      fail.
cbc_check(_NOW) :-
     option(cbc_refinement),
     check_loaded(cbc_refinement),
     print('% Starting Constraint-Based static refinement check '),nl,
     start_xml_feature(cbc_refinement_check,FINFO),
     tcltk_cbc_refinement_check(list(Result),ErrorsWereFound),
     print('% Constraint-Based Refinement Check Result: '),nl,print(Result),nl,
     (ErrorsWereFound = time_out -> print('*** TIME_OUT occurred ***'),nl,error_occurred(cbc_refinement_time_out) ;
      ErrorsWereFound = true -> print('*** Refinement Violation found ***'),nl,error_occurred(cbc_refinement) ;
      otherwise -> print('No static Refinement Violation found'),nl
     ),
      write_xml_element_to_log(cbc_check_result,[errors_were_found/ErrorsWereFound]),
      stop_xml_feature(cbc_refinement_check,FINFO),
      fail.
cbc_check(_NOW) :-
     option(cbc_redundant_invariants(NrExpected)),
     check_loaded(cbc_redundant_invariants),
     print('% Starting Constraint-Based invariant redundancy check'),nl,
     start_xml_feature(cbc_redundant_invariants,FINFO),
     cbc_find_redundant_invariants(Res,TimeoutOccured),
     length(Res,NrInvs),
     (Res = [] -> print('No redundant invariants found'),nl
      ; format('*** REDUNDANT INVARIANTS (~w) ***~n',[NrInvs]),
        prnt(1,Res), nl
      ),
      (NrExpected = NrInvs -> true ;  error_occurred(cbc_redundant_invariants)),
      write_xml_element_to_log(cbc_redundant_invariants,[redundant_invariants/NrInvs, timeout_occured/TimeoutOccured]),
      stop_xml_feature(cbc_redundant_invariants,FINFO),
      fail.
cbc_check(_).

prnt(_,[]).
prnt(N,[H|T]) :- format(' ~w : ~w~n',[N,H]), N1 is N+1, prnt(N1,T).

:- use_module(solver_interface,[predicate_uses_unfixed_deferred_set/2, unfixed_typed_id_in_list/3]).
:- use_module(bsyntaxtree,[get_texpr_id/2]).
process_cbc_assertion_result(time_out,_) :- !,
        write_result_to_file(no_counterexample_found('"TIME_OUT"')),
	    print('*** TIME_OUT occurred ***'),nl,
	    error_occurred(cbc_assertions_time_out).
process_cbc_assertion_result(no_counterexample_exists(Constants,TotPredicate,OtherInfo),AllowEnumWarning) :- !,
		print('No counter-example to ASSERTION exists '),(OtherInfo=[] -> true ; print(OtherInfo)),nl,
		(unfixed_typed_id_in_list(TID,CType,Constants) % TO DO: look only at component
		  -> write_deferred_set_used(AllowEnumWarning),
			 get_texpr_id(TID,CID),pretty_type(CType,CTypeS),
		     format('Warning: Some constants use deferred sets (e.g., ~w:~w) which have only been checked for a single cardinality!~n',[CID,CTypeS])
		  ; predicate_uses_unfixed_deferred_set(TotPredicate,CType)
		    -> write_deferred_set_used(AllowEnumWarning),pretty_type(CType,CTypeS),
		       format('Warning: Some quantified variables use deferred sets (e.g., ~w) which have only been checked for a single cardinality!~n',[CTypeS]) % happens for tests 1173, 1174
		  ; write_result_to_file(no_counterexample_exists)
		    %,print('Computing unsat core: '),nl,unsat_cores:unsat_core(TotPredicate,Core),print('CORE: '),translate:print_bexpr(Core),nl
		). % WE HAVE A PROOF
process_cbc_assertion_result(no_counterexample_found,AllowEnumWarning) :- !,
		write_result_to_file(no_counterexample_found('"Enumeration Warning"')),
		print('No counter-example for ASSERTION found (*enumeration warning occured*)'),nl,
		(AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).
process_cbc_assertion_result(counterexample_found,_) :- !,
		write_result_to_file(counterexample_found),
        print('*** Counter-example for ASSERTION found ***'),nl,
		error_occurred(cbc_assertions),
		print('*** STATE = '),nl,
		current_b_expression(DBState), translate:print_bstate(DBState),nl,
		print('*** END ASSERTION counter-example STATE '),nl,
		(get_dot_file('cbc_assertions',DFile) -> generate_dot_from_assertions(DFile) ; true).
process_cbc_assertion_result(Res,A) :-
		write_result_to_file(Res),
		add_internal_error('Unknown: ',process_cbc_assertion_result(Res,A)).


write_deferred_set_used(AllowEnumWarning) :-
             write_result_to_file(no_counterexample_found('"Deferred Sets Used"')),
			 (AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).

:- use_module(tools_io,[safe_open_file/4]).
write_result_to_file(Result) :- option(cbc_result_file(FILE)),
   safe_open_file(FILE,write,Stream,[]),
   !,
   write(Stream,Result),
   close(Stream).
write_result_to_file(_).



if_option_set(Option,Call) :-
    if_option_set(Option,Call,true).
if_option_set(Option,Then,Else) :-
    ( option(Option) -> call(Then)
    ; otherwise      -> call(Else)).
ifm_option_set(Option,Call) :-
    ifm_option_set(Option,Call,true).
ifm_option_set(Option,Then,Else) :- % can perform multiple options
   findall(Then,option(Option),As),
   (As=[] -> call(Else) ; perform(As)).
perform([]).
perform([A|T]) :- call(A), perform(T).
if_option_set_loaded(Option,Action,Call) :-
    ( (option(Option),check_loaded(Action)) ->
        call(Call)
    ; otherwise -> true).
ifm_option_set_loaded(Option,Action,Call) :- % can perform multiple options
   findall(Call,(option(Option),check_loaded(Action)),As), perform(As).



if_options_set(Option,Call) :- % allow multiple solutions for Option
   option(Option),call(Call),fail.
if_options_set(_,_).

print_options :- print('CLI OPTIONS: '),nl,
                 option(Option), print(Option), nl, fail.
print_options :- nl.

:- use_module(enabling_analysis,[infeasible_operation_cache/1]).
:- use_module(sap,[explore_and_generate_testcases/7,cbc_gen_test_cases_from_string/5, tcl_get_stored_test_cases/1]).

mcm_test_case_generation(ADepth,AMaxStates,ATarget,Output) :-
    arg_is_number(ADepth,MaxDepth),
    arg_is_number(AMaxStates,MaxStates),
    bmachine:b_parse_machine_predicate(ATarget,Target),!,
    get_comma_or_space_separated_options(mcm_cover,Events),
    print('mcm test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('mcm test case generation, maximum number of states: '),print(MaxStates),nl,
    print('mcm test case generation, target state predicate: '),print_bexpr(Target),nl,
    print('mcm test case generation, output file: '),print(Output),nl,
    print('mcm test case generation, events to cover: '),print_list(Events),nl,
    explore_and_generate_testcases(Events,Target,MaxDepth,MaxStates,Output,NumTests,Uncovered),
    print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    print_uncovered('mcm test case generation, ',Uncovered).
mcm_test_case_generation(_ADepth,_AMaxStates,_ATarget,_Output) :-
    print_error('### MCM Test Case Generation failed'),
    error_occurred(mcm_tests).

cbc_test_case_generation(ADepth,TargetString,Output) :-
    arg_is_number(ADepth,MaxDepth),
    ( option(cbc_cover_all) -> Events=all
       ; (get_comma_or_space_separated_options(cbc_cover,Events), Events \= []) -> true
       ; Events=all ),
    (\+ option(cbc_cover_final) -> FEvents = Events
     ; Events=all -> FEvents=all,
         add_error(cbc_cover_final,'Option cbc_cover_final not compatible with trying to cover all events')
     ; FEvents = final(Events),
       print('constraint based test case generation, target events considered final'),nl),
    print('constraint based test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('constraint based test case generation, target state predicate: '),print(TargetString),nl,
    print('constraint based test case generation, output file: '),print(Output),nl,
    (TargetString = '#not_invariant' ->
       print('constraint based test case generation, performing bounded model checking'),nl
    ;  print('constraint based test case generation, events to cover: '),print_list(Events),nl),
    cbc_gen_test_cases_from_string(FEvents,TargetString,MaxDepth,Output,Uncovered),
    !,
    %print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    (TargetString = '#not_invariant'
      -> tcl_get_stored_test_cases(list(Tests)), %print(tests(Tests)),nl,
        (Tests=[] -> print('No counterexample found'),nl
         ; Tests = [_|_] -> add_error(invariant_violation,'Invariant violation found by bmc')
         ; add_internal_error('Unexpected bmc result: ',Tests))
      ; (Uncovered=[_|_],option(strict_raise_error))
       -> add_error(cbc_tests,'Uncovered events: ',Uncovered)
      ; print_uncovered('constraint based test case generation, ',Uncovered)
    ).
cbc_test_case_generation(_ADepth,_ATarget,_Output) :-
    print_error('### CBC Test Case Generation failed'),
    error_occurred(cbc_tests).

print_uncovered(Msg,Uncovered) :-
   include(infeasible_operation_cache,Uncovered,Infeasible),
    (Infeasible=[]
      -> format('~wuncovered events: ',[Msg]),print_list(Uncovered),nl
      ;  format('~winfeasible uncovered events: ',[Msg]),print_list(Infeasible),nl,
         exclude(infeasible_operation_cache,Uncovered,Feasible),
         format('~wuncovered events: ',[Msg]),print_list(Feasible),nl
    ).

print_list(all) :- print('** all **').
print_list([]) :- print('** none **').
print_list([H|T]) :- length([H|T],Len), format('(~w) ',[Len]),
    print(H),print(' '),print_list2(T).
print_list2([]).
print_list2([H|T]) :- print(H),print(' '),print_list2(T).

get_comma_or_space_separated_options(Option,Selection) :-
    functor(O,Option,1),
    findall(E, (option(O),arg(1,O,CommaSep),
                   split_by_seperator(CommaSep,Es),
                   member(E,Es)),
            Selection).

split_by_seperator(NonAtomic,Res) :- \+ atomic(NonAtomic),!, Res=[NonAtomic].
split_by_seperator(String,Strings) :-
    atom_chars(String,Chars),
    split_by_seperator2(Chars,Strings).
split_by_seperator2(Chars,Result) :-
    append(AChars,[X|B],Chars),seperator(X),!,
    (AChars=[] -> Result=Rest ; atom_chars(A,AChars), Result=[A|Rest]),
    split_by_seperator2(B,Rest).
split_by_seperator2(Chars,[String]) :- atom_chars(String,Chars).

seperator(',').
seperator(' ').
seperator(';').

ltl_check_assertions :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print('LTL Model checking assertions'),nl,
    ltl_check_assertions(Limit,Outcome),
    ( Outcome = pass -> print('LTL check passed'),nl
    ; Outcome = fail -> print('*** LTL check failed'),nl,error_occurred(ltl)
    ; Outcome = no_tests -> print('*** No LTL assertions found, test failed'),nl,definite_error_occurred
    ; otherwise -> print('*** An error occurred in the LTL assertion test'),nl,
	definite_error_occurred).

:- use_module(probltlsrc(ltl),[parse_ltlfile/2]).
ltl_check_file(Filename) :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print('LTL Model checking assertions'),nl,
    parse_ltlfile(Filename, Formulas),
    ( real_error_occurred ->
        print('An error occurred while parsing the LTL file.\n'), fail
    ; otherwise ->
        ltl_check_formulas(Formulas,Limit)).

:- use_module(probltlsrc(ltl),[ltl_model_check2/4]).
ltl_check_formulas([],_) :-
    print('All LTL formulas checked.\n').
ltl_check_formulas([formula(Name,F)|Rest],Limit) :-
    print('Checking formula '),print(Name),print(':\n'),
    ltl_model_check2(F,Limit,init,Status),
    ( Status == no ->
        print('Counter-example found for formula \"'),print(Name),
        print('\", saving trace file.\n'),
        ajoin(['ltlce_', Name, '.trace'], Tracefile),
        tcltk_save_history_as_trace_file(Tracefile),
        add_error(ltl_counterexample,'Counter-example was found')
    ; Status == ok ->
        ltl_check_formulas(Rest,Limit)
    ; Status == incomplete ->
        ajoin(['Model was not completly model-checked, aborted after ',Limit,' new states'],
              Msg),
        add_error(ltl,Msg)
    ; otherwise ->
        ajoin(['Model checker returns unexpected result (',Status,')'],Msg),
        add_error(ltl,Msg)).

cli_ltl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    start_xml_feature(ltl_model_check,formula,Formula,FINFO),
    ltl_model_check(Formula,Max,init,Status),
    write_xml_element_to_log(model_check_result,[status/Status,expected_status/ExpectedStatus]),
    check_status(Status,ExpectedStatus,Formula,ltl),
    stop_xml_feature(ltl_model_check,FINFO).

cli_ctl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    start_xml_feature(ctl_model_check,formula,Formula,FINFO),
    ctl_model_check(Formula,Max,init,Status),
    write_xml_element_to_log(model_check_result,[status/Status,expected_status/ExpectedStatus]),
    check_status(Status,ExpectedStatus,Formula,ctl),
    stop_xml_feature(ctl_model_check,FINFO).

check_expected(St,Exp,Mode) :-
   (St=Exp -> true ; add_error(Mode,'Unexpected model checking result: ',Mode:St)).

check_status(ok,Expected,Formula,ltl) :- !, % TO DO: make uniform ? CTL returns true; LTL returns ok
    format('LTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    check_expected(true,Expected,ltl).
check_status(true,Expected,Formula,ctl) :- !,
    format('CTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    check_expected(true,Expected,ctl).
check_status(incomplete,Expected,Formula,LTLorCTL) :- !,
    incomplete_warning(LTLorCTL,Warning),
    add_warning(Warning, 'Warning: Model Check incomplete for: ', Formula),nl,
    format('No counter example found so far for ~w.~n',[Formula]),
    check_expected(incomplete,Expected,LTLorCTL).
check_status(NO,Expected,Formula,LTLorCTL) :- (NO=no ; NO=false),!, % TO DO: make uniform
    (Expected==false -> format('Model Check Counter-Example found for: ~w~n',[Formula])
      ; add_error(LTLorCTL, 'Model Check Counter-Example found for: ', Formula),nl
    ),
    print('Formula '), print('FALSE.'),nl,
        %print('Counter example found for '),print(Formula),print('.'),nl,
    nl,
    check_expected(false,Expected,LTLorCTL).
check_status(Status,Expected,Formula,LTLorCTL) :-
    add_internal_error('Unknown status: ', check_status(Status,Expected,Formula,LTLorCTL)).

incomplete_warning(ltl,ltl_incomplete) :- !.
incomplete_warning(ctl,ctl_incomplete) :- !.
incomplete_warning(X,X).

evaluate_from_commandline :-
    retractall(eval_result(_,_)),
    option(eval_string_or_file(A,B,Q,E)),   %print(eval(A,B,Q,E)),nl,
    % treat eval_string and eval_file together to ensure proper order of evaluation
    %  (only possible side-effect at the moment: formula can add new machine_string facts)
    eval_string_or_file(A,B,Q,E),
    fail.
evaluate_from_commandline :- print_eval_results,
    % treat -repl option or -replay File option
    (option(eval_repl([File1|TF]))
      -> (evaluate_expressions([File1|TF]) -> true ; true)
      ; option(eval_repl([]))
      -> (evaluate_expressions([]) -> true ; true)
      ; true).

:- dynamic eval_result/2.
add_eval_result(R) :- retract(eval_result(R,N)),!,
   N1 is N+1, assert(eval_result(R,N1)).
add_eval_result(R) :- assert(eval_result(R,1)).
print_eval_results :- findall(R/N, eval_result(R,N), L), sort(L,SL),
   (SL=[] -> true ; format('Evaluation results: ~w~n',[SL])).

eval_string_or_file(string,String,_,Expected) :-
    set_current_probcli_command(eval_string(String)),
    print('eval: '), print(String),nl,
    eval_string_with_time_out(String,StringResult,EnumWarning,_LS),
    % print(enumwarning(EnumWarning,Expected)),nl,
    add_eval_result(StringResult),
    (Expected\=StringResult
      -> print_error('### Error, expected result to be: '), print_error(Expected),
         error_occurred(eval_string)
      ; EnumWarning = false -> true
      ; gen_enum_warning(Expected) % now also generate enum warnings when Expected not set (eval instead of evalt/evalf)
    ),
   unset_current_probcli_command.
eval_string_or_file(file,File,Quantifier,Expected0) :-
    % evaluate a single formula stored in a file
    set_current_probcli_command(eval_file(File)),
    turn_show_error_source_off, % reduce clutter in user feedback; eval_file used in ProB Logic Calculator for example
    (Expected0 == print_file -> format('~nEvaluating file: ~w~n',[File])
      ; Expected = Expected0, debug_println(10,eval_file(File,Quantifier,Expected)),nl
    ),
    error_manager:reset_error_scopes, % TO DO: avoid that exceptions mess up error scopes in eval_string/file
    statistics(runtime,[Start,_]),nl,
    (Expected=='TRUE' -> TypeInfo=predicate(_) % avoids parsing as expression
     ; true),
    (eval_file(File,Quantifier,Result,EnumWarning,TypeInfo)
      -> add_eval_result(Result),
         (EnumWarning = false -> true ;  gen_enum_warning(EnumWarning)),
         (Result\=Expected -> print_error('### Error, expected result to be: '),
                              print_error(Expected),
                              print_error('### File: '), print_error(File),
                              error_occurred(eval_file)
           ; true)
       ;  add_eval_result(eval_file_failed),
          print_error('### Eval from file failed: '), print_error(File),
          error_occurred(eval_file)
    ),
   (debug_mode(on) -> statistics(runtime,[Stop,_]), Time is Stop - Start,
                      print(File),print(' time: '), print(Time), print(' ms'),nl
      ; true),
  turn_show_error_source_on,
  unset_current_probcli_command.

gen_enum_warning(time_out) :- !,error_occurred(eval_string_time_out).
gen_enum_warning(_) :- print_error('### Enumeration warning occurred'),
                       error_occurred(eval_string_enum_warning,warning).
repl :- evaluate_expressions([]).
:- use_module(parsercall,[ensure_console_parser_launched/0]).
evaluate_expressions(StartFiles) :- nl,
                        print('ProB Interactive Expression and Predicate Evaluator '), nl,
                        print('Type ":help" for more information.'),nl,
                        turn_show_error_source_off, % reduce clutter in user feedback
                        (option(evaldot(File))
                          -> print('Solutions written to dot file: '), print(File),nl
                          ;  true
                        ),
                        print('>>> '),
                        (ensure_console_parser_launched
                           -> maplist(set_repl_input_file(verbose),StartFiles),
                              top_level_eval
                            ; print('>>> ABORTING REPL'),nl),
                        turn_show_error_source_on.

:- use_module(user_interrupts,[interruptable_call/1]).
top_level_eval :-
  catch(top_level_eval1, halt(0), (format('~s', ["Bye."]), nl)).

top_level_eval1 :-
 (interruptable_call(eval1) -> true
  ; print('Evaluation failed or interrupted'),nl,
    print('Use :q to quit REPL'),nl),
  print('>>> '),
  top_level_eval1.
eval :- store_last_error_location_for_repl,
        reset_errors, % get_errors prints errors again and quits in -strict mode
        print('>>> '), garbage_collect, eval1.
eval1 :- repl_multi_read_line(Expr), eval_probcli_repl_line(Expr).

:- dynamic last_repl_error/2.
store_last_error_location_for_repl :-
        retractall(last_repl_error(_,_)),
        check_error_span_file_linecol(_,File,Line,_,_,_),!,
        assert(last_repl_error(File,Line)).
store_last_error_location_for_repl.

:- dynamic current_repl_input_stream/2.
close_repl_input_stream(file_closed) :- retract(current_repl_input_stream(X,File)),!,
   format(":replayed ~w~n",[File]),
   close(X).
close_repl_input_stream(no_file).
:- use_module(tools_io,[safe_open_file/4]).
set_repl_input_file(_,File) :- current_repl_input_stream(_,File),!,
   add_error(set_repl_input_file,'Cyclic file replay: ',File).
set_repl_input_file(Verbose,File) :-
   % close_repl_input_stream, % this way we allow one REPL file to call another
   safe_open_file(File,read,Stream,[]),!,
   (Verbose=verbose -> format('Replaying REPL commands in file: ~w~n',[File]) ; true),
   asserta(current_repl_input_stream(Stream,File)).
set_repl_input_file(_,_).

repl_multi_read_line(Line) :-
   (current_repl_input_stream(Stream,_)
    -> repl_multi_read_line_aux(Stream,[],Line),
       format('~s~n',[Line])
     ; repl_multi_read_line_aux(user_input,[],Line)
    ).
repl_multi_read_line_aux(Stream,SoFar,Line) :-
  read_line(Stream,L),
  (L=end_of_file -> close_repl_input_stream(FileC),
                    ((SoFar=[], FileC = file_closed)
                     -> print('>>> '),repl_multi_read_line(Line) % last line of file empty; do not process
                     ; FileC = file_closed -> Line=SoFar
                     ; Line=end_of_file) % user pressed CTRL-D
    ; append(LFront,[92],L)  % line ends with slash
	   -> append(SoFar,LFront,NewSoFar), print('... '),
		  repl_multi_read_line_aux(Stream,NewSoFar,Line)
	; append(SoFar,L,Line)).

:- use_module(eval_strings).
:- dynamic trace_eval/0.

generate_atom_list([],[],R) :- !, R=[].
generate_atom_list([],Last,[NewAtom]) :-
   reverse(Last,RL),
   atom_codes(NewAtom,RL).
generate_atom_list([39|X],[],[QuotedAtom|T]) :- !,
   get_quoted_atom(X,[],QuotedAtom,Rest),
   strip_leading_ws(Rest,X2),
   generate_atom_list(X2,[],T).
generate_atom_list([32|X],Last,[NewAtom|T]) :- !,
   reverse(Last,RL),
   atom_codes(NewAtom,RL),
   strip_leading_ws(X,X2),
   generate_atom_list(X2,[],T).
generate_atom_list([H|X],Last,Res) :- generate_atom_list(X,[H|Last],Res).

get_quoted_atom([],Acc,QuotedAtom,[]) :- reverse(Acc,R), atom_codes(QuotedAtom,R).
get_quoted_atom([39|T],Acc,QuotedAtom,T) :- !, reverse(Acc,R), atom_codes(QuotedAtom,R).
get_quoted_atom([H|T],Acc,QuotedAtom,Rest) :- get_quoted_atom(T,[H|Acc],QuotedAtom,Rest).


strip_leading_ws([32|X],R) :- !, strip_leading_ws(X,R).
strip_leading_ws(X,X).

:- meta_predicate call_probcli_option(0).
call_probcli_option(_:Option) :- just_assert_option(Option), !,
   (option(Option) -> true ; assert_option(Option)).
call_probcli_option(_:statistics) :- !, % avoid calling SICS version
   cli_print_statistics.
call_probcli_option(Option) :-
   on_exception(error(existence_error(A,B),E),
       call(Option),treat_existence_error(A,B,E,Option)).

% commands that require no execution; just asserting option(.)
just_assert_option(depth_first).
just_assert_option(breadth_first).
just_assert_option(strict_raise_error).
just_assert_option(no_deadlocks).
just_assert_option(no_invariant_violations).
just_assert_option(no_goal).
just_assert_option(no_ltl).
just_assert_option(no_assertion_violations).
just_assert_option(no_counter_examples).

treat_existence_error(source_sink,File,E,Option) :- !,
     format('* Could not find file ~w~n* for probcli command ~w~n* Detailed error: ~w~n',[File,Option,E]).
treat_existence_error(_,_,E,Option) :-
     format('* probcli command not yet supported in REPL: ~w~n* Error: ~w~n',[Option,E]).

reload_mainfile :- file_loaded(_,MainFile),
      print('Reloading and initialising file: '), print(MainFile),nl,
      clear_loaded_files,
      load_main_file(MainFile,0,_),
      get_errors,
      cli_start_animation(0),
      cli_start_initialisation(0).

% REPL EVAL LOOP:
eval_probcli_repl_line(end_of_file) :- !, eval_line(end_of_file).
eval_probcli_repl_line(Line) :- strip_ws(Line,SLine),
   on_exception(E,eval_line(SLine),
                (E=halt(_) -> throw(E) % e.g., coming from :quit; will be caught above
                 ; add_error(repl,'Uncaught Exception in REPL: ',E))).

% strip whitespace at end and beginning
strip_ws([H|T],Res) :- is_ws(H),!, strip_ws(T,Res).
strip_ws(C,Res) :- reverse(C,CR), strip_ws2(CR,SCR), reverse(SCR,Res).
strip_ws2([H|T],Res) :- is_ws(H),!, strip_ws2(T,Res).
strip_ws2(R,R).

is_ws(32).

eval_line([]) :- !, print('Type :q or :quit to quit.'),nl,eval.
eval_line(end_of_file) :- !, halt_exception(0).
% Haskell GHCI like syntax
eval_line(":r") :- !, eval_line("--reload").
eval_line(":reload") :- !, eval_line("--reload").
eval_line("--reload") :- !,
      (reload_mainfile -> true ; get_errors,print('### Error(s) occured during reload (use :e to jump to first error)'),nl),
      eval.
% TO DO: other Haskell commands :info E   :l FILE ,  let pattern = expression
eval_line("--prefs") :- !,print_eclipse_prefs, eval.
eval_line([45|Command]) :- % -command
   generate_atom_list([45|Command],[],ArgV),
   %print(argv(ArgV)),nl,
   % try and parse like commands passed to probcli
   get_options(ArgV,recognised_cli_option,Options,[],fail),
   print('Executing probcli command: '),print(Options),nl,!,
   (maplist(call_probcli_option,Options) -> true
     ; print('Failed to execute probcli arguments'),nl),
   eval.
eval_line("+") :- !, add_last_expression_to_unit_tests, eval.
eval_line("$+") :- !, preferences:temporary_set_preference(expand_avl_upto,-1,CHNG),
    print_last_value,preferences:reset_temporary_preference(expand_avl_upto,CHNG),
    eval.
%eval_line("$$") :- !, print_last_expression, eval.   % now in eval_strings
eval_line([36,36,Num]) :- Num>47, Num<58,!, % $$0 - $$9 commands to print last expression with indentation
   N1 is Num-48,indent_print_last_expression(N1), eval.
%eval_line("$") :- !, print_last_info, eval.  % now in eval_strings
eval_line("!trace") :- !, eval_line("^").
eval_line("^") :- !, (retract(trace_eval) -> print('TRACING OFF'),nl ; assert(trace_eval), print('TRACING ON'),nl), eval.
eval_line("!observe") :- !, toggle_observe_evaluation.
eval_line("!v") :- !, tcltk_turn_debugging_off.
eval_line("!p") :- !, debug:show_perfmessages.
eval_line("%") :- !, print('PROFILING : '), %spy([avl:avl_size/2]),
   (current_prolog_flag(profiling,on)
    -> set_prolog_flag(profiling,off), print('OFF') ;
       set_prolog_flag(profiling,on), print('ON')),
    nl,print('USE %% to print profile info'),nl,eval.
eval_line("%%") :- !, nl,print('PROFILE INFORMATION:'), nl,
   on_exception(error(existence_error(_,_),_),print_profile, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl, debug:timer_statistics,
   eval.
eval_line("%%%") :- !, nl,print('PROFILE INFORMATION (Starting TK Viewer):'), nl,
   on_exception(error(existence_error(_,_),_),(use_module(library(gauge)),view), print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,eval.
eval_line("!debug") :- !, nl,print('ENTERING PROLOG DEBUG MODE:'),
   on_exception(error(existence_error(_,_),_),debug, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,
   eval.
eval_line("@") :- !, get_preference(find_abort_values,OldVal),
  print('Try more aggressively to detect ill-defined expressions: '),
  (OldVal=true -> Val=false ; Val=true), print(Val),nl,
  temporary_set_preference(find_abort_values,Val) , eval.
eval_line("!") :- !, toggle_eval_det,eval.
eval_line("!norm") :- !, toggle_normalising,eval.
eval_line(Codes) :- parse_eval_command(Codes,CommandName,Argument),!,
       debug_println(9,executing_eval_command(CommandName,Argument)),
       (exec_eval_command(CommandName,Argument) -> eval
         ; format('Command ~w failed~n',[CommandName]), eval).
eval_line(ExpressionOrPredicate) :- (trace_eval -> trace ; true),
      (eval_codes(ExpressionOrPredicate,exists,_,_,_,_)
        -> eval ; print('Evaluation failed'),nl,eval).

parse_eval_command([C|Rest],CommandName,Argument) :- [C]=":",
   eval_command(Cmd,CommandName),
   append(Cmd,RestArg,Rest),
   (RestArg = [Letter1|_] -> is_ws(Letter1) /* otherwise command name continues */ ; true),
   strip_ws(RestArg,Argument),
   ((eval_command_help(CommandName,[],_), Argument = [_|_])
     -> format('WARNING: Command ~w does not take arguments!~n',[CommandName])
    ; (eval_command_help(CommandName,[_|_],_), Argument = [])
     -> format('WARNING: Command ~w requires arguments!~n',[CommandName])
     ; true).

% TO DO: some of these commands should also be made available in the Tcl/Tk Console
eval_command("q",quit).
eval_command("quit",quit).
eval_command("halt",quit).
eval_command("x",exit).
eval_command("exit",exit).
eval_command("f",find).
eval_command("find",find).
eval_command("*",apropos).
eval_command("apropos",apropos).
eval_command("help",help).
eval_command("h",help).
eval_command("?",help).
eval_command("ctl",ctl).
eval_command("ltl",ltl).
eval_command("statistics",statistics).
eval_command("stats",statistics).
eval_command("states",state_space_stats).
eval_command("state",show_state_info).
eval_command("u",unsat_core).
eval_command("core",unsat_core).
eval_command("show",show_last_as_table).
eval_command("dot",show_last_as_dot).
eval_command("browse",browse).
eval_command("b",browse).
eval_command("replay",replay_repl_file).
eval_command("src",show_source).
eval_command("source",show_source).
eval_command("origin",show_origin).
eval_command("edit",edit_main_file).
eval_command("e",edit_main_file). % :e
eval_command("comment",comment).

available_commands(SLC) :-
  findall(Cmd,(eval_command(Cs,_),atom_codes(Cmd,[58|Cs])), LC),
  sort(LC,SLC).

eval_command_help(exit,[],'Exit ProB').
eval_command_help(find,['P'],'Find state in state-space which makes LTL atomic proposition P true; LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink').
eval_command_help(ltl,['F'],'Check LTL formula F; LTL Operators: G,F,X,U,W,R,not,&,or,=>; LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink; Past-LTL Operators: Y,H,O,S,T (dual to X,G,F,U,R)').
eval_command_help(ctl,['F'],'Check CTL formula F; CTL Syntax: ExUy,EXx,AXx,EFx,AGx,EX[Op]x,e(Op),{B-Pred}').
eval_command_help(browse,opt('PAT'),'Browse available constants, variables, sets and lets introduced in REPL').
eval_command_help(apropos,['PAT'],'Find constant or variable whose names contains PAT').
eval_command_help(show_last_as_table,[],'Show last evaluated expression in tabular form').
eval_command_help(show_last_as_dot,['F'],'Show expression or predicate F as dot graph').
eval_command_help(unsat_core,[],'Compute Unsatisfiable Core of last evaluated predicate').
eval_command_help(help,opt('CMD'),'Provide help about REPL command CMD').
eval_command_help(replay_repl_file,['FILE'],'Replay FILE of REPL commands').
eval_command_help(show_source,['ID'],'Show origin and source code definition of identifier ID').
eval_command_help(show_origin,['ID'],'Show origin of identifier ID and try opening in EDITOR').
eval_command_help(state_space_stats,[],'Show statistics about state space').
eval_command_help(show_state_info,[],'Show current state').
eval_command_help(statistics,[],'Show statistics about last evaluation').
% implemented in eval_strings:
eval_command_help(type,['E'],'Show type of expression E').
eval_command_help(cvc4,['P'],'Solve predicate P using CVC4 solver').
eval_command_help(kodkod,['P'],'Solve predicate P using SAT solver via Kodkod').
eval_command_help(z3,['P'],'Solve predicate P using Z3 solver').
eval_command_help(edit_main_file,opt('ID'),'Edit main file (or origin of identifier ID) using EDITOR (path_to_text_editor preference)').
eval_command_help(comment,['STRING'],'provide STRING as a comment (mainly useful for :replay files)').

print_eval_command_help(Codes) :-
   eval_command(Codes,Cmd),
   eval_command_help(Cmd,Args,Descr),
   (Args = []
     -> format('Command ~w~n Syntax :~s~n        ~w~n',[Cmd,Codes,Descr])
     ; Args=[Arg] -> format('Command ~w~n Syntax :~s ~w~n       ~w~n',[Cmd,Codes,Arg,Descr])
     ; Args=opt(Arg) -> format('Command ~w~n Syntax :~s [~w]~n       ~w~n',[Cmd,Codes,Arg,Descr])
      ; format('Command ~w~n Syntax :~s ~w~n        ~w~n',[Cmd,Codes,Args,Descr])).

exec_eval_command(quit,_) :- !, halt_exception(0).
exec_eval_command(exit,_) :- !,halt.
exec_eval_command(browse,CodesToMatch) :- !,
     (CodesToMatch=[] -> browse  % maybe merge with apropos functionality
      ; exec_eval_command(apropos,CodesToMatch)).
exec_eval_command(find,FORMULA) :-
    atom_codes(APF,FORMULA),cli_find_ltl_ap(APF).
exec_eval_command(apropos,CodesToMatch) :-  /* :* Pattern (apropos command) */
       browse_machine(CodesToMatch).
exec_eval_command(help,Arg) :-
  (Arg=[] -> eval_help
   ;  print_eval_command_help(Arg) -> true
   ;  (Arg=[58|RA],print_eval_command_help(RA)) -> true  % remove : at front
   ;  format('Cannot provide help about ~s~n',[Arg]),
      available_commands(LC), format('Available commands: ~w~n',[LC])
  ).
exec_eval_command(ctl,FORMULA) :-
    atom_codes(F,FORMULA),
    (cli_ctl_model_check(F,_,Status)
     -> (Status=false -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('CTL Syntax: ExUy,EXx,AXx,EFx,AGx,EX[Op]x,e(Op),{B-Pred}'),nl).
exec_eval_command(ltl,FORMULA) :-
    atom_codes(F,FORMULA),
    (cli_ltl_model_check(F,_,Status)
     -> (Status=no -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('LTL Operators: G,F,X,U,W,R,not,&,or,=>'),nl,
       print('LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink'),nl,
       print('Past-LTL Operators: Y,H,O,S,T (dual to X,G,F,U,R)'),nl
     ).
exec_eval_command(statistics,_) :- !, print_last_info.
exec_eval_command(state_space_stats,_) :- !, state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,ProcessedTotal),
   format('Statespace: ~w states (~w processed) and ~w transitions.~n',[TotalNodeSum,ProcessedTotal,TotalTransSum]).
exec_eval_command(show_state_info,_) :- !,
    current_expression(ID,CurState),
    format('Current state ~w : ',[ID]),translate:print_bstate(CurState),nl.
exec_eval_command(unsat_core,_) :- !,   % :core
    unsat_core_last_expression.
exec_eval_command(show_last_as_table,_) :- !, % :show
    show_last_expression_as_table.
exec_eval_command(show_last_as_dot,Arg) :- !,
    (Arg = [] -> print('*** :dot requires an expression or predicate as argument.'),nl
     ; File='~/probcli_repl.dot',
       set_eval_dot_file(File),
       format('Displaying evaluation result in: ~w~n',[File]),
       (eval_codes(Arg,exists,_,_,_,_) -> true ; true), unset_eval_dot_file).
exec_eval_command(replay_repl_file,FILEC) :- !,
    atom_codes(File,FILEC),
    set_repl_input_file(not_verbose,File).
exec_eval_command(show_source,IDC) :- !,
    atom_codes(ID,IDC),
    show_source(ID).
exec_eval_command(show_origin,IDC) :- !,
    atom_codes(ID,IDC),
    show_origin(ID).
exec_eval_command(edit_main_file,Arg) :- !,
    (Arg=[] -> edit_main_file ; exec_eval_command(show_origin,Arg)).
exec_eval_command(comment,_Arg) :- !. % do nothing; argument was a comment; mainly useful for :replay files

:- use_module(tools_commands,[edit_file/2]).
edit_main_file :- last_repl_error(File,Line),!,
     format('Showing first error from last command~n',[]),
     edit_file(File,Line).
edit_main_file :- file_loaded(true,MainFile),!,edit_file(MainFile,unknown).
% Note: for the bbedit command we can also specify line numbers bbedit +LINE FILE
edit_main_file :- format('No file loaded, cannot open EDITOR!~n',[]).



open_file_at_position(OriginTerm) :-
    error_manager:extract_line_col(OriginTerm,LINE,_COL,_Erow,_Ecol),
    error_manager:extract_file_name(OriginTerm,FILE),
    edit_file(FILE,LINE).


show_source(ID) :- source_code_for_identifier(ID,Kind,_Type,OriginStr,OriginTerm,Source),!,
    translate:translate_subst_or_bexpr(Source,PPS),
    %format('~w: ~w (Type: ~w)~norigin: ~w~nsource: ~w~n',[Kind,ID,_Type,Origin,PPS]).
    format('~w: ~w~norigin: ~w~nsource: ~w~n',[Kind,ID,OriginStr,PPS]),
    ((OriginTerm=b(_,_,_),bsyntaxtree:get_texpr_description(OriginTerm,Description))
     -> format('description: ~w~n',[Description]) ; true).
show_source(ID) :- format('### Could not find source for ~w~n',[ID]).

show_origin('') :- last_repl_error(_,_),!, % error occured: show error in editor like :e would
   edit_main_file.
show_origin('') :- !,format('### You need to provided an identifier~n',[]).
show_origin(ID) :- source_code_for_identifier(ID,Kind,_Type,OriginStr,OriginTerm,_Source),!,
    format('~w: ~w~norigin: ~w~n',[Kind,ID,OriginStr]),
    open_file_at_position(OriginTerm).
show_origin(ID) :- format('### Could not find origin for ~w~n',[ID]).

profiling_on :- set_prolog_flag(profiling,on), print('% PROFILING ON'),nl.

% find a state satisfying LTL atomic property
cli_find_ltl_ap(APF) :-
    if(ltl:find_atomic_property_formula(APF,ID),
         (format('Found state (id = ~w) satisfying LTL atomic property.~n',[ID]),
          user:tcltk_goto_state('LTL FIND',ID)),
         format('No explored state satsifies LTL atomic property.~n',[])).

eval_help :-
       print('ProB Interactive Expression and Predicate Evaluator '), nl,
       print('Type a valid B expressions or predicates, followed by RETURN or ENTER.'),nl,
       print('You can spread input over multiple lines by ending lines with "\\".'),nl,
       browse_machine([]),
       print('You can also type one of the following commands: '),nl,
       (option_verbose ->
          print('    + to save last expression to ProB unit tests.'),nl,
          print('    ! to go to deterministic propagation only mode.'),nl
         ; true),
       print('    let x = E to define a new local variable x and :b to browse the variables'),nl,
       print('    :t E to get the type of an expression and :r to reload the machine'),nl,
       print('    :show to display the last result as a table (if possible)'),nl,
       print('    :list CAT to display information with CAT : {files,variables,help,...}'),nl,
       print('    :* P to display constants/variables containing pattern P'),nl,
       print('    :core to compute the unsat core for last result'),nl,
       print('    :stats to print the type and evaluation time for last query'),nl,
       print('    -PROBCLIARGS to pass command-line probcli arguments to the REPL'),nl,
       print('    :ctl F or :ltl F to check a CTL or LTL formula.'),nl,
       print('    :f F to find a state satisfying LTL atomic property.'),nl,
       print('    :help CMD to obtain more help about a command.'),nl,
       print('    :replay FILE to replay a file of commands.'),nl,
       print('    :z3 P, :cvc4 P, :kodkod P to solve predicate P using alternate solver'),nl,
       print('    :states, :statistics, :source, :orgin, :dot - use :help for more info'),nl,
       print('    :q to exit.'),nl.


browse :- browse_machine([]), browse_repl_lets.

% the CodesToMatch parameters mimics the apropos command of the Clojure-REPL
browse_machine(CodesToMatch) :-
       findall(FID, (bmachine:b_filenumber(FID,Type,_,_),Type \= def),MN),
       display_match('MACHINES',CodesToMatch,MN),
       (CodesToMatch = [] -> print_sets ; true), % TO DO: also match inside SETS
       findall(FID, bmachine:b_filenumber(FID,def,_,_),DFN),
       (DFN=[] -> true ; display_match('DEFINITIONS FILES',CodesToMatch,DFN)),
       findall(DefID,bmachine:b_get_definition(DefID,_,_,_,_),DN),
       (DN=[] -> true ; display_match('DEFINITIONS',CodesToMatch,DN)),
       bmachine:b_get_machine_constants(Constants),
       bsyntaxtree:get_texpr_ids(Constants,CN),
       display_match('CONSTANTS',CodesToMatch,CN),
       bmachine:b_get_machine_variables(Variables),
       bsyntaxtree:get_texpr_ids(Variables,VN),
       display_match('VARIABLES',CodesToMatch,VN).

display_match(KIND,CodesToMatch,Ids) :-
       include(atom_contains_codes(CodesToMatch),Ids,MatchingIds),
       length(MatchingIds,LenMIds),
       sort(MatchingIds,SMatchingIds),
       (CodesToMatch=[]
        -> format('Available ~w: ~w ~w~n',[KIND,LenMIds,SMatchingIds])
         ; length(Ids,LenIds),
           format('Matching ~w: ~w/~w ~w~n',[KIND,LenMIds,LenIds,SMatchingIds])).

% check if an atom contains a list of codes in its name
atom_contains_codes([],_) :- !.
atom_contains_codes(Codes,Name) :- atom_codes(Name,NC), append(Codes,_,C),
   append(_,C,NC).

print_sets :- print('Available SETS: '), b_global_set(GS), print_set(GS),fail.
print_sets :- nl.

print_set(GS) :- print(GS), \+ is_b_global_constant(GS,_,_),!, print(' ').
print_set(GS) :- print(' = {'), is_b_global_constant(GS,_,Cst), print(Cst), print(' '),fail.
print_set(_) :- print(' } ').

% showing relations as tables:

:- use_module(table_tools).
show_last_expression_as_table :- \+ last_expression(_,_Expr,_Value),!,
   print('Please evaluate an expression or predicate first.'),nl.
show_last_expression_as_table :-
   get_last_result_value(Expr,_,Value),
   print_value_as_table(Expr,Value).


% a few definitions so that probcli commands work in REPL:

:- public pretty_print_internal_rep/1, pretty_print_all_types/1, pretty_print_internal_rep_to_B/1.
pretty_print_internal_rep(PPFILE) :- b_write_machine_representation_to_file(needed,PPFILE).
pretty_print_all_types(PPFILE) :- b_write_machine_representation_to_file(all,PPFILE).
pretty_print_internal_rep_to_B(PPFILE) :- b_write_eventb_machine_to_classicalb_to_file(PPFILE).


% Simple Animator

animate_machine :- nl,print('IMPORTANT: Do not use this mode for automatic tools.'),nl,
                   print('The output format can change arbitrarily in future versions.'),nl,
                   print('Please terminate your input with a dot (.) and then type return.'),nl,nl,
                   animate_machine2.

animate_machine2 :-
  print_current_state,
  cli_computeOperations(Ops),
  length(Ops,Max),
  print('Enabled Operations: '),nl,
  print_options(Ops,1),
  print(' ==> '),!,
  read(Nr),
  ((number(Nr),Nr>0,Nr=<Max)
    -> cli_animateOperationNr(Nr,Ops,0)
    ;  fail
  ),!,
  animate_machine2.
animate_machine2.

print_current_state :- getCurrentStateID(CurID), print('ID ==> '), print(CurID),nl,
  getStateValues(CurID,State),
  print_bindings(State),
  ((specfile:b_or_z_mode,\+is_initialised_state(CurID))
  -> print(' Not yet initialised.'),print_mode_info, debug_println(10,state(State)) ; nl).

print_mode_info :- animation_mode(M), (animation_minor_mode(MM) -> true ; MM=''),
  format('Animation Mode = ~w [~w]~n',[M,MM]).

cli_computeOperations(Ops) :- option(animate_stats),!, % provide statistics about the animation
   nl,
   start_probcli_timer(Timer),
   computeOperations(Ops),
   stop_probcli_timer(Timer,'Time: ').
cli_computeOperations(Ops) :- computeOperations(Ops).

cli_animateOperationNr(Nr,Options,StepNr) :-
   (option(animate_stats)
     -> nth1(Nr,Options,Action),
        truncate_animate_action(Action,TA),
        (StepNr>1 -> format('performing step ~w : ~w~n',[StepNr,TA])
          ; format('performing ~w~n',[TA]))
     ; true),
   executeOperationNr_on_computed_node(Nr).

% optionally truncate animation action atom for printing:
truncate_animate_action(Action,TA) :-
   (option_verbose -> TA = Action
	  ; \+ atom(Action) -> TA = Action
	  ; truncate_atom(Action,100,TA)).

perform_random_step(StepNr) :- perform_random_step(_Ops,_Len,_RanChoice,StepNr).
perform_random_step(Ops,Len,RanChoice,StepNr) :-
  cli_computeOperations(Ops),
  getCurrentStateID(CurID), check_for_errors(CurID,StepNr),
  length(Ops,Len), Len>0,
  debug_println(20,perform_random_step(Len,StepNr)),
  L1 is Len+1,
  ((do_det_checking, Len>1)
    ->  print_error('### Non-deterministic step in animate or init'),
        print_error('### State:'),
        print_current_state,  print_error('### Enabled Operations: '), print_options(Ops,1),
        error_occurred(det_check)
    ; true),
  random(1,L1,RanChoice),
  debug_println(20,random(L1,RanChoice)),
  cli_animateOperationNr(RanChoice,Ops,StepNr).

:- use_module(state_space,[visited_expression/2]).
check_for_errors(CurID,StepNr) :- invariant_violated(CurID),
   \+ option(no_invariant_violations),
   get_preference(do_invariant_checking,true),
   tools:ajoin(['INVARIANT VIOLATED after ',StepNr,' steps (state id ',CurID,').'],ErrMsg),
   format('~w~n',[ErrMsg]),
   visited_expression(CurID,CurState), print_state_silent(CurState),
   error_occurred_with_msg(invariant_violation,ErrMsg),
   fail.
check_for_errors(CurID,_) :- get_state_errors(CurID).
% TO DO: also check for assertion errors, goal,  state_errors with abort

do_det_checking :- option(det_check),!.
do_det_checking :- option(det_constants_check),getCurrentStateID(root),
   b_mode, b_machine_has_constants_or_properties.

perform_random_steps(Nr,_) :- \+ number(Nr),!,
   print_error('### Argument to animate not a number'), print_error(Nr),error_occurred(animate).
perform_random_steps(Nr,_) :- Nr<0, !,
   print_error('### Argument to animate is a negative number'), print_error(Nr),error_occurred(animate).
perform_random_steps(0,_) :- !.
perform_random_steps(Nr,ErrorOnDeadlock) :-
   perform_random_initialisation_if_necessary(Steps), % if Nr=1 we currently will also execute the INITIALISATION ! TO DO: fix
   perform_random_steps_aux(Steps,Nr,ErrorOnDeadlock).

perform_random_steps_aux(Nr,Max,_) :- Nr >= Max,!, debug_println(9,performed_random_steps(Nr)).
perform_random_steps_aux(Nr,Max,ErrorOnDeadlock) :-
   N1 is Nr+1,
   (perform_random_step(N1)
     -> perform_random_steps_aux(N1,Max,ErrorOnDeadlock)
     ;  /* deadlock */
      write_xml_element_to_log(deadlock_found,[step/Nr]),
      (ErrorOnDeadlock=true, \+ option(no_deadlocks)) ->
       print_error('### Deadlock occurred during -animate, at step number:'), print_error(Nr),
       error_occurred(animate)
     ; print('% Deadlock occurred during -animate, at step number:'), print(Nr),nl
    ).

perform_random_initialisation_if_necessary(Steps) :-
   b_mode, getCurrentStateID(State), State=root,!, perform_random_initialisation(Steps).
perform_random_initialisation_if_necessary(0).

perform_random_initialisation :- perform_random_initialisation(_).
perform_random_initialisation(Steps) :- getCurrentStateID(State), State \= root, !,
   print_error('### init can only be used in initial state'), print_error(State),error_occurred(initialisation),
   Steps=0.
perform_random_initialisation(Steps) :- b_mode, b_machine_has_constants_or_properties,!,
   (perform_random_step(Ops,_Len,RanChoice,1)
    -> nth1(RanChoice,Ops,Choice), %print(Choice),nl,
       (Choice = 'PARTIAL_SETUP_CONSTANTS'
         -> error_occurred(setup_constants_inconsistent)
         ; true)
    ; error_occurred(setup_constants_fails),fail), % $setup_constants
   perform_random_init_after_setup_constants, Steps=2. % $initialise_machine
perform_random_initialisation(Steps) :- (perform_random_step(1) -> Steps=1 ; error_occurred(initialisation_fails),fail).


perform_random_init_after_setup_constants :- \+ option(initialise), we_need_only_static_assertions(_),!,
         print('% NOT INITIALISING MACHINE (not required)'),nl.
 %  debug_println(20,'% NOT INITIALISING MACHINE (not required)').
perform_random_init_after_setup_constants :-
   (perform_random_step(2)  % 2 is the step nr not the number of steps
    -> true ; error_occurred(initialisation_fails),fail).

:- use_module(enabling_analysis,[tcltk_cbc_enabling_analysis/1, print_enable_table/1, is_timeout_enabling_result/1]).
do_enabling_analysis_csv(EnablingCsvFile,NOW) :-
    start_probcli_timer(Timer1),
    start_xml_feature(enabling_analysis,file,EnablingCsvFile,FINFO),
    tcltk_cbc_enabling_analysis(list(R)),
    stop_probcli_timer(Timer1,'% Finished CBC Enabling Analysis',_TotWallTime),
    print_cbc_stats(R,NOW),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table(R),
    told,!,
    stop_xml_feature(enabling_analysis,FINFO).
do_enabling_analysis_csv(EnablingCsvFile,_) :-
    add_error(enabling_analysis,'Enabling analysis failed',EnablingCsvFile),
    stop_xml_group_in_log(enabling_analysis).

print_cbc_stats(Res,_NOW) :- length(Res,Len), Ops is Len-2, % Header + Init
      CBC_Calls is Ops*(Ops+1), % +1 for INITIALISATION
      findall(TO,(member(list([_|T]),Res), member(TO,T),is_timeout_enabling_result(TO)),TOS),
      length(TOS,NrTOS),
      format('% CBC Enabling Stats:~n% Nr of events: ~w~n% Nr of cbc calls: ~w, Timeout results: ~w~n',[Ops,CBC_Calls,NrTOS]),
      write_xml_element_to_log(cbc_enabling_stats,[nr_events/Ops,cbc_calls/CBC_Calls,nr_timeouts/NrTOS]).


:- use_module(enabling_analysis,[feasible_operation_with_timeout/3]).
do_feasibility_analysis(ATimeOut,EnablingCsvFile) :-
    arg_is_number(ATimeOut,TimeOut),
    start_xml_feature(feasibility_analysis,file,EnablingCsvFile,FINFO),
    findall(list([Op,Res]),feasible_operation_with_timeout(Op,TimeOut,Res),R),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table([list(['Event','Feasibility'])|R]),
    told,!,
    stop_xml_feature(feasibility_analysis,FINFO).
do_feasibility_analysis(_,EnablingCsvFile) :-
    add_error(feasibility_analysis,'Feasibility analysis failed',EnablingCsvFile),
    stop_xml_group_in_log(feasibility_analysis).

:- use_module(b_read_write_info,[tcltk_read_write_matrix/1]).
generate_read_write_matrix(CsvFile) :-
    tcltk_read_write_matrix(list(Matrix)),
    my_tell(CsvFile),
    print_enable_table(Matrix),
    told,!.
generate_read_write_matrix(CsvFile) :-
    add_error(read_write_matrix,'Generating Read-Write-Matrix failed',CsvFile).


my_tell(File) :- on_exception(error(_E,_), % existence_error(_,_)
                   tell(File), add_error_fail(tell,'File cannot be written to: ',File)).

print_options([],_).
print_options([H|T],N) :-
  print(' '), print(N), print(':'), print(H),nl,
  N1 is N+1,
  print_options(T,N1).

print_list([],_) :- !.
print_list([H|T],N) :- !,
  print(' '), print(N), print(': '),print(H), print(': '),
  (visited_expression(H,State) -> print(State) ; true),nl,
  N1 is N+1,
  print_list(T,N1).
print_list(X,_) :- print('### not a list: '), print(X),nl.

print_bindings([]) :- !.
print_bindings([binding(Var,_,PPV)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings([binding(Var,_,PPV,_Tag)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings(X) :- print('### Internal Error: illegal binding list: '), print(X),nl.

:- dynamic expected_error_occurred/1.
:- dynamic error_did_not_occur/1.
reset_expected_error_occurred :- retractall(expected_error_occurred(_)).
check_all_expected_errors_occurred(NOW) :-
  %error_manager:display_error_statistics,
  get_errors, get_state_space_errors,
  retractall(error_did_not_occur(_)),
  expected_error(Type),
  \+ expected_error_occurred(Type),
  print('*** Expected Error of following type to occur: '), print(Type),nl,
  writeln_log_time(expected_error_did_not_occur(NOW,Type)),
  assert(error_did_not_occur(Type)),
  (option(strict_raise_error) -> definite_error_occurred ; fail).
check_all_expected_errors_occurred(_NOW) :-
   (expected_error(_)
    -> (error_did_not_occur(_) -> print('*** Some expected errors did NOT occur !')
         ; print('All expected errors occurred.')),nl
    ;  true).

expected_error(Type) :- option(expect_error(Type)).
expected_error(Type) :- option(expect_error_pos(Type,_Line,_Col)).

error_occurred(warning(Type)) :- !, error_occurred(Type,warning).
error_occurred(Type) :- error_occurred(Type,error).

get_error_category_and_type(warning(Cat),Category,Type) :- !, Category=Cat,Type=warning.
get_error_category_and_type(C,C,error).

error_occurred_with_msg(warning(Type),Msg) :- !, error_occurred(Type,warning,not_yet_extracted,Msg).
error_occurred_with_msg(Type,Msg) :- error_occurred(Type,error,not_yet_extracted,Msg).

error_occurred(Type,ErrOrWarn) :- error_occurred(Type,ErrOrWarn,not_yet_extracted,'').

error_occurred(Type,ErrOrWarning,ExtractedSpan,Msg) :- option(expect_error_pos(Type,Line,Col)),!,
    write_xml_element_to_log(expected_error_occurred,[category/Type, type/ErrOrWarning, message/Msg]),
    assert(expected_error_occurred(Type)),
    (get_error_or_warning_span(ExtractedSpan,Type,EL,EC)
      -> ((EL,EC)=(Line,Col) -> debug_println(9,expect_error_pos_ok(Type,EL,EC))
           ; format('Unexpected line ~w and column ~w for error ~w! Expected ~w:~w.~n',[EL,EC,Type,Line,Col]),
             definite_error_occurred
         )
      ;  format('*** Could not obtain position information for error ~w! Expected ~w:~w.~n',[Type,Line,Col]),
         %display_error_statistics,
         definite_error_occurred).
error_occurred(Type,ErrOrWarning,_,Msg) :-
    option(expect_error(Type)),!,
    write_xml_element_to_log(expected_error_occurred,[category/Type, type/ErrOrWarning, message/Msg]),
    assert(expected_error_occurred(Type)).
error_occurred(Type,ErrOrWarning,_,Msg) :-
    (probcli_time_stamp(NOW) -> true ; NOW=unknown),
    writeln_log(error_occurred(NOW,Type)),
    (option(optional_error(Type)) ->
        write_xml_element_to_log(optional_error_occurred,[category/Type, type/ErrOrWarning, message/Msg]),
        format(user_output,'% Optional error occured: ~w~n',[Type])
     ; write_xml_element_to_log(error_occurred,[category/Type, type/ErrOrWarning, message/Msg]), % TO DO: maybe extract position info
       (ErrOrWarning = warning -> inc_counter(cli_warnings) ; inc_counter(cli_errors)),
       (option(strict_raise_error) ->
			print_error('*** Unexpected error occurred ***'),
			print_error(Type),
            findall(Err,option(expect_error(Err)),Ls), (Ls=[] -> true ; print_error(expected(Ls))),
			definite_error_occurred
		 ; (ErrOrWarning=error,serious_error(Type))
		  -> print_error('*** Serious error occurred ***'),
			 print_error(Type),
			 definite_error_occurred
		 ;  print_probcli_error_non_strict(Type,ErrOrWarning)
       )
    ).


get_error_or_warning_span(not_yet_extracted,Type,EL,EC) :- check_error_span_file_linecol(Type,_File,EL,EC,_,_).
get_error_or_warning_span(not_yet_extracted,Type,EL,EC) :- check_error_span_file_linecol(warning(Type),_File,EL,EC,_,_).
get_error_or_warning_span(Span,_,EL,EC) :- Span \= not_yet_extracted, extract_line_col(Span,EL,EC,_,_).


% a list of serious errors: if these occur; then return code different from 0 even in non-strict mode
serious_error(get_java_command_path).
serious_error(internal_error(_)).

print_probcli_error_non_strict(parse_machine_predicate_error,_) :-
    !. % have already been reported
print_probcli_error_non_strict(Type,ErrOrWarning) :-
  (ErrOrWarning=warning -> print_error('*** warning occurred ***')
     ; print_error('*** error occurred ***')),
  print_error(Type).

definite_error_occurred :- print_error('*** Abnormal termination of probcli !'),
  (file_loaded(_,File) -> print_error('*** for_file'(File)) ; true),
  (current_probcli_command(Cmd) -> print_error('*** for_command'(Cmd)) ; true),
  (probcli_time_stamp(NOW) -> halt1_prob(NOW)
   ; writeln_log(halt(1)),
     halt_exception(1)
  ).

:- dynamic current_probcli_command/1.
set_current_probcli_command(X) :- retractall(current_probcli_command(_)),
  assert(current_probcli_command(X)).
unset_current_probcli_command :- retractall(current_probcli_command(_)).

halt_prob(NOW) :-
  write_xml_element_to_log(probcli_halted_prematurely,[now/NOW]),
  check_all_expected_errors_occurred(NOW),
  close_all_xml_groups_in_log,
  halt_exception.
halt1_prob(NOW) :-
  write_xml_element_to_log(probcli_halted_prematurely,[now/NOW]),
  close_all_xml_groups_in_log,
  halt_exception(1).


:- dynamic accumulated_infos/3.
accumulate_infos(_,[]).
accumulate_infos(Context,[Info|T]) :- get_info(Info,FF,Nr),Nr>0,!,
  (retract(accumulated_infos(Context,FF,OldNr))->true ; OldNr=0),
  N1 is OldNr+Nr,
  assert(accumulated_infos(Context,FF,N1)),
  accumulate_infos(Context,T).
accumulate_infos(Context,[_|T]) :- accumulate_infos(Context,T).
get_info(FF-Nr,FF,Nr).
get_info(FF/Nr,FF,Nr).
print_accumulated_infos(NrFilesProcessed) :- NrFilesProcessed>1, accumulated_infos(_,_,_),!,
    format('Analysis summary (~w files processed): ',[NrFilesProcessed]),
    start_xml_group_in_log(summary,files_processed,NrFilesProcessed),
    findall(Context-F-Nr,accumulated_infos(Context,F,Nr),L), sort(L,SL),
    maplist(pracc,SL),nl,
    % TO DO: write infos to XML log
    stop_xml_group_in_log_no_statistics(summary).
print_accumulated_infos(_NrFilesProcessed).
pracc(Context-F-Nr) :- format('~w:~w:~w ',[Context,F,Nr]).

write_important_xml_element_to_log(Category,Infos) :-
   include(important_info,Infos,II),
   write_xml_element_to_log(Category,II).
important_info(FF/Nr) :- Nr \= 0, \+ irrelevant_xml_info(FF).
irrelevant_xml_info(true_after_expansion).
irrelevant_xml_info(false_after_expansion).
irrelevant_xml_info(unknown_after_expansion).
irrelevant_xml_info(total_after_expansion).

check_required_infos([],_,_).
check_required_infos([H|T],Infos,ErrType) :-
  (member(H,Infos) -> check_required_infos(T,Infos,ErrType)
     ; print('*** Unexpected result while checking: '),
       translate_err_type(ErrType,ES), print(ES),nl,
       print('***    expected : '), print(H),nl,
       print('***          in : '), print(Infos),nl,
       error_occurred(ErrType)).
translate_err_type(check_assertions,'ASSERTIONS') :- !.
translate_err_type(cli_check_assertions,'ASSERTIONS') :- !.
translate_err_type(check_goal,'GOAL') :- !.
translate_err_type(load_po_file,'PROOF OBLIGATIONS') :- !.
translate_err_type(X,X).

:- public mc_ok_arg/2.
mc_ok_arg(Arg,X) :- Arg==all,!,prolog_flag(max_tagged_integer,X).
mc_ok_arg(Arg,N) :- tools:arg_is_number(Arg,N).


:- dynamic option/1.
assert_all_options([]).
assert_all_options([Opt|T]) :- assert_option(Opt),
    assert_all_options(T).

assert_option(Opt) :- assert(option(Opt)),
    (Opt=silent -> set_silent_mode(on) ; true).


% replace a leading double-dash -- by a single dash and replace inner dashes by underscores
normalise_option_atom(X,RX) :- atom(X),!,
    atom_codes(X,CodesX),
    % remove leading dash
    ((CodesX=[45,45,H|T], H\=45)
     -> maplist(convert_dash_to_underscore,[H|T],HT2),
        RXCodes=[45|HT2]
     ; CodesX = [45|T]
     -> maplist(convert_dash_to_underscore,T,T2),
        RXCodes=[45|T2]
     ; maplist(convert_dash_to_underscore,CodesX,RXCodes)
    ),
    atom_codes(RX,RXCodes).
normalise_option_atom(T,T).

% replace dashes by underscores
normalise_pref_name(X,RX) :- atom(X),!,
    atom_codes(X,CodesX),
    maplist(convert_dash_to_underscore,CodesX,C2),
    atom_codes(RX,C2).
normalise_pref_name(T,T).

convert_dash_to_underscore(45,R) :- !, R=95.
convert_dash_to_underscore(X,X).

recognised_cli_option(X,Opt,Args,Condition) :- normalise_option_atom(X,RX),
   recognised_option(RX,Opt,Args,Condition).

recognised_option(X,Opt,[],true) :- recognised_option(X,Opt). % options without arguments
recognised_option(X,Opt,Args,true) :- recognised_option(X,Opt,Args). % options with arguments but no code needed to check arguments

recognised_option('-mc',cli_mc(N),[Arg],user:mc_ok_arg(Arg,N)).
recognised_option('-model_check',cli_mc(LimitNr),[],true) :- max_tagged_integer(LimitNr).
recognised_option('-timeout',timeout(N),[Arg],tools:arg_is_number(Arg,N)). % for model checking, refinement checking and for disprover per PO
recognised_option('-time_out',timeout(N),[Arg],tools:arg_is_number(Arg,N)).
recognised_option('-s',socket(S),[Arg],tools:arg_is_number(Arg,S)).
recognised_option('-cc',coverage(N,N2,just_check_stats),[Arg,Arg2],
                   (arg_is_number_or_wildcard(Arg,N),arg_is_number_or_wildcard(Arg2,N2))).
recognised_option('-hash',check_statespace_hash(H,_),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash64',check_statespace_hash(H,'64bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash32',check_statespace_hash(H,'32bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-ltllimit',ltl_limit(Nr),[Arg], tools:arg_is_number(Arg,Nr)).
recognised_option('-check_disprover_result',cli_check_disprover_result([true-TNr,false-FNr,unknown-UNr,failure-0]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-aa',cli_check_assertions(all,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-ma',cli_check_assertions(main,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-kodkod_comparision',kodkod_comparision(Nr),[Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-kodkod_performance',kodkod_performance(File,Nr),[File,Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-animate',cli_random_animate(N,true),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-execute',execute(N,true,current_state),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-logxml_write_vars',logxml_write_vars(Prefix),[Prefix],true).
recognised_option('-expcteval',expect_evaluation(List),[Arg],user:evalchecklist(Arg,List)).

recognised_option('-cbc_tests', cbc_tests(Depth,EndPred,Output),[Depth,EndPred,Output],
                  tools:check_filename_arg(Output,'cbc_tests')).
recognised_option('-mcm_tests', mcm_tests(Depth,MaxStates,EndPred,Output),[Depth,MaxStates,EndPred,Output],
                  tools:check_filename_arg(Output,'mcm_tests')).
recognised_option('-test_description', test_description(File), [File],
                  tools:check_filename_arg(File,'test_description')).
recognised_option('-all_paths', all_deadlocking_paths(File), [File],
                  tools:check_filename_arg(File,'all_paths')).
recognised_option('-dot',dot_command(Category,File),[Category,File],
                  tools:check_filename_arg(File,'dot')).
recognised_option('-spdot',dot_command(state_space,File),[File], tools:check_filename_arg(File,'spdot')). % we keep this : it is shown in Wiki
% recognised_option('-spmdot',dot_command(signature_merge,File),[File], tools:check_filename_arg(File,'spmdot')).
% recognised_option('-spddot',dot_command(dfa_merge,File),[File], tools:check_filename_arg(File,'spddot')).
% recognised_option('-sgdot',dot_command(state_as_graph,File),[File],  tools:check_filename_arg(File,'sgdot')).
recognised_option('-dotexpr',dot_command_for_expr(Category,Expr,File,[]),[Category,Expr,File],
                  tools:check_filename_arg(File,'dot')).
%recognised_option('-sgedot',dot_command_for_expr(expr_as_graph,Expr,File,[]),[Expr,File], tools:check_filename_arg(File,'sgedot')).
% recognised_option('-sptdot',dot_command_for_expr(transition_diagram,Expr,File,[]),[Expr,File],tools:check_filename_arg(File,'sptdot')).
%recognised_option('-invdot',generate_dot_from_invariant(File),[File],   tools:check_filename_arg(File,'invdot')).
%recognised_option('-propdot',generate_dot_from_properties(File),[File],    tools:check_filename_arg(File,'propdot')).
%recognised_option('-assdot',generate_dot_from_assertions(File),[File],  tools:check_filename_arg(File,'assdot')).
%recognised_option('-deaddot',generate_dot_from_deadlock_po(File),[File],   tools:check_filename_arg(File,'deaddot')).
recognised_option('-csvhist',evaluate_expression_over_history_to_csv_file(Expr,File),[Expr,File],
                  tools:check_filename_arg(File,'csvhist')).
recognised_option('-read_write_matrix_csv',generate_read_write_matrix_csv(CsvFile),
                  [CsvFile],
                  tools:check_filename_arg(CsvFile,'read_write_matrix_csv')).
recognised_option('-feasibility_analysis_csv',feasibility_analysis_csv(TimeOut,EnablingCsvFile),
                  [TimeOut,EnablingCsvFile],
                  tools:check_filename_arg(EnablingCsvFile,'feasibility_analysis_csv')).
recognised_option('-l',log(Log,prolog),[Log],
                  tools:check_filename_arg(Log,'l')).
recognised_option('-log',log(Log,prolog),[Log],
                  tools:check_filename_arg(Log,'log')).
recognised_option('-logxml',log(Log,xml),[Log],
                  tools:check_filename_arg(Log,'logxml')).
recognised_option('-pp',pretty_print_internal_rep(File),[File],
                  tools:check_filename_arg(File,'pp')).
recognised_option('-ppf',pretty_print_all_types(File),[File],
                  tools:check_filename_arg(File,'ppf')).
recognised_option('-ppB',pretty_print_internal_rep_to_B(File),[File],
                  tools:check_filename_arg(File,'ppB')).
recognised_option('-save_state',save_state(Filename),[Filename],
                  tools:check_filename_arg(Filename,'save_state')).
recognised_option('-save',save_state_for_refinement(Filename),[Filename],
                  tools:check_filename_arg(Filename,'save')).
recognised_option('-sptxt',print_values(Filename),[Filename],
                  tools:check_filename_arg(Filename,'sptxt')).
recognised_option('-sstxt',print_all_values(Dirname),[Dirname],
                  tools:check_filename_arg(Dirname,'sstxt')).
recognised_option('-latex',process_latex_file(In,Out),[In,Out],
                  (tools:check_filename_arg(In,'latex'),tools:check_filename_arg(Out,'latex'))).

recognised_option('-zmq_master2',zmq_master2(N,Max,PortStart,Recover,Logfile),[SN,SMax,SPortStart,SRecover,Logfile],
                  tools:(arg_is_number(SN,N),arg_is_integer(SMax,Max),
                         arg_is_number(SPortStart,PortStart),arg_is_number(SRecover,Recover))).
recognised_option('-zmq_assertions',zmq_assertion(PortStart, Logfile),[SPortStart,Logfile],
                  tools:(arg_is_number(SPortStart, PortStart))).
recognised_option('-zmq_master',zmq_master(Logfile),[Logfile], true).
recognised_option('-zmq_worker2',zmq_worker2(MasterIP, Port, Recover),[MasterIP, SPort, SRecover],
                  tools:(arg_is_number(SPort,Port), arg_is_number(SRecover, Recover))).
recognised_option('-p',set_pref(NPREF,PREFVAL),[PREF,PREFVAL],user:normalise_pref_name(PREF,NPREF)).
recognised_option('-pref',set_pref(NPREF,PREFVAL),[PREF,PREFVAL],user:normalise_pref_name(PREF,NPREF)).
recognised_option('-cbc_redundant_invariants',cbc_redundant_invariants(Nr),[X],tools:arg_is_number(X,Nr)).

recognised_option('-pref_group',set_preference_group(Group,Val),[Group,Val]).
recognised_option('-prefs',set_prefs_from_file(PREFFILE),[PREFFILE]).
recognised_option('-plugin',plugin(Plugin), [Plugin]).
recognised_option('-card',set_card(SET,SCOPE),[SET,SCOPE]).
recognised_option('-argv',set_argv(ARGV),[ARGV]).
recognised_option('-goal',set_goal(GOAL),[GOAL]).
recognised_option('-property',add_additional_property(PRED),[PRED]).
recognised_option('-scope',set_searchscope(GOAL),[GOAL]).
recognised_option('-searchscope',set_searchscope(GOAL),[GOAL]).
recognised_option('-eval',eval_string_or_file(string,E,exists,_),[E]).
recognised_option('-evalt',eval_string_or_file(string,E,exists,'TRUE'),[E]).
recognised_option('-evalf',eval_string_or_file(string,E,exists,'FALSE'),[E]).
recognised_option('-evalu',eval_string_or_file(string,E,exists,'UNKNOWN'),[E]).
recognised_option('-eval_file',eval_string_or_file(file,F,exists,print_file),[F]).
recognised_option('-evalt_file',eval_string_or_file(file,F,exists,'TRUE'),[F]).
recognised_option('-eval_rule_file',eval_string_or_file(file,F,forall,print_file),[F]).
recognised_option('-parsercp',(L),[L]).
recognised_option('-lg',analyse_log(Log,timestamp,true),[Log]).
recognised_option('-lgr',analyse_log(Log,revision,false),[Log]).
recognised_option('-check_log',check_log(Log),[Log]).
recognised_option('-expcterr',expect_error(Type),[Type]).
recognised_option('-expcterrpos',expect_error_pos(Type,Line,Col),[Type,Line,Col]).
recognised_option('-expecterr',expect_error(Type),[Type]).
recognised_option('-expect',expect_error(Type),[Type]).
recognised_option('-opterr',optional_error(Type),[Type]).
recognised_option('-his',history(File),[File]).
recognised_option('-his_option',history_option(Option),[Option]).
recognised_option('-cache',cache_storage(D),[D]).
recognised_option('-show_cache',show_cache,[]).

recognised_option('-MAIN',csp_main(ProcessName),[ProcessName]).

recognised_option('-ltlfile',ltl_file(Filename),[Filename]).
recognised_option('-ltlformula',ltl_formula_model_check(Formula,_),[Formula]).
recognised_option('-ltlformulat',ltl_formula_model_check(Formula,true),[Formula]).
recognised_option('-ltlformulaf',ltl_formula_model_check(Formula,false),[Formula]).
recognised_option('-ctlformula',ctl_forumla_model_check(Formula,_),[Formula]).
recognised_option('-ctlformulat',ctl_forumla_model_check(Formula,true),[Formula]).
recognised_option('-ctlformulaf',ctl_forumla_model_check(Formula,false),[Formula]).
recognised_option('-load_state',load_state(Filename),[Filename]).
recognised_option('-refchk',refinement_check(Filename,0,100000),[Filename]).

%recognised_option('-cspref',csp_in_situ_refinement_check(assertRef('False',val_of(AbsP1,no_loc_info_available),Type,val_of(ImplP2,no_loc_info_available),no_loc_info_available),'False'),[AbsP1,Type,ImplP2]).
recognised_option('-cspref',csp_in_situ_refinement_check(AbsP1,Type,ImplP2),[AbsP1,Type,ImplP2]).
% -cspref R [F= Q
recognised_option('-cspdeadlock',csp_checkAssertion(Proc,Model,'deadlock free'),[Proc,Model]).
% -cspdeadlock R F
recognised_option('-cspdeterministic',csp_checkAssertion(Proc,Model,'deterministic'),[Proc,Model]).
% -cspdeterministic R F
recognised_option('-csplivelock',csp_checkAssertion(Proc,'FD','livelock free'),[Proc]).
% -csplivelock R
% -csp_assertion "P [F= Q"
recognised_option('-csp_assertion',check_csp_assertion(Assertion),[Assertion]).
recognised_option('-csp_eval', eval_csp_expression(Expr),[Expr]).
recognised_option('-csp_add_unicode', csp_add_unicode(File), [File]).
recognised_option('-csp_remove_unicode', csp_remove_unicode(File), [File]).
recognised_option('-get_csp_assertions_as_string',csp_get_assertions,[]).
recognised_option('-prologOut',csp_translate_to_file(PlFile),[PlFile]).
recognised_option('-csp_guide',add_csp_guide(Filename),[Filename]).
recognised_option('-get_min_max_coverage',get_min_max_coverage(FileName),[FileName]).
recognised_option('-get_coverage_information',get_coverage_information(FileName),[FileName]).
recognised_option('-vacuity_check',vacuity_check,[]).
recognised_option('-cbc',constraint_based_check(OpName),[OpName]).
recognised_option('-cbc_deadlock',cbc_deadlock_check(true),[]).
recognised_option('-cbc_assertions',cbc_assertions(true,[]),[]).
recognised_option('-cbc_main_assertions',cbc_assertions(true,[main_assertions]),[]).
recognised_option('-cbc_assertions_proof',cbc_assertions(false,[]),[]). % do not allow enumeration warnings
recognised_option('-cbc_assertions_tautology_proof',cbc_assertions(false,[tautology_check]),[]). % do not allow enumeration warnings + disregard PROPERTIES
recognised_option('-cbc_assertions_tautology_proof_check',cbc_assertions(false,[tautology_check,contradiction_check]),[]).
recognised_option('-cbc_option',cbc_option(OPT),[OPT]). % should be tautology_check,contradiction_check, unsat_core
recognised_option('-cbc_result_file',cbc_result_file(FILE),[FILE]). % write result to FILE
recognised_option('-cbc_refinement',cbc_refinement,[]).
recognised_option('-cbc_deadlock_pred',cbc_deadlock_check(GoalPred),[GoalPred]).
recognised_option('-cbc_sequence',cbc_sequence(OpSequence,'',single_solution),[OpSequence]).
recognised_option('-cbc_sequence_all',cbc_sequence(OpSequence,'',findall),[OpSequence]).
recognised_option('-cbc_sequence_with_target',cbc_sequence(OpSequence,TargetPredString,single_solution),[OpSequence,TargetPredString]).
recognised_option('-cbc_sequence_with_target_all',cbc_sequence(OpSequence,TargetPredString,findall),[OpSequence,TargetPredString]).
recognised_option('-comment',comment(UserComment),[UserComment]). % not processed by tool, but will be stored in log-file and used by log_analyser
recognised_option('-junit',junit(Dir),[Dir]).
recognised_option('-mcm_cover', mcm_cover(Event),[Event]).
recognised_option('-cbc_cover', cbc_cover(Event),[Event]).
recognised_option('-cbc_cover_match', cbc_cover(match_event(Event)),[Event]). % find events which have Event String occuring somewhere in name
recognised_option('-cbc_cover_all', cbc_cover_all,[]).  % is now default if no cbc_cover provided
recognised_option('-cbc_cover_final', cbc_cover_final,[]).
recognised_option('-bmc', cbc_tests(Depth,'#not_invariant',''),[Depth]).
recognised_option('-dot_output',dot_analyse_output_prefix(Path),[Path]).
recognised_option('-evaldot',evaldot(File),[File]).
recognised_option('-enabling_analysis',enabling_analysis_csv(user_output),[]).
recognised_option('-enabling_analysis_csv',enabling_analysis_csv(EnablingCsvFile),[EnablingCsvFile]).
recognised_option('-feasibility_analysis',feasibility_analysis_csv(1000,user_output),[]).
recognised_option('-read_write_matrix',generate_read_write_matrix_csv(user_output),[]).
recognised_option('-state_trace',state_trace(File),[File]).
recognised_option('-typecheckertest',typechecker_test(File),[File]).
recognised_option('-scc_trace',check_scc_for_ltl_formula(LtlFormula,SCC),[LtlFormula,SCC]).
recognised_option('-selfcheck_module',selfcheck(M),[M]).
recognised_option('-mc_mode',depth_breadth_first_mode(M),[M]). % can be mixed, hash, heuristic
recognised_option('-assertion',cli_check_assertions(specific(X),[false/0,unknown/0]),[X]).
recognised_option('-cbc_assertion',cbc_assertions(true,[specific(X)]),[X]). % check only a specific assertion
recognised_option('-symbolic_model_check', cli_symbolic_model_check(Algorithm), [Algorithm]).
recognised_option('-replay',eval_repl([File]),[File]). % used to be -eval
recognised_option('-ltsmin2',ltsmin2(EndpointPath), [EndpointPath]).
recognised_option('-ltsmin_ltl_output',ltsmin_ltl_output(Path), [Path]).
recognised_option('-ltsmin_option', ltsmin_option(X),[X]).

recognised_option('-dot_all',dot_generate_for_all_formulas). % generate dot also for true formulas
recognised_option('-animate_all',cli_random_animate(2147483647,false)).
recognised_option('-execute_all',execute(2147483647,false,current_state)).
recognised_option('-execute_all_inits',execute(2147483647,false,from_all_initial_states)).
recognised_option('-animate_stats',animate_stats).
recognised_option('-check_goal',check_goal).
recognised_option('-ltlassertions',ltl_assertions).
recognised_option('-assertions',cli_check_assertions(all,[false/0,unknown/0])).
recognised_option('-main_assertions',cli_check_assertions(main,[false/0,unknown/0])).
recognised_option('-properties',cli_check_properties).
recognised_option('-selfcheck',selfcheck(_)).
recognised_option('-pacheck',pa_check).
recognised_option('-det_check',det_check). % check if animation is deterministic
recognised_option('-det_constants',det_constants_check). % check if animation for setup_constants is deterministic
recognised_option('-bf',breadth_first).
recognised_option('-breadth',breadth_first).
recognised_option('-df',depth_first).
recognised_option('-depth',depth_first).
recognised_option('-strict',strict_raise_error).
recognised_option('-silent',silent).
recognised_option('-statistics',statistics).
recognised_option('-profile_statistics',profile_statistics).
recognised_option('-reset_profile_statistics',reset_profiler). % mainly for use in REPL
recognised_option('-nodead',no_deadlocks).
recognised_option('-noinv',no_invariant_violations).
recognised_option('-nogoal',no_goal).
recognised_option('-noltl',no_ltl). % just used for TLC at the moment
recognised_option('-noass',no_assertion_violations).
recognised_option('-nocounter',no_counter_examples).
recognised_option('-disable_time_out',set_preference_group(time_out,disable_time_out)).
recognised_option('-disable_timeout',set_preference_group(time_out,disable_time_out)).
%recognised_option('-POR',with_reduction).
recognised_option('-i',animate).
recognised_option('-repl',eval_repl([])). % used to be -eval
recognised_option('-c',coverage(false)).
recognised_option('-cs',coverage(just_summary)).
recognised_option('-coverage',coverage(false)).
recognised_option('-coverage_summary',coverage(just_summary)).
recognised_option('-machine_stats',cli_print_machine_statistics).
recognised_option('-vc',variable_coverage).
recognised_option('-variable_coverage',variable_coverage).
recognised_option('-cv',coverage(true)).
recognised_option('-v',verbose).
recognised_option('-vv',very_verbose).
recognised_option('-verbose',verbose).
recognised_option('-very_verbose',very_verbose).
recognised_option('-profiling_on',profiling_on). % Prolog profiling
recognised_option('-profile',prob_profile). % ProB Operation profiling
recognised_option('-version',print_version).
recognised_option('-svers',print_short_version).
recognised_option('-short_version',print_short_version).
recognised_option('-check_java_version',check_java_version).
recognised_option('-file_info',file_info).
recognised_option('-t',trace_check).
recognised_option('-init',initialise).
recognised_option('-initialise',initialise).
recognised_option('-ll',log('/tmp/prob_cli_debug.log',prolog)).
recognised_option('-ss',socket(9000)).
recognised_option('-sf',socket(_)).
recognised_option('-help',help).
recognised_option('-h',help).
recognised_option('-rc',runtimechecking).
recognised_option('-test_mode',test_mode).
recognised_option('-check_complete',check_complete).
recognised_option('-check_complete_operation_coverage', check_complete_operation_coverage).
recognised_option('-mc_with_tlc', cli_start_mc_with_tlc).
recognised_option('-mc_with_lts_sym', cli_start_sym_mc_with_lts(symbolic)).
recognised_option('-mc_with_lts_seq', cli_start_sym_mc_with_lts(sequential)).
%recognised_option('-csp',csp_pl_mode).
%recognised_option('-cspm',csp_mode).
recognised_option('-core',disprover_options([disprover_option(unsat_core),unsat_core_algorithm/linear])).

recognised_option('-zmq_worker',zmq_worker).
recognised_option('-ltsmin',ltsmin).

% some utilities to be able to call the above options directly from repl:
:- public variable_coverage/0, silent/0, coverage/1, help/0.
verbose :- tcltk_turn_debugging_on(19).
very_verbose :- tcltk_turn_debugging_on(5).
file_info :- file_loaded(true,MainFile), print_file_info(MainFile).
coverage(ShowEnabledInfo) :- probcli_time_stamp(NOW), cli_show_coverage(ShowEnabledInfo,NOW).
variable_coverage :- probcli_time_stamp(NOW), cli_show_variable_coverage(NOW).
silent :- (option(silent) -> true ; assert_option(silent)).
help :- eval_help.
dot_command(DCommand,DotFile) :- call_dot_command(DCommand,DotFile,[]).
dot_command_for_expr(DECommand,Expr,DotFile,Opts) :- call_dot_command_for_expr(DECommand,Expr,DotFile,Opts).

option_verbose :- (option(verbose) -> true ; option(very_verbose)).

% new profiler
%:- use_module(('../extensions/profiler/profiler.pl').
%prob_profiler :- pen.

% old profiler
:- use_module(runtime_profiler,[print_runtime_profile/0]).
prob_profile :- print_runtime_profile.

% utilities for command-line arguments
% convert a string of the form "+x,-y,z" to a list [evaluated(x),not_evaluated(y),evaluated(z)]
evalchecklist(Arg,List) :-
    split_atom(Arg,[',',' '],AList),
    evalchecklist2(AList,List).
evalchecklist2([],[]).
evalchecklist2([Atom|Arest],[Term|Trest]) :-
    atom_chars(Atom,Chars),Chars=[C|Crest],
    evalchecklist3(C,Crest,Term),
    evalchecklist2(Arest,Trest).
evalchecklist3(+,Crest,evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(-,Crest,not_evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(C,Crest,evaluated(Pred)) :- atom_chars(Pred,[C|Crest]).

set_random_seed_to_deterministic_start_seed :-
     % in test_mode we do not change the random number generator's initial seed
     true. %getrand(CurrState),setrand(CurrState). % this seems to be a no-op

set_new_random_seed :-
   now(TimeStamp), % getting the unix time
   setrand(TimeStamp). % setting new random seed by every execution of probcli

halt_exception :- halt_exception(0).
halt_exception(Code) :- throw(halt(Code)).

% -----------------

start_xml_feature(FeatureName,[CErrs1,CWarns1]) :-
    debug_format(20,'% Starting ~w~n',[FeatureName]),
    get_counter(cli_errors,CErrs1), get_counter(cli_warnings,CWarns1),
    start_xml_group_in_log(FeatureName).

start_xml_feature(FeatureName,Attr,Value,[CErrs1,CWarns1]) :-
    debug_format(20,'% Starting ~w (~w=~w)~n',[FeatureName,Attr,Value]),
    get_counter(cli_errors,CErrs1), get_counter(cli_warnings,CWarns1),
    start_xml_group_in_log(FeatureName,Attr,Value).

stop_xml_feature(FeatureName,[CErrs1,CWarns1]) :-
	get_counter(cli_errors,CErrs2), get_counter(cli_warnings,CWarns2),
	CErrs is CErrs2-CErrs1, CWarns is CWarns2-CWarns1,
	write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns]),
    debug_format(20,'% Finished ~w (errors=~w, warnings=~w)~n',[FeatureName,CErrs,CWarns]),
	stop_xml_group_in_log(FeatureName).
	%(CErrs>0 -> (file_loaded(_,MainFile) -> true ; MainFile=unknown), Time=unknown, % TO DO: determine time
    %   create_and_print_junit_result(FeatureName,'Feature', MainFile, Time, error) ; true).
% Note: call stop_xml_group_in_log if the feature stops unexpectedly and you do not have the Info list available

% -----------------

runtime_entry(start) :- go_cli.

save :- save_program('probcli.sav').

:- announce_event(compile_prob).
