:- module(synthesis_tests,[run_tests/0,run_tests/1]).

:- ensure_loaded('../../src/prob_cli').
:- use_module(library(plunit)).
:- use_module(probsrc(solver_interface),[solve_predicate/3]).
:- use_module(probsrc(b_interpreter)).
:- use_module(probsrc(bmachine),[b_load_machine_from_file/1,b_machine_precompile/0]).
:- use_module(synthesis(b_synthesis)).
:- use_module(synthesis(library_setup)).
:- use_module(synthesis(synthesis_util)).

% run from prob_prolog root folder

:-  set_search_pathes,
    init_eclipse_preferences,
    external_functions:reset_argv,
    set_prefs , 
    logging:disable_logging , 
    preferences:set_preference(randomise_enumeration_order,true).

:- begin_tests(location_vars_to_program).

test(lVar2Prog_Guard1) :- 
    b_synthesis:location_vars_to_program(([card-set(integer),constant-set(integer),constant-integer,greater],[]),guard,
            [b(identifier(var),set(integer),[])],
            [b(identifier(l0),set(integer),[synthesis(input(0)),synthesis(machinevar,var)]),b(identifier(li68),integer,[synthesis(card1,input),synthesis(type,set(integer))]),b(identifier(li74),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo69),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo70),integer,[synthesis(constant71,output),synthesis(type,set(integer))]),b(identifier(lo72),integer,[synthesis(constant73,output),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(greater1,output),synthesis(type,boolean)])],
            4,1,
            solution([binding(constant71,[],{}),binding(constant73,int(1),'1'),binding(l0set,int(0),'0'),binding(li68,int(0),'0'),binding(li74,int(3),'3'),binding(li75,int(2),'2'),binding(lo0boolean_false,int(4),'4'),binding(lo0boolean_true,int(4),'4'),binding(lo69,int(3),'3'),binding(lo70,int(1),'1'),binding(lo72,int(2),'2'),binding(lo76,int(4),'4'),binding(o72,int(-10),'-10')]),
            [b(identifier(lo76),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo69),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo72),integer,[synthesis(constant73,output),synthesis(type,integer)])],
            ProgramAST) , 
    ProgramAST = b(greater(b(card(b(identifier(var),set(integer),_)),integer,_),b(value(int(1)),integer,_)),boolean,_).

test(lVar2Prog_Guard2) :- 
    b_synthesis:location_vars_to_program(([greater,equal-integer,constant-integer],[]),guard,
            [b(identifier(var),integer,[])],
            [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,var)]),b(identifier(li86),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li87),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li89),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li90),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo88),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo91),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo92),integer,[synthesis(constant93,output),synthesis(type,integer)])],
            3,1,
            solution([binding(constant93,int(0),'0'),binding(l00,int(0),'0'),binding(l01,int(0),'0'),binding(li86,int(0),'0'),binding(li87,int(1),'1'),binding(li89,int(0),'0'),binding(li90,int(0),'0'),binding(lo0boolean_false,int(3),'3'),binding(lo0boolean_true,int(3),'3'),binding(lo88,int(3),'3'),binding(lo91,int(2),'2'),binding(lo92,int(1),'1'),binding(o92,int(-10),'-10')]),
            [b(identifier(lo88),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo92),integer,[synthesis(constant93,output),synthesis(type,integer)])],
            ProgramAST) , 
    ProgramAST = b(greater(b(identifier(var),integer,_),b(value(int(0)),integer,_)),boolean,_).

test(lVar2Prog_Guard3) :- 
    b_synthesis:location_vars_to_program(([greater,equal-integer,constant-integer],[add,minus]),guard,
        [b(identifier(var1),integer,[]),b(identifier(var2),integer,[])],
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,var1)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,var2)]),b(identifier(li448),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li449),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li451),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li452),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li456),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li457),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li459),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li460),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo450),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo453),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo454),integer,[synthesis(constant455,output),synthesis(type,integer)]),b(identifier(lo458),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo461),integer,[synthesis(minus1,output),synthesis(type,integer)])],
        6,2,
        solution([binding(constant455,int(1),'1'),binding('l0-1',int(0),'0'),binding(l00,int(0),'0'),binding(l01,int(0),'0'),binding(l02,int(0),'0'),binding(l10,int(1),'1'),binding(l11,int(1),'1'),binding(l12,int(1),'1'),binding(li448,int(3),'3'),binding(li449,int(4),'4'),binding(li451,int(0),'0'),binding(li452,int(0),'0'),binding(li456,int(1),'1'),binding(li457,int(0),'0'),binding(li459,int(0),'0'),binding(li460,int(0),'0'),binding(lo0boolean_false,int(6),'6'),binding(lo0boolean_true,int(6),'6'),binding(lo450,int(6),'6'),binding(lo453,int(5),'5'),binding(lo454,int(4),'4'),binding(lo458,int(3),'3'),binding(lo461,int(2),'2'),binding(o454,int(-10),'-10')]),
        [b(identifier(lo450),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo458),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo454),integer,[synthesis(constant455,output),synthesis(type,integer)])],
        ProgramAST) , 
    ProgramAST = b(greater(b(add(b(identifier(var2),integer,_),b(identifier(var1),integer,_)),integer,_),b(value(int(1)),integer,_)),boolean,_).

test(lVar2Prog_Guard4) :- 
    b_synthesis:location_vars_to_program(([greater,equal-integer,constant-integer,greater,equal-integer,constant-integer],[add,minus]),guard,
        [b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(li100),integer,[synthesis(type,integer),synthesis(greater2,right_input)]),b(identifier(li102),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li103),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li107),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li108),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li110),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li111),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li91),integer,[synthesis(type,integer),synthesis(greater1,left_input)]),b(identifier(li92),integer,[synthesis(type,integer),synthesis(greater1,right_input)]),b(identifier(li94),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li95),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li99),integer,[synthesis(type,integer),synthesis(greater2,left_input)]),b(identifier(lo101),integer,[synthesis(type,boolean),synthesis(greater2,output)]),b(identifier(lo104),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo106),integer,[synthesis(type,integer),synthesis(constant105,output),synthesis(type,constant-integer)]),b(identifier(lo109),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo112),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo93),integer,[synthesis(type,boolean),synthesis(greater1,output)]),b(identifier(lo96),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo98),integer,[synthesis(type,integer),synthesis(constant97,output),synthesis(type,constant-integer)])],
        9,2,
        solution([binding(constant105,int(0),'0'),binding(constant97,int(2),'2'),binding(l02,int(0),'0'),binding(l03,int(0),'0'),binding(l10,int(1),'1'),binding(l13,int(1),'1'),binding(li100,int(0),'0'),binding(li102,int(5),'5'),binding(li103,int(0),'0'),binding(li107,int(0),'0'),binding(li108,int(0),'0'),binding(li110,int(0),'0'),binding(li111,int(0),'0'),binding(li91,int(0),'0'),binding(li92,int(0),'0'),binding(li94,int(0),'0'),binding(li95,int(0),'0'),binding(li99,int(0),'0'),binding(lo0boolean_false,int(9),'9'),binding(lo0boolean_true,int(9),'9'),binding(lo101,int(4),'4'),binding(lo104,int(9),'9'),binding(lo106,int(6),'6'),binding(lo109,int(3),'3'),binding(lo112,int(2),'2'),binding(lo93,int(8),'8'),binding(lo96,int(7),'7'),binding(lo98,int(5),'5')]),
        [b(identifier(lo104),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo98),integer,[synthesis(type,integer),synthesis(constant97,output),synthesis(type,constant-integer)])],
        ProgramAST) , 
    ProgramAST = b(equal(b(value(int(2)),integer,_),b(identifier(coins),integer,_)),boolean,_).

test(lVar2Prog_Invariant1) :- 
    b_synthesis:location_vars_to_program(([greater,equal-integer,constant-integer],[]),invariant,
            [b(identifier(var),integer,[])],
            [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,var)]),b(identifier(li102),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li103),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li105),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li106),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo104),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo107),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo108),integer,[synthesis(constant109,output),synthesis(type,integer)])],
            3,1,
            solution([binding(constant109,int(-1),'-1'),binding('l0-1',int(0),'0'),binding('l0-2',int(0),'0'),binding(l00,int(0),'0'),binding(l01,int(0),'0'),binding(l02,int(0),'0'),binding(l03,int(0),'0'),binding(l04,int(0),'0'),binding(l05,int(0),'0'),binding(li102,int(0),'0'),binding(li103,int(1),'1'),binding(li105,int(0),'0'),binding(li106,int(0),'0'),binding(lo0boolean_false,int(3),'3'),binding(lo0boolean_true,int(3),'3'),binding(lo104,int(3),'3'),binding(lo107,int(2),'2'),binding(lo108,int(1),'1'),binding(o108,int(-10),'-10')]),
            [b(identifier(lo104),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo108),integer,[synthesis(constant109,output),synthesis(type,integer)])],
            ProgramAST) , 
    ProgramAST = b(greater(b(identifier(var),integer,_),b(value(int(-1)),integer,_)),boolean,_).

test(lVar2Prog_Action1) :- 
    b_synthesis:location_vars_to_program(([equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,multiplication,div,add,minus,multiplication,div]),action,
        [b(identifier(coins),integer,[]),b(identifier(soda),integer,[])],
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(li183),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li184),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li188),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li189),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li193),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li194),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li196),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li197),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li199),integer,[synthesis(type,integer),synthesis(multiplication1,left_input)]),b(identifier(li200),integer,[synthesis(type,integer),synthesis(multiplication1,right_input)]),b(identifier(li202),integer,[synthesis(type,integer),synthesis(div1,left_input)]),b(identifier(li203),integer,[synthesis(type,integer),synthesis(div1,right_input)]),b(identifier(li205),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li206),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li208),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li209),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li211),integer,[synthesis(type,integer),synthesis(multiplication2,left_input)]),b(identifier(li212),integer,[synthesis(type,integer),synthesis(multiplication2,right_input)]),b(identifier(li214),integer,[synthesis(type,integer),synthesis(div2,left_input)]),b(identifier(li215),integer,[synthesis(type,integer),synthesis(div2,right_input)]),b(identifier(lo185),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo187),integer,[synthesis(type,integer),synthesis(constant186,output),synthesis(type,constant-integer)]),b(identifier(lo190),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo192),integer,[synthesis(type,integer),synthesis(constant191,output),synthesis(type,constant-integer)]),b(identifier(lo195),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo198),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo201),integer,[synthesis(type,integer),synthesis(multiplication1,output)]),b(identifier(lo204),integer,[synthesis(type,integer),synthesis(div1,output)]),b(identifier(lo207),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo210),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo213),integer,[synthesis(type,integer),synthesis(multiplication2,output)]),b(identifier(lo216),integer,[synthesis(type,integer),synthesis(div2,output)])],
        13,2,
        solution([binding(constant186,int(2),'2'),binding(constant191,int(3),'3'),binding(l02,int(0),'0'),binding(l04,int(0),'0'),binding(l05,int(0),'0'),binding(l06,int(0),'0'),binding(l13,int(1),'1'),binding(l14,int(1),'1'),binding(l15,int(1),'1'),binding(l19,int(1),'1'),binding(li183,int(5),'5'),binding(li184,int(0),'0'),binding(li188,int(0),'0'),binding(li189,int(0),'0'),binding(li193,int(0),'0'),binding(li194,int(0),'0'),binding(li196,int(1),'1'),binding(li197,int(9),'9'),binding(li199,int(0),'0'),binding(li200,int(0),'0'),binding(li202,int(0),'0'),binding(li203,int(0),'0'),binding(li205,int(0),'0'),binding(li206,int(0),'0'),binding(li208,int(0),'0'),binding(li209,int(10),'10'),binding(li211,int(0),'0'),binding(li212,int(0),'0'),binding(li214,int(0),'0'),binding(li215,int(0),'0'),binding(lo00,int(13),'13'),binding(lo02,int(13),'13'),binding(lo03,int(13),'13'),binding(lo04,int(13),'13'),binding(lo10,int(12),'12'),binding(lo11,int(12),'12'),binding(lo12,int(12),'12'),binding(lo16,int(12),'12'),binding(lo185,int(6),'6'),binding(lo187,int(10),'10'),binding(lo190,int(2),'2'),binding(lo192,int(9),'9'),binding(lo195,int(5),'5'),binding(lo198,int(12),'12'),binding(lo201,int(11),'11'),binding(lo204,int(4),'4'),binding(lo207,int(8),'8'),binding(lo210,int(13),'13'),binding(lo213,int(7),'7'),binding(lo216,int(3),'3')]),
        [b(identifier(lo187),integer,[synthesis(type,integer),synthesis(constant186,output),synthesis(type,constant-integer)]),b(identifier(lo192),integer,[synthesis(type,integer),synthesis(constant191,output),synthesis(type,constant-integer)]),b(identifier(lo198),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo210),integer,[synthesis(type,integer),synthesis(minus2,output)])],
        ProgramAST) , 
    ProgramAST = b(conjunct(b(equal(b(identifier('coins\''),integer,_),b(minus(b(identifier(coins),integer,_),b(value(int(2)),integer,_)),integer,_)),pred,_),b(equal(b(identifier('soda\''),integer,_),b(minus(b(identifier(soda),integer,_),b(value(int(3)),integer,_)),integer,_)),pred,_)),pred,_).

test(lVar2Prog_Action2) :- 
    b_synthesis:location_vars_to_program(([equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,multiplication,div,add,minus,multiplication,div]),action,
        [b(identifier(coins),integer,[]),b(identifier(soda),integer,[])],
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(li273),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li274),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li278),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li279),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li283),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li284),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li286),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li287),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li289),integer,[synthesis(type,integer),synthesis(multiplication1,left_input)]),b(identifier(li290),integer,[synthesis(type,integer),synthesis(multiplication1,right_input)]),b(identifier(li292),integer,[synthesis(type,integer),synthesis(div1,left_input)]),b(identifier(li293),integer,[synthesis(type,integer),synthesis(div1,right_input)]),b(identifier(li295),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li296),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li298),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li299),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li301),integer,[synthesis(type,integer),synthesis(multiplication2,left_input)]),b(identifier(li302),integer,[synthesis(type,integer),synthesis(multiplication2,right_input)]),b(identifier(li304),integer,[synthesis(type,integer),synthesis(div2,left_input)]),b(identifier(li305),integer,[synthesis(type,integer),synthesis(div2,right_input)]),b(identifier(lo275),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo277),integer,[synthesis(type,integer),synthesis(constant276,output),synthesis(type,constant-integer)]),b(identifier(lo280),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo282),integer,[synthesis(type,integer),synthesis(constant281,output),synthesis(type,constant-integer)]),b(identifier(lo285),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo288),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo291),integer,[synthesis(type,integer),synthesis(multiplication1,output)]),b(identifier(lo294),integer,[synthesis(type,integer),synthesis(div1,output)]),b(identifier(lo297),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo300),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo303),integer,[synthesis(type,integer),synthesis(multiplication2,output)]),b(identifier(lo306),integer,[synthesis(type,integer),synthesis(div2,output)])],
        13,2,
        solution([binding(constant276,int(-2),'-2'),binding(constant281,int(3),'3'),binding(l02,int(0),'0'),binding(l04,int(0),'0'),binding(l13,int(1),'1'),binding(l14,int(1),'1'),binding(li273,int(0),'0'),binding(li274,int(0),'0'),binding(li278,int(0),'0'),binding(li279,int(0),'0'),binding(li283,int(11),'11'),binding(li284,int(0),'0'),binding(li286,int(1),'1'),binding(li287,int(10),'10'),binding(li289,int(0),'0'),binding(li290,int(0),'0'),binding(li292,int(0),'0'),binding(li293,int(0),'0'),binding(li295,int(0),'0'),binding(li296,int(0),'0'),binding(li298,int(0),'0'),binding(li299,int(0),'0'),binding(li301,int(0),'0'),binding(li302,int(0),'0'),binding(li304,int(0),'0'),binding(li305,int(0),'0'),binding(lo00,int(13),'13'),binding(lo02,int(13),'13'),binding(lo10,int(12),'12'),binding(lo11,int(12),'12'),binding(lo275,int(5),'5'),binding(lo277,int(11),'11'),binding(lo280,int(4),'4'),binding(lo282,int(10),'10'),binding(lo285,int(13),'13'),binding(lo288,int(12),'12'),binding(lo291,int(9),'9'),binding(lo294,int(3),'3'),binding(lo297,int(8),'8'),binding(lo300,int(7),'7'),binding(lo303,int(6),'6'),binding(lo306,int(2),'2')]),
        [b(identifier(lo277),integer,[synthesis(type,integer),synthesis(constant276,output),synthesis(type,constant-integer)]),b(identifier(lo282),integer,[synthesis(type,integer),synthesis(constant281,output),synthesis(type,constant-integer)]),b(identifier(lo285),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo288),integer,[synthesis(type,integer),synthesis(minus1,output)])],
        ProgramAST) , 
    ProgramAST = b(conjunct(b(equal(b(identifier('coins\''),integer,_),b(add(b(value(int(-2)),integer,_),b(identifier(coins),integer,_)),integer,_)),pred,_),b(equal(b(identifier('soda\''),integer,_),b(minus(b(identifier(soda),integer,_),b(value(int(3)),integer,_)),integer,_)),pred,_)),pred,_).

test(lVar2Prog_Action3) :- 
    b_synthesis:location_vars_to_program(([equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,multiplication,div,add,minus,multiplication,div]),action,
        [b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(li397),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li398),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li402),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li403),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li407),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li408),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li410),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li411),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li413),integer,[synthesis(type,integer),synthesis(multiplication1,left_input)]),b(identifier(li414),integer,[synthesis(type,integer),synthesis(multiplication1,right_input)]),b(identifier(li416),integer,[synthesis(type,integer),synthesis(div1,left_input)]),b(identifier(li417),integer,[synthesis(type,integer),synthesis(div1,right_input)]),b(identifier(li419),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li420),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li422),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li423),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li425),integer,[synthesis(type,integer),synthesis(multiplication2,left_input)]),b(identifier(li426),integer,[synthesis(type,integer),synthesis(multiplication2,right_input)]),b(identifier(li428),integer,[synthesis(type,integer),synthesis(div2,left_input)]),b(identifier(li429),integer,[synthesis(type,integer),synthesis(div2,right_input)]),b(identifier(lo399),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo401),integer,[synthesis(type,integer),synthesis(constant400,output),synthesis(type,constant-integer)]),b(identifier(lo404),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo406),integer,[synthesis(type,integer),synthesis(constant405,output),synthesis(type,constant-integer)]),b(identifier(lo409),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo412),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo415),integer,[synthesis(type,integer),synthesis(multiplication1,output)]),b(identifier(lo418),integer,[synthesis(type,integer),synthesis(div1,output)]),b(identifier(lo421),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo424),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo427),integer,[synthesis(type,integer),synthesis(multiplication2,output)]),b(identifier(lo430),integer,[synthesis(type,integer),synthesis(div2,output)])],
        13,2,
        solution([binding(constant400,int(3),'3'),binding(constant405,int(0),'0'),binding(l02,int(0),'0'),binding(l13,int(1),'1'),binding(l14,int(1),'1'),binding(li397,int(5),'5'),binding(li398,int(0),'0'),binding(li402,int(0),'0'),binding(li403,int(0),'0'),binding(li407,int(0),'0'),binding(li408,int(0),'0'),binding(li410,int(1),'1'),binding(li411,int(7),'7'),binding(li413,int(0),'0'),binding(li414,int(0),'0'),binding(li416,int(0),'0'),binding(li417,int(0),'0'),binding(li419,int(0),'0'),binding(li420,int(0),'0'),binding(li422,int(0),'0'),binding(li423,int(0),'0'),binding(li425,int(0),'0'),binding(li426,int(0),'0'),binding(li428,int(0),'0'),binding(li429,int(0),'0'),binding(lo00,int(13),'13'),binding(lo10,int(12),'12'),binding(lo11,int(12),'12'),binding(lo399,int(6),'6'),binding(lo401,int(7),'7'),binding(lo404,int(2),'2'),binding(lo406,int(13),'13'),binding(lo409,int(5),'5'),binding(lo412,int(12),'12'),binding(lo415,int(11),'11'),binding(lo418,int(4),'4'),binding(lo421,int(10),'10'),binding(lo424,int(9),'9'),binding(lo427,int(8),'8'),binding(lo430,int(3),'3')]),
        [b(identifier(lo401),integer,[synthesis(type,integer),synthesis(constant400,output),synthesis(type,constant-integer)]),b(identifier(lo406),integer,[synthesis(type,integer),synthesis(constant405,output),synthesis(type,constant-integer)]),b(identifier(lo412),integer,[synthesis(type,integer),synthesis(minus1,output)])],
        ProgramAST) , 
    ProgramAST = b(conjunct(b(equal(b(identifier('coins\''),integer,_),b(value(int(0)),integer,_)),pred,_),b(equal(b(identifier('soda\''),integer,[nodeid(pos(6,1,2,18,2,21))]),b(minus(b(identifier(soda),integer,_),b(value(int(3)),integer,_)),integer,_)),pred,_)),pred,_).

:- end_tests(location_vars_to_program).

:- begin_tests(synthesis).

% precompile the machines since synthesis_from_examples uses predicates like b_get_machine_variables/1

test(synthesis_1,[nondet]) :- 
    b_load_machine_from_file('../prob_examples/public_examples/Synthesis_Tests/lift2.mch') , b_machine_precompile , 
    preferences:set_preference(time_out,2500) , 
    library_setup(guard,1,[b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(machinevar,ee)])],Library) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,[b(identifier(ee),set(couple(integer,integer)),[nodeid(pos(15,1,5,2,5,3)),loc(local,test_seq_mch,abstract_variables)])],[b(identifier(ee),set(couple(integer,integer)),[nodeid(pos(15,1,5,2,5,3)),loc(local,test_seq_mch,abstract_variables)])],[b(subset(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(33,1,9,29,9,31)),readonly]),b(identifier(ee),set(couple(integer,integer)),[nodeid(pos(34,1,9,36,9,37)),loc(local,test_seq_mch,abstract_variables)])),pred,[nodeid(pos(32,1,9,29,9,37))]),b(disjunct(b(equal(b(card(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(38,1,9,48,9,50)),readonly])),integer,[contains_wd_condition,nodeid(pos(37,1,9,43,9,51))]),b(integer(1),integer,[nodeid(pos(39,1,9,55,9,55))])),pred,[contains_wd_condition,nodeid(pos(36,1,9,43,9,55))]),b(equal(b(card(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(42,1,9,65,9,67)),readonly])),integer,[contains_wd_condition,nodeid(pos(41,1,9,60,9,68))]),b(integer(0),integer,[nodeid(pos(43,1,9,72,9,72))])),pred,[contains_wd_condition,nodeid(pos(40,1,9,60,9,72))])),pred,[contains_wd_condition,nodeid(pos(35,1,9,43,9,72))])],guard,[[b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(machinevar,ee)])],[b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(machinevar,ee)])],[b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(machinevar,ee)])]],[[b(value(avl_set(node((int(1),int(2)),true,0,empty,empty))),set(couple(integer,integer)),[synthesis(machinevar,ee)])]],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(o5),integer,[synthesis(constant6,output)]),b(interval(b(integer(-10),integer,[]),b(integer(11),integer,[])),set(integer),[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(interval(b(integer(0),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(interval(b(integer(1),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(interval(b(integer(1),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(interval(b(integer(1),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(interval(b(integer(1),integer,[]),b(integer(4),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[]),b(less(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o9),boolean,[synthesis(greater1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(4),integer,[])),pred,[]),b(equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equal(b(card(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])),integer,[]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[]),b(equal(b(identifier(constant4),set(couple(integer,integer)),[]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[]),b(equal(b(identifier(constant6),integer,[]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(greater(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(member(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant4,constant6,i1,i7,i8,l0set,li1,li7,li8,lo0boolean_true,lo2,lo3,lo5,lo9,o2,o3,o5,o9])])),pred,[]),b(exists([b(identifier(o9),boolean,[synthesis(greater1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(4),integer,[])),pred,[]),b(equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equal(b(card(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])),integer,[]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[]),b(equal(b(identifier(constant4),set(couple(integer,integer)),[]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[]),b(equal(b(identifier(constant6),integer,[]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(greater(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(member(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant4,constant6,i1,i7,i8,l0set,li1,li7,li8,lo0boolean_true,lo2,lo3,lo5,lo9,o2,o3,o5,o9])])),pred,[]),b(exists([b(identifier(o9),boolean,[synthesis(greater1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(4),integer,[])),pred,[]),b(equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equal(b(card(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])),integer,[]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[]),b(equal(b(identifier(constant4),set(couple(integer,integer)),[]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[]),b(equal(b(identifier(constant6),integer,[]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(greater(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(value(avl_set(node((int(0),int(1)),true,1,empty,node((int(1),int(2)),true,0,empty,empty)))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(member(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant4,constant6,i1,i7,i8,l0set,li1,li7,li8,lo0boolean_true,lo2,lo3,lo5,lo9,o2,o3,o5,o9])])),pred,[]),b(exists([b(identifier(o9),boolean,[synthesis(greater1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(4),integer,[])),pred,[]),b(equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equal(b(card(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)])),integer,[]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[]),b(equal(b(identifier(constant4),set(couple(integer,integer)),[]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[]),b(equal(b(identifier(constant6),integer,[]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(greater(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(i1),set(couple(integer,integer)),[synthesis(card1,input)]),b(value(avl_set(node((int(1),int(2)),true,0,empty,empty))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(i8),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o2),integer,[synthesis(card1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o2),integer,[synthesis(card1,output)]),b(identifier(o5),integer,[synthesis(constant6,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[]),b(equal(b(identifier(o3),set(couple(integer,integer)),[synthesis(constant4,output)]),b(value(avl_set(node((int(1),int(2)),true,0,empty,empty))),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(not_equal(b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o9),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l0set),integer,[synthesis(input(0)),synthesis(machinevar,ee),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(member(b(identifier(i7),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i8),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant4,constant6,i1,i7,i8,l0set,li1,li7,li8,lo0boolean_false,lo2,lo3,lo5,lo9,o2,o3,o5,o9])])),pred,[]),
        [b(identifier(l0),set(couple(integer,integer)),[synthesis(input(0)),synthesis(machinevar,ee)]),b(identifier(li1),integer,[synthesis(card1,input),synthesis(type,set(couple(integer,integer)))]),b(identifier(li7),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li8),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo2),integer,[synthesis(card1,output),synthesis(type,integer)]),b(identifier(lo3),integer,[synthesis(constant4,output),synthesis(type,set(couple(integer,integer)))]),b(identifier(lo5),integer,[synthesis(constant6,output),synthesis(type,integer)]),b(identifier(lo9),integer,[synthesis(greater1,output),synthesis(type,boolean)])],4,1,solution([binding(constant4,[],{}),binding(constant6,int(1),'1'),binding(l0set,int(0),'0'),binding(li1,int(0),'0'),binding(li7,int(3),'3'),binding(li8,int(2),'2'),binding(lo0boolean_false,int(4),'4'),binding(lo0boolean_true,int(4),'4'),binding(lo2,int(3),'3'),binding(lo3,int(1),'1'),binding(lo5,int(2),'2'),binding(lo9,int(4),'4'),binding(o5,int(-10),'-10')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    b_test_boolean_expression_wf(b(equivalence(Synthesized,b(conjunct(b(conjunct(b(greater(b(card(b(identifier(ee),set(couple(integer,integer)),[])),set(couple(integer,integer)),[]),b(integer(1),integer,[])),boolean,[]),b(subset(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(33,1,9,29,9,31)),readonly]),b(identifier(ee),set(couple(integer,integer)),[nodeid(pos(34,1,9,36,9,37)),loc(local,test_seq_mch,abstract_variables)])),pred,[nodeid(pos(32,1,9,29,9,37))])),pred,[]),b(disjunct(b(equal(b(card(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(38,1,9,48,9,50)),readonly])),integer,[contains_wd_condition,nodeid(pos(37,1,9,43,9,51))]),b(integer(1),integer,[nodeid(pos(39,1,9,55,9,55))])),pred,[contains_wd_condition,nodeid(pos(36,1,9,43,9,55))]),b(equal(b(card(b(identifier(sub),set(couple(integer,integer)),[nodeid(pos(42,1,9,65,9,67)),readonly])),integer,[contains_wd_condition,nodeid(pos(41,1,9,60,9,68))]),b(integer(0),integer,[nodeid(pos(43,1,9,72,9,72))])),pred,[contains_wd_condition,nodeid(pos(40,1,9,60,9,72))])),pred,[contains_wd_condition,nodeid(pos(35,1,9,43,9,72))])),pred,[contains_wd_condition])),pred,[]),[],_).
    
test(synthesis_2,[nondet]) :- 
    preferences:set_preference(time_out,2500) , 
    library_setup(guard,1,[b(integer(-1),integer,[synthesis(machinevar,c)]),b(integer(3),integer,[synthesis(machinevar,d)])],Library) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,[b(identifier(c),integer,[nodeid(pos(18,1,4,30,4,30)),loc(local,lift,abstract_variables)]),b(identifier(d),integer,[nodeid(pos(19,1,4,34,4,34)),loc(local,lift,abstract_variables)])],[b(identifier(c),integer,[nodeid(pos(18,1,4,30,4,30)),loc(local,lift,abstract_variables)]),b(identifier(d),integer,[nodeid(pos(19,1,4,34,4,34)),loc(local,lift,abstract_variables)])],any,guard,[[b(integer(-1),integer,[synthesis(machinevar,c)]),b(integer(3),integer,[synthesis(machinevar,d)])],[b(integer(4),integer,[synthesis(machinevar,c)]),b(integer(-2),integer,[synthesis(machinevar,d)])],[b(integer(0),integer,[synthesis(machinevar,c)]),b(integer(3),integer,[synthesis(machinevar,d)])]],[[b(integer(0),integer,[synthesis(machinevar,c)]),b(integer(1),integer,[synthesis(machinevar,d)])]],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(o77),integer,[synthesis(constant78,output)]),b(interval(b(integer(-10),integer,[]),b(integer(11),integer,[])),set(integer),[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(interval(b(integer(2),integer,[]),b(integer(6),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(less(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[]),b(less(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(less(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(less(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(less(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o84),integer,[synthesis(minus1,output)]),b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)]),b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i71),integer,[synthesis(greater1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(6),integer,[])),pred,[]),b(conjunct(b(equal(b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant78),integer,[]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,c)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant78,i71,i72,i74,i75,i79,i80,i82,i83,'l0-1',l13,li71,li72,li74,li75,li79,li80,li82,li83,lo0boolean_true,lo73,lo76,lo77,lo81,lo84,o73,o76,o77,o81,o84])])),pred,[]),b(exists([b(identifier(o84),integer,[synthesis(minus1,output)]),b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)]),b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i71),integer,[synthesis(greater1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(6),integer,[])),pred,[]),b(conjunct(b(equal(b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant78),integer,[]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,c)]),b(integer(-2),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier('l1-2'),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant78,i71,i72,i74,i75,i79,i80,i82,i83,l04,'l1-2',li71,li72,li74,li75,li79,li80,li82,li83,lo0boolean_true,lo73,lo76,lo77,lo81,lo84,o73,o76,o77,o81,o84])])),pred,[]),b(exists([b(identifier(o84),integer,[synthesis(minus1,output)]),b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)]),b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i71),integer,[synthesis(greater1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(6),integer,[])),pred,[]),b(conjunct(b(equal(b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant78),integer,[]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l13),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant78,i71,i72,i74,i75,i79,i80,i82,i83,l00,l13,li71,li72,li74,li75,li79,li80,li82,li83,lo0boolean_true,lo73,lo76,lo77,lo81,lo84,o73,o76,o77,o81,o84])])),pred,[]),b(exists([b(identifier(o84),integer,[synthesis(minus1,output)]),b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)]),b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i71),integer,[synthesis(greater1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(6),integer,[])),pred,[]),b(conjunct(b(equal(b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant78),integer,[]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i72),integer,[synthesis(greater1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i74),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i75),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i79),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i82),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o77),integer,[synthesis(constant78,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(identifier(o76),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o73),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o76),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o81),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o77),integer,[synthesis(constant78,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(identifier(o84),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o81),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o84),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,c)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,d)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,c),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l11),integer,[synthesis(input(1)),synthesis(machinevar,d),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i71),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i74),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i75),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i79),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i82),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant78,i71,i72,i74,i75,i79,i80,i82,i83,l00,l11,li71,li72,li74,li75,li79,li80,li82,li83,lo0boolean_false,lo73,lo76,lo77,lo81,lo84,o73,o76,o77,o81,o84])])),pred,[]),
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,c)]),b(identifier(l1),integer,[synthesis(input(1)),synthesis(machinevar,d)]),b(identifier(li71),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(li72),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li74),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li75),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li79),integer,[synthesis(add1,left_input),synthesis(type,integer)]),b(identifier(li80),integer,[synthesis(add1,right_input),synthesis(type,integer)]),b(identifier(li82),integer,[synthesis(minus1,left_input),synthesis(type,integer)]),b(identifier(li83),integer,[synthesis(minus1,right_input),synthesis(type,integer)]),b(identifier(lo73),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo76),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo77),integer,[synthesis(constant78,output),synthesis(type,integer)]),b(identifier(lo81),integer,[synthesis(add1,output),synthesis(type,integer)]),b(identifier(lo84),integer,[synthesis(minus1,output),synthesis(type,integer)])],6,2,solution([binding(constant78,int(1),'1'),binding('l0-1',int(0),'0'),binding(l00,int(0),'0'),binding(l04,int(0),'0'),binding('l1-2',int(1),'1'),binding(l11,int(1),'1'),binding(l13,int(1),'1'),binding(li71,int(3),'3'),binding(li72,int(4),'4'),binding(li74,int(0),'0'),binding(li75,int(0),'0'),binding(li79,int(1),'1'),binding(li80,int(0),'0'),binding(li82,int(0),'0'),binding(li83,int(0),'0'),binding(lo0boolean_false,int(6),'6'),binding(lo0boolean_true,int(6),'6'),binding(lo73,int(6),'6'),binding(lo76,int(5),'5'),binding(lo77,int(4),'4'),binding(lo81,int(3),'3'),binding(lo84,int(2),'2'),binding(o77,int(-10),'-10')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    b_test_boolean_expression_wf(b(equivalence(Synthesized,b(greater(b(add(b(identifier(c),integer,[]),b(identifier(d),integer,[])),integer,[]),b(integer(1),integer,[])),pred,[])),pred,[]),[],_).

test(synthesis_3,[nondet]) :- 
    b_load_machine_from_file('../prob_examples/public_examples/Synthesis_Tests/lift.mch') , b_machine_precompile , 
    preferences:set_preference(time_out,10000) , 
    b_synthesis:prepare_synthesis_of_guard_by_examples([c],[[b(integer(0),integer,[synthesis(machinevar,c),synthesis(initial_example)])],[b(integer(1),integer,[synthesis(machinevar,c),synthesis(initial_example)])],[b(integer(2),integer,[synthesis(machinevar,c),synthesis(initial_example)])]],[[b(integer(-1),integer,[synthesis(machinevar,c),synthesis(initial_example)])],[b(integer(-2),integer,[synthesis(machinevar,c),synthesis(initial_example)])]],Positive,Negative,Library,ParameterAmount,M1,
                                       CurrentVars,BehavioralConstraint,LocationVars,Solution) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,CurrentVars,CurrentVars,any,guard,Positive,Negative,
                            BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,SynthesizedGuard,_) , 
    ! , SynthesizedGuard \= none , 
    b_test_boolean_expression_wf(b(equivalence(SynthesizedGuard,b(greater(b(identifier(c),integer,[]),b(integer(0),integer,[])),pred,[])),pred,[]),[],_).

test(synthesis_4,[nondet]) :- 
    b_load_machine_from_file('../prob_examples/public_examples/Synthesis_Tests/simple_vending_machine.mch') , b_machine_precompile , 
    library_setup(invariant,1,[b(integer(0),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],Library) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,[b(identifier(soda),integer,[nodeid(pos(10,1,4,2,4,5)),loc(local,simple_vending_machine,abstract_variables)])],[b(identifier(soda),integer,[nodeid(pos(10,1,4,2,4,5)),loc(local,simple_vending_machine,abstract_variables)])],[b(greater_equal(b(identifier(coins),integer,[nodeid(pos(13,1,4,13,4,17)),loc(local,simple_vending_machine,abstract_variables)]),b(integer(0),integer,[nodeid(pos(14,1,4,22,4,22))])),pred,[section(invariant),nodeid(pos(12,1,4,13,4,22))])],invariant,[[b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(2),integer,[synthesis(machinevar,soda)])],[b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(4),integer,[synthesis(machinevar,soda)])],[b(integer(5),integer,[synthesis(machinevar,soda)])]],[[b(integer(-1),integer,[synthesis(machinevar,soda)])],[b(integer(-1),integer,[synthesis(machinevar,soda)])],[b(integer(-2),integer,[synthesis(machinevar,soda)])]],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(o15),integer,[synthesis(constant16,output)]),b(interval(b(integer(-10),integer,[]),b(integer(11),integer,[])),set(integer),[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(interval(b(integer(0),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(member(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(interval(b(integer(1),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(interval(b(integer(1),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(interval(b(integer(1),integer,[]),b(integer(3),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(less(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[]),b(less(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(less(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l00),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l00,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l01),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l01,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l02),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l02,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l03),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l03,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l04),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l04,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(5),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(5),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(5),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(5),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(5),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(l05),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_true),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,l05,li10,li12,li13,li9,lo0boolean_true,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(-1),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier('l0-1'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,'l0-1',li10,li12,li13,li9,lo0boolean_false,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),b(exists([b(identifier(o15),integer,[synthesis(constant16,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)]),b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)]),b(integer(3),integer,[])),pred,[]),b(equal(b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(integer(0),integer,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(equivalence(b(greater(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(less_equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(i10),integer,[synthesis(greater1,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant16),integer,[]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i12),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(integer(-2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i13),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(integer(-2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(greater1,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(integer(-2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(identifier(o15),integer,[synthesis(constant16,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(integer(-2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(identifier(o14),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(greater1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[]),b(equal(b(identifier(o14),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[synthesis(output(0))])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)])),pred,[]),b(equal(b(identifier(o15),integer,[synthesis(constant16,output)]),b(integer(-2),integer,[synthesis(input(0)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[]),b(not_equal(b(identifier('l0-2'),integer,[synthesis(input(0)),synthesis(machinevar,soda),synthesis(type,integer)]),b(identifier(lo0boolean_false),integer,[synthesis(output(0)),synthesis(type,boolean)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(greater1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i12),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i13),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(greater1,left_input)]),b(interval(b(integer(-60),integer,[]),b(integer(66),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([constant16,i10,i12,i13,i9,'l0-2',li10,li12,li13,li9,lo0boolean_false,lo11,lo14,lo15,o11,o14,o15])])),pred,[]),
        [b(identifier(l0),integer,[synthesis(input(0)),synthesis(machinevar,soda)]),b(identifier(li10),integer,[synthesis(greater1,right_input),synthesis(type,integer)]),b(identifier(li12),integer,[synthesis(equal1,left_input),synthesis(type,integer)]),b(identifier(li13),integer,[synthesis(equal1,right_input),synthesis(type,integer)]),b(identifier(li9),integer,[synthesis(greater1,left_input),synthesis(type,integer)]),b(identifier(lo11),integer,[synthesis(greater1,output),synthesis(type,boolean)]),b(identifier(lo14),integer,[synthesis(equal1,output),synthesis(type,boolean)]),b(identifier(lo15),integer,[synthesis(constant16,output),synthesis(type,integer)])],3,1,solution([binding(constant16,int(-1),'-1'),binding('l0-1',int(0),'0'),binding('l0-2',int(0),'0'),binding(l00,int(0),'0'),binding(l01,int(0),'0'),binding(l02,int(0),'0'),binding(l03,int(0),'0'),binding(l04,int(0),'0'),binding(l05,int(0),'0'),binding(li10,int(1),'1'),binding(li12,int(0),'0'),binding(li13,int(0),'0'),binding(li9,int(0),'0'),binding(lo0boolean_false,int(3),'3'),binding(lo0boolean_true,int(3),'3'),binding(lo11,int(3),'3'),binding(lo14,int(2),'2'),binding(lo15,int(1),'1'),binding(o15,int(-10),'-10')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    b_test_boolean_expression_wf(b(equivalence(Synthesized,b(conjunct(b(greater(b(identifier(soda),integer,[]),b(integer(-1),integer,[])),pred,[]),b(greater_equal(b(identifier(coins),integer,[]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),[],_).

test(synthesis_5,[nondet]) :- 
    preferences:set_preference(time_out,10000) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[[b(integer(0),integer,[synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(1),integer,[synthesis(machinevar,coins)]),b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(2),integer,[synthesis(machinevar,coins)]),b(integer(2),integer,[synthesis(machinevar,soda)])],[b(integer(2),integer,[synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(0),integer,[synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(machinevar,soda)])]]),
        ([conjunct,equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,add,minus]),1,
        [b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],[b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],any,action,
        [[b(integer(2),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(3),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(4),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(4),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(6),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(5),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(12),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(9),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(8),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(13),integer,[synthesis(machinevar,soda),synthesis(initial_example)])]],
        [[b(integer(0),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(0),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(2),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(1),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(4),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(2),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(10),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(6),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(6),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(10),integer,[synthesis(machinevar,soda,synthesis(initial_example))])]],
        b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(less(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(less(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(less(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(less(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(less(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(less(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(less(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(less(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(o6),boolean,[synthesis(equal1,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i1),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant7),integer,[]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant12),integer,[]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o8,o6,o3,o25,o22,o19,o16,o13,o11,lovalue10,lovalue00,lo8,lo6,lo3,lo25,lo22,lo19,lo16,lo13,lo11,li9,li5,li4,li24,li23,li21,li20,li2,li18,li17,li15,li14,li10,li1,l13,l02,i9,i5,i4,i24,i23,i21,i20,i2,i18,i17,i15,i14,i10,i1,constant7,constant12])])),pred,[]),b(exists([b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(o6),boolean,[synthesis(equal1,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i1),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant7),integer,[]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant12),integer,[]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-88),integer,[]),b(integer(154),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o8,o6,o3,o25,o22,o19,o16,o13,o11,lovalue11,lovalue02,lo8,lo6,lo3,lo25,lo22,lo19,lo16,lo13,lo11,li9,li5,li4,li24,li23,li21,li20,li2,li18,li17,li15,li14,li10,li1,l14,l04,i9,i5,i4,i24,i23,i21,i20,i2,i18,i17,i15,i14,i10,i1,constant7,constant12])])),pred,[]),[b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(l04),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l14),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])],10,2,solution([binding(constant12,int(-2),'-2'),binding(constant7,int(3),'3'),binding(l02,int(0),'0'),binding(l04,int(0),'0'),binding(l13,int(1),'1'),binding(l14,int(1),'1'),binding(li1,int(2),'2'),binding(li10,int(0),'0'),binding(li14,int(8),'8'),binding(li15,int(0),'0'),binding(li17,int(1),'1'),binding(li18,int(7),'7'),binding(li2,int(2),'2'),binding(li20,int(0),'0'),binding(li21,int(0),'0'),binding(li23,int(0),'0'),binding(li24,int(0),'0'),binding(li4,int(0),'0'),binding(li5,int(0),'0'),binding(li9,int(0),'0'),binding(lo11,int(4),'4'),binding(lo13,int(8),'8'),binding(lo16,int(10),'10'),binding(lo19,int(9),'9'),binding(lo22,int(6),'6'),binding(lo25,int(5),'5'),binding(lo3,int(3),'3'),binding(lo6,int(2),'2'),binding(lo8,int(7),'7'),binding(lovalue00,int(10),'10'),binding(lovalue02,int(10),'10'),binding(lovalue10,int(9),'9'),binding(lovalue11,int(9),'9')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    (Synthesized = b(precondition(Guard,Action),subst,_)  
    ->  b_synthesis:parallel_to_conjunction(Action,ActionPred) , 
        solve_predicate(b(equivalence(Guard,b(conjunct(b(greater(b(identifier(coins),integer,[]),b(integer(1),integer,[])),pred,[]),b(greater(b(identifier(soda),integer,[]),b(integer(2),integer,[])),pred,[])),pred,[])),pred,[]),_,_) , 
        solve_predicate(b(equivalence(ActionPred,b(conjunct(b(equal(b(identifier('coins\''),integer,[nodeid(pos(5,1,2,11,2,15))]),b(minus(b(identifier(coins),integer,[]),b(integer(2),integer,[])),integer,[])),pred,[]),b(equal(b(identifier('soda\''),integer,[nodeid(pos(6,1,2,18,2,21))]),b(add(b(integer(-3),integer,[]),b(identifier(soda),integer,[])),integer,[])),pred,[])),pred,[])),pred,[]),_,_)
    ;   b_synthesis:parallel_to_conjunction(Synthesized,SynthesizedPred) , 
        solve_predicate(b(equivalence(SynthesizedPred,b(conjunct(b(equal(b(identifier('coins\''),integer,[nodeid(pos(5,1,2,11,2,15))]),b(minus(b(identifier(coins),integer,[]),b(integer(2),integer,[])),integer,[])),pred,[]),b(equal(b(identifier('soda\''),integer,[nodeid(pos(6,1,2,18,2,21))]),b(add(b(integer(-3),integer,[]),b(identifier(soda),integer,[])),integer,[])),pred,[])),pred,[])),pred,[]),_,_)).

test(synthesis_6,[nondet]) :- 
    preferences:set_preference(time_out,5000) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),([conjunct,equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,add,minus]),1,
        [b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],[b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],any,action,
        [[b(integer(0),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(0),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(2),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(1),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(3),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(5),integer,[synthesis(machinevar,soda),synthesis(initial_example)])]],
        [[b(integer(1),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(0),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(3),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(1),integer,[synthesis(machinevar,soda),synthesis(initial_example)])],[b(integer(4),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(5),integer,[synthesis(machinevar,soda),synthesis(initial_example)])]],
        b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(less(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(less(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(less(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(less(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(less(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(less(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(less(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(less(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(o6),boolean,[synthesis(equal1,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i1),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant7),integer,[]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant12),integer,[]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue01),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue10),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(1),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o8,o6,o3,o25,o22,o19,o16,o13,o11,lovalue10,lovalue01,lo8,lo6,lo3,lo25,lo22,lo19,lo16,lo13,lo11,li9,li5,li4,li24,li23,li21,li20,li2,li18,li17,li15,li14,li10,li1,l10,l00,i9,i5,i4,i24,i23,i21,i20,i2,i18,i17,i15,i14,i10,i1,constant7,constant12])])),pred,[]),b(exists([b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(o6),boolean,[synthesis(equal1,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i1),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant7),integer,[]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant12),integer,[]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(1),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue03),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o8,o6,o3,o25,o22,o19,o16,o13,o11,lovalue11,lovalue03,lo8,lo6,lo3,lo25,lo22,lo19,lo16,lo13,lo11,li9,li5,li4,li24,li23,li21,li20,li2,li18,li17,li15,li14,li10,li1,l11,l02,i9,i5,i4,i24,i23,i21,i20,i2,i18,i17,i15,i14,i10,i1,constant7,constant12])])),pred,[]),b(exists([b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(o6),boolean,[synthesis(equal1,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i1),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o6),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant7),integer,[]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(i10),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant12),integer,[]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i2),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i1),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i14),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i15),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i17),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i18),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i20),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o11),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i2),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i21),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i23),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i24),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i4),integer,[synthesis(equal1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i5),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(i9),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o3),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o11),boolean,[synthesis(equal2,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o16),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o13),integer,[synthesis(constant12,output),synthesis(type,constant-integer)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o19),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o16),integer,[synthesis(add1,output)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o22),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o19),integer,[synthesis(minus1,output)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o25),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o22),integer,[synthesis(add2,output)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o25),integer,[synthesis(minus2,output)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o3),boolean,[synthesis(conjunct1,output)]),b(identifier(o6),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o8),integer,[synthesis(constant7,output),synthesis(type,constant-integer)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(5),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue04),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue15),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(4),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(5),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i10),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i14),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i15),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i17),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i18),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i20),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i21),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i23),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i24),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i4),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i5),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i9),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(165),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o8,o6,o3,o25,o22,o19,o16,o13,o11,lovalue15,lovalue04,lo8,lo6,lo3,lo25,lo22,lo19,lo16,lo13,lo11,li9,li5,li4,li24,li23,li21,li20,li2,li18,li17,li15,li14,li10,li1,l15,l03,i9,i5,i4,i24,i23,i21,i20,i2,i18,i17,i15,i14,i10,i1,constant7,constant12])])),pred,[]),[b(identifier(li1),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li10),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li14),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li15),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li17),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li18),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li2),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li20),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li21),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li23),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li24),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(li4),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li5),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li9),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo11),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo13),integer,[synthesis(type,integer),synthesis(constant12,output),synthesis(type,constant-integer)]),b(identifier(lo16),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo19),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo22),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo25),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lo3),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo6),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo8),integer,[synthesis(type,integer),synthesis(constant7,output),synthesis(type,constant-integer)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l11),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(l03),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l15),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])],10,2,solution([binding(constant12,int(0),'0'),binding(constant7,int(1),'1'),binding(l00,int(0),'0'),binding(l02,int(0),'0'),binding(l03,int(0),'0'),binding(l10,int(1),'1'),binding(l11,int(1),'1'),binding(l15,int(1),'1'),binding(li1,int(2),'2'),binding(li10,int(0),'0'),binding(li14,int(7),'7'),binding(li15,int(0),'0'),binding(li17,int(1),'1'),binding(li18,int(5),'5'),binding(li2,int(2),'2'),binding(li20,int(0),'0'),binding(li21,int(0),'0'),binding(li23,int(0),'0'),binding(li24,int(0),'0'),binding(li4,int(0),'0'),binding(li5,int(0),'0'),binding(li9,int(0),'0'),binding(lo11,int(4),'4'),binding(lo13,int(8),'8'),binding(lo16,int(10),'10'),binding(lo19,int(9),'9'),binding(lo22,int(6),'6'),binding(lo25,int(5),'5'),binding(lo3,int(3),'3'),binding(lo6,int(2),'2'),binding(lo8,int(7),'7'),binding(lovalue01,int(10),'10'),binding(lovalue03,int(10),'10'),binding(lovalue04,int(10),'10'),binding(lovalue10,int(9),'9'),binding(lovalue11,int(9),'9'),binding(lovalue15,int(9),'9')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    b_synthesis:parallel_to_conjunction(Synthesized,SynthesizedPred) , 
    b_test_boolean_expression_wf(b(equivalence(SynthesizedPred,b(conjunct(b(equal(b(identifier('coins\''),integer,[nodeid(pos(5,1,2,11,2,15))]),b(minus(b(identifier(coins),integer,[]),b(integer(-1),integer,[])),integer,[])),pred,[]),b(equal(b(identifier('soda\''),integer,[nodeid(pos(6,1,2,18,2,21))]),b(add(b(integer(0),integer,[]),b(identifier(soda),integer,[])),integer,[])),pred,[])),pred,[])),pred,[]),[],_).

test(synthesis_7,[nondet]) :- 
    preferences:set_preference(time_out,5000) , 
    b_synthesis:synthesis_from_examples(b(truth,pred,[]),([],[]),([conjunct,equal-integer,constant-integer,equal-integer,constant-integer],[add,minus,add,minus]),1,
        [b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],[b(identifier(coins),integer,[nodeid(pos(5,1,2,11,2,15))]),b(identifier(soda),integer,[nodeid(pos(6,1,2,18,2,21))])],any,action,
        [[b(integer(0),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(2),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(6),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(6),integer,[synthesis(machinevar,soda),synthesis(initial_example)])]],
        [[b(integer(0),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(2),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(4),integer,[synthesis(machinevar,soda)])],[b(integer(6),integer,[synthesis(machinevar,coins),synthesis(initial_example)]),b(integer(7),integer,[synthesis(machinevar,soda),synthesis(initial_example)])]],
        b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(interval(b(integer(2),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(less(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(less(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(less(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(less(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(less(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(less(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(less(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(less(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(less(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[])),pred,[])),pred,[]),b(exists([b(identifier(o88),integer,[synthesis(minus2,output)]),b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o74),boolean,[synthesis(equal2,output)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i64),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant70),integer,[]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant75),integer,[]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue00),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue11),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(0),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(1),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o88,o85,o82,o79,o76,o74,o71,o69,o66,lovalue11,lovalue00,lo88,lo85,lo82,lo79,lo76,lo74,lo71,lo69,lo66,li87,li86,li84,li83,li81,li80,li78,li77,li73,li72,li68,li67,li65,li64,l10,l00,i87,i86,i84,i83,i81,i80,i78,i77,i73,i72,i68,i67,i65,i64,constant75,constant70])])),pred,[]),b(exists([b(identifier(o88),integer,[synthesis(minus2,output)]),b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o74),boolean,[synthesis(equal2,output)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i64),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant70),integer,[]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant75),integer,[]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(3),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue02),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue14),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(2),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(4),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o88,o85,o82,o79,o76,o74,o71,o69,o66,lovalue14,lovalue02,lo88,lo85,lo82,lo79,lo76,lo74,lo71,lo69,lo66,li87,li86,li84,li83,li81,li80,li78,li77,li73,li72,li68,li67,li65,li64,l13,l02,i87,i86,i84,i83,i81,i80,i78,i77,i73,i72,i68,i67,i65,i64,constant75,constant70])])),pred,[]),b(exists([b(identifier(o88),integer,[synthesis(minus2,output)]),b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o74),boolean,[synthesis(equal2,output)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(i64),boolean,[synthesis(conjunct1,left_input)])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(equal(b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)]),b(integer(9),integer,[])),pred,[]),b(equal(b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(10),integer,[])),pred,[])),pred,[]),b(conjunct(b(equal(b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(1),integer,[])),pred,[]),b(equal(b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(0),integer,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(disjunct(b(conjunct(b(conjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_true,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(conjunct(b(disjunct(b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(boolean_false,boolean,[])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(boolean_false,boolean,[])),pred,[])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant70),integer,[]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(conjunct(b(equivalence(b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_true,boolean,[])),pred,[])),pred,[]),b(equivalence(b(not_equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(o74),boolean,[synthesis(equal2,output)]),b(boolean_false,boolean,[])),pred,[])),pred,[])),pred,[])),pred,[]),b(equal(b(identifier(constant75),integer,[]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),integer,[]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),integer,[]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[]),b(equal(b(add(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),integer,[]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[]),b(equal(b(minus(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),integer,[]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(i65),boolean,[synthesis(conjunct1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i64),boolean,[synthesis(conjunct1,left_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o66),boolean,[synthesis(conjunct1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(i65),boolean,[synthesis(conjunct1,right_input)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i68),integer,[synthesis(equal1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i72),integer,[synthesis(equal2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i73),integer,[synthesis(equal2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i77),integer,[synthesis(add1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i78),integer,[synthesis(add1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i80),integer,[synthesis(minus1,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i81),integer,[synthesis(minus1,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i83),integer,[synthesis(add2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i84),integer,[synthesis(add2,right_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i86),integer,[synthesis(minus2,left_input)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(i87),integer,[synthesis(minus2,right_input)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o69),boolean,[synthesis(equal1,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o66),boolean,[synthesis(conjunct1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[]),b(equal(b(identifier(o69),boolean,[synthesis(equal1,output)]),b(identifier(o74),boolean,[synthesis(equal2,output)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o71),integer,[synthesis(constant70,output),synthesis(type,constant-integer)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[]),b(not_equal(b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o79),integer,[synthesis(add1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o76),integer,[synthesis(constant75,output),synthesis(type,constant-integer)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o82),integer,[synthesis(minus1,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o79),integer,[synthesis(add1,output)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o85),integer,[synthesis(add2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o82),integer,[synthesis(minus1,output)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(identifier(o88),integer,[synthesis(minus2,output)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o85),integer,[synthesis(add2,output)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(identifier(o88),integer,[synthesis(minus2,output)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(input(0)),synthesis(machinevar,coins)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(input(1)),synthesis(machinevar,soda)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[]),b(implication(b(equal(b(identifier(lovalue06),integer,[synthesis(type,integer),synthesis(output(0)),synthesis(machinevar,coins)]),b(identifier(lovalue17),integer,[synthesis(type,integer),synthesis(output(1)),synthesis(machinevar,soda)])),pred,[]),b(equal(b(integer(6),integer,[synthesis(output(0)),synthesis(machinevar,coins)]),b(integer(7),integer,[synthesis(output(1)),synthesis(machinevar,soda)])),pred,[])),pred,[])),pred,[])),pred,[]),b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(i67),integer,[synthesis(equal1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[]),b(member(b(identifier(i68),integer,[synthesis(equal1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i72),integer,[synthesis(equal2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i73),integer,[synthesis(equal2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i77),integer,[synthesis(add1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i78),integer,[synthesis(add1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i80),integer,[synthesis(minus1,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i81),integer,[synthesis(minus1,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i83),integer,[synthesis(add2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i84),integer,[synthesis(add2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i86),integer,[synthesis(minus2,left_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[]),b(member(b(identifier(i87),integer,[synthesis(minus2,right_input)]),b(interval(b(integer(-110),integer,[]),b(integer(176),integer,[])),set(integer),[])),pred,[])),pred,[])),pred,[])),pred,[used_ids([o88,o85,o82,o79,o76,o74,o71,o69,o66,lovalue17,lovalue06,lo88,lo85,lo82,lo79,lo76,lo74,lo71,lo69,lo66,li87,li86,li84,li83,li81,li80,li78,li77,li73,li72,li68,li67,li65,li64,l16,l06,i87,i86,i84,i83,i81,i80,i78,i77,i73,i72,i68,i67,i65,i64,constant75,constant70])])),pred,[]),[b(identifier(li64),integer,[synthesis(type,boolean),synthesis(conjunct1,left_input)]),b(identifier(li65),integer,[synthesis(type,boolean),synthesis(conjunct1,right_input)]),b(identifier(li67),integer,[synthesis(type,integer),synthesis(equal1,left_input)]),b(identifier(li68),integer,[synthesis(type,integer),synthesis(equal1,right_input)]),b(identifier(li72),integer,[synthesis(type,integer),synthesis(equal2,left_input)]),b(identifier(li73),integer,[synthesis(type,integer),synthesis(equal2,right_input)]),b(identifier(li77),integer,[synthesis(type,integer),synthesis(add1,left_input)]),b(identifier(li78),integer,[synthesis(type,integer),synthesis(add1,right_input)]),b(identifier(li80),integer,[synthesis(type,integer),synthesis(minus1,left_input)]),b(identifier(li81),integer,[synthesis(type,integer),synthesis(minus1,right_input)]),b(identifier(li83),integer,[synthesis(type,integer),synthesis(add2,left_input)]),b(identifier(li84),integer,[synthesis(type,integer),synthesis(add2,right_input)]),b(identifier(li86),integer,[synthesis(type,integer),synthesis(minus2,left_input)]),b(identifier(li87),integer,[synthesis(type,integer),synthesis(minus2,right_input)]),b(identifier(lo66),integer,[synthesis(type,boolean),synthesis(conjunct1,output)]),b(identifier(lo69),integer,[synthesis(type,boolean),synthesis(equal1,output)]),b(identifier(lo71),integer,[synthesis(type,integer),synthesis(constant70,output),synthesis(type,constant-integer)]),b(identifier(lo74),integer,[synthesis(type,boolean),synthesis(equal2,output)]),b(identifier(lo76),integer,[synthesis(type,integer),synthesis(constant75,output),synthesis(type,constant-integer)]),b(identifier(lo79),integer,[synthesis(type,integer),synthesis(add1,output)]),b(identifier(lo82),integer,[synthesis(type,integer),synthesis(minus1,output)]),b(identifier(lo85),integer,[synthesis(type,integer),synthesis(add2,output)]),b(identifier(lo88),integer,[synthesis(type,integer),synthesis(minus2,output)]),b(identifier(l00),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l10),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(l02),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l13),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)]),b(identifier(l06),integer,[synthesis(type,integer),synthesis(input(0)),synthesis(machinevar,coins)]),b(identifier(l16),integer,[synthesis(type,integer),synthesis(input(1)),synthesis(machinevar,soda)])],10,2,solution([binding(constant70,int(0),'0'),binding(constant75,int(-1),'-1'),binding(l00,int(0),'0'),binding(l02,int(0),'0'),binding(l06,int(0),'0'),binding(l10,int(1),'1'),binding(l13,int(1),'1'),binding(l16,int(1),'1'),binding(li64,int(2),'2'),binding(li65,int(2),'2'),binding(li67,int(0),'0'),binding(li68,int(0),'0'),binding(li72,int(0),'0'),binding(li73,int(0),'0'),binding(li77,int(5),'5'),binding(li78,int(0),'0'),binding(li80,int(1),'1'),binding(li81,int(7),'7'),binding(li83,int(0),'0'),binding(li84,int(0),'0'),binding(li86,int(0),'0'),binding(li87,int(0),'0'),binding(lo66,int(3),'3'),binding(lo69,int(4),'4'),binding(lo71,int(8),'8'),binding(lo74,int(2),'2'),binding(lo76,int(7),'7'),binding(lo79,int(10),'10'),binding(lo82,int(9),'9'),binding(lo85,int(6),'6'),binding(lo88,int(5),'5'),binding(lovalue00,int(10),'10'),binding(lovalue02,int(10),'10'),binding(lovalue06,int(10),'10'),binding(lovalue11,int(9),'9'),binding(lovalue14,int(9),'9'),binding(lovalue17,int(9),'9')]),Synthesized,_) , 
    ! , Synthesized \= none , 
    b_synthesis:parallel_to_conjunction(Synthesized,SynthesizedPred) , 
    b_test_boolean_expression_wf(b(equivalence(SynthesizedPred,b(conjunct(b(equal(b(identifier('coins\''),integer,[nodeid(pos(5,1,2,11,2,15))]),b(minus(b(add(b(identifier(coins),integer,[]),b(identifier(coins),integer,[])),integer,[]),b(identifier(coins),integer,[])),integer,[])),pred,[]),b(equal(b(identifier('soda\''),integer,[nodeid(pos(6,1,2,18,2,21))]),b(add(b(integer(1),integer,[]),b(identifier(soda),integer,[])),integer,[])),pred,[])),pred,[])),pred,[]),[],_).

:- end_tests(synthesis).


:- begin_tests(synthesis_util).

test(get_single_constant_or_set_operator_id,[]) :- 
    synthesis_util:get_single_constant_or_set_operator_id(seq-integer_set,OutputId1,ValueId1) , 
    OutputId1 = b(identifier(_),set(seq(integer)),[synthesis(type,seq-integer_set)]) , 
    ValueId1 = b(seq(b(integer_set('INTEGER'),set(integer),[])),set(seq(integer)),[]) , 

    synthesis_util:get_single_constant_or_set_operator_id(seq-integer_set_natural1,OutputId2,ValueId2) , 
    OutputId2 = b(identifier(_),set(seq(integer)),[synthesis(type,seq-integer_set_natural1)]) , 
    ValueId2 = b(seq(b(integer_set('NATURAL1'),set(integer),[])),set(seq(integer)),[]) , 

    synthesis_util:get_single_constant_or_set_operator_id(seq-string_set,OutputId3,ValueId3) , 
    OutputId3 = b(identifier(_),set(seq(string)),[synthesis(type,seq-string_set)]) , 
    ValueId3 = b(seq(b(string_set,set(string),[])),set(seq(string)),[]) , 

    synthesis_util:get_single_constant_or_set_operator_id(constant-set(seq(integer)),OutputId4,ValueId4) , 
    OutputId4 = b(identifier(_),set(seq(integer)),[synthesis(type,constant-set(seq(integer)))]) , 
    ValueId4 = b(identifier(ConstantName4),set(seq(integer)),[]) , 
    synthesis_util:delete_numbers_from_atom(ConstantName4,RawConstantName4) , 
    RawConstantName4 = constant , 

    synthesis_util:get_single_constant_or_set_operator_id(constant-string,OutputId5,ValueId5) , 
    OutputId5 = b(identifier(_),string,[synthesis(type,constant-string)]) , 
    ValueId5 = b(identifier(ConstantName5),string,[]) , 
    synthesis_util:delete_numbers_from_atom(ConstantName5,RawConstantName5) , 
    RawConstantName5 = constant , 

    synthesis_util:get_single_constant_or_set_operator_id(constant-set(boolean),OutputId6,ValueId6) , 
    OutputId6 = b(identifier(_),set(boolean),[synthesis(type,constant-set(boolean))]) , 
    ValueId6 = b(identifier(ConstantName6),set(boolean),[]) , 
    synthesis_util:delete_numbers_from_atom(ConstantName6,RawConstantName6) , 
    RawConstantName6 = constant.

% true
%contains_equivalent_state([b(integer(-1),integer,[synthesis(machinevar,soda)])],[[b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(0),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(1),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(2),integer,[synthesis(machinevar,soda)])],[b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(4),integer,[synthesis(machinevar,soda)])],[b(unary_minus(b(integer(1),integer,[nodeid(none)])),integer,[synthesis(machinevar,soda),nodeid(none)])]]).
% false
%contains_equivalent_state([b(integer(-2),integer,[synthesis(machinevar,soda)])],[[b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(0),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(1),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(2),integer,[synthesis(machinevar,soda)])],[b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(4),integer,[synthesis(machinevar,soda)])],[b(unary_minus(b(integer(1),integer,[nodeid(none)])),integer,[synthesis(machinevar,soda),nodeid(none)])]]).
% true
%contains_equivalent_state([b(integer(-2),integer,[synthesis(machinevar,soda)])],[[b(integer(0),integer,[synthesis(machinevar,soda)])],[b(integer(0),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(1),integer,[synthesis(machinevar,soda)])],[b(integer(1),integer,[synthesis(machinevar,soda),nodeid(none)])],[b(integer(2),integer,[synthesis(machinevar,soda)])],[b(integer(3),integer,[synthesis(machinevar,soda)])],[b(integer(4),integer,[synthesis(machinevar,soda)])],[b(unary_minus(b(integer(2),integer,[nodeid(none)])),integer,[synthesis(machinevar,soda),nodeid(none)])]]).
:- end_tests(synthesis_util).
